/* Computes a sequence of MCMC-EM map estimates. An error model 
   is *** NOT *** used.                  */
int mcemits (int sxs_flg, int sz_flg, int nsij, int *conv,
             int ***Recounts, double **Temp, double **mle, double **probs)
{
   int   i, j, c, jj, t, tally = 0;
   int   nscans;
   double **rcz;

   conv[0] = 1;                         /* assume convergence */
   rcz = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, rcz);
   for (i = 0; i < 2; i++)
   {
      rcz[i] = malloc (3 * sizeof (double));
      alloc_chk (WHERE, rcz[i]);
      for (j = 0; j < 3; j++)
         rcz[i][j] = 0;
   }

   nscans = NIt / 2;
   while (!sz_flg)
   {
      tally++;
      nscans = nscans * 2;

      for (j = 0; j < NMarks; j++)
         for (c = 0; c < n_compnts; c++)
            for (jj = 0; jj < 6; ++jj)
               Recounts[c][j][jj] = 0;

      for (t = 0; t < nscans; ++t)
      {
         sampling (0, 1, NMarks, nsij, nFGL, probs);
         for (c = 0; c < n_compnts; ++c)
            full_seg_prob (NMarks, c, Recounts, Temp);
      }
      for (j = 1; j < NMarks; j++)
      {
         set_rc (j, Recounts, rcz);
         if (sxs_flg)                   /* generate sxs map estimates */
         {
            for (i = 0; i < 2; i++)
            {
               mle[i][j] = rcz[i][0] / rcz[i][2];
               if (mle[i][j] < RF_bound)
                  mle[i][j] = RF_bound;
               if (mle[i][j] > (.5 - RF_bound))
                  mle[i][j] = .5 - RF_bound;
            }
         }
         else                           /* generate sxa map estimates */
         {
            mle[0][j] = (rcz[0][0] + rcz[1][0]) / (rcz[0][2] + rcz[1][2]);
            if (mle[0][j] < RF_bound)
               mle[0][j] = RF_bound;
            if (mle[0][j] > (.5 - RF_bound))
               mle[0][j] = .5 - RF_bound;
            mle[1][j] = mle[0][j];
         }
      }
      sz_flg = comp_map (EM_bound, mle, Marek);
      reset_map (1, nsij, mle, Marek, probs);
      if (tally >= ItEM)
      {
         sz_flg = 1;
         conv[0] = 0;
      }                                 /* toggle convergence flag */
   }
   for (i = 0; i < 2; i++)
      free (rcz[i]);
   free (rcz);
   return (nscans);
}

/* Computes the observed information. An error model is NOT used.   */
void obs_info (int sxs_flg, double **oi_mat, int ***Scount,
               int ***Recounts, int nsij, double **probs,
               double **Temp, double **mle, int T, int R)
{
   int   c, i, j, t, jj, nLoci;
   double **rcz;

   rcz = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, rcz);
   for (i = 0; i < 2; i++)
   {
      rcz[i] = malloc (3 * sizeof (double));
      alloc_chk (WHERE, rcz[i]);
      for (j = 0; j < 3; j++)
         rcz[i][j] = 0;
   }

   nLoci = 2 * NMarks;

   for (j = 0; j < NMarks; j++)
      for (i = 0; i < n_compnts; i++)
         for (jj = 0; jj < 6; jj++)
            Scount[i][j][jj] = Recounts[i][j][jj] = 0;
   for (i = 0; i < nLoci; i++)
      for (j = 0; j < nLoci; j++)
         oi_mat[i][j] = 0;

   for (t = 0; t < T; t++)
   {
      sampling (0, 1, NMarks, nsij, nFGL, probs);
      for (c = 0; c < n_compnts; ++c)
         full_seg_prob (NMarks, c, Recounts, Temp);
      for (j = 1; j < NMarks; j++)
         for (c = 0; c < n_compnts; c++)
            for (jj = 0; jj < 6; jj++)
               Scount[c][j][jj] = Recounts[c][j][jj] - Scount[c][j][jj];
      if (sxs_flg)                      /* handle the sex-specific case */
      {
         for (j = 1; j < NMarks; j++)
         {
            set_rc (j, Scount, rcz);
            for (i = 0; i < 2; i++)
               oi_mat[(2 * j - 1 + i)][0] = rcz[i][0] / mle[i][j]
                  - rcz[i][1] / (1 - mle[i][j]);
         }
         for (j = 1; j < (nLoci - 1); j++)
            for (i = 1; i < (nLoci - 1); i++)
               oi_mat[j][i] += oi_mat[j][0] * oi_mat[i][0];

         for (j = 1; j < NMarks; j++)
            for (c = 0; c < n_compnts; c++)
               for (jj = 0; jj < 6; jj++)
                  Scount[c][j][jj] = Recounts[c][j][jj];
      }
      else                              /* handle the sex-averaged case */
      {
         for (j = 1; j < NMarks; j++)
         {
            set_rc (j, Scount, rcz);
            oi_mat[j][0] = (rcz[0][0] + rcz[1][0]) / mle[0][j]
               - (rcz[0][1] + rcz[1][1]) / (1 - mle[0][j]);
         }
         for (j = 1; j < NMarks; j++)
            for (i = 1; i < NMarks; i++)
               oi_mat[j][i] += oi_mat[j][0] * oi_mat[i][0];

         for (j = 1; j < NMarks; j++)
            for (c = 0; c < n_compnts; c++)
               for (jj = 0; jj < 6; jj++)
                  Scount[c][j][jj] = Recounts[c][j][jj];
      }
   }
   if (sxs_flg)                         /* more sex-specific components of obs_info */
   {
      for (i = 1; i < (nLoci - 1); i++) /* compute E(delta^2)    */
         for (j = 1; j < (nLoci - 1); j++)
            oi_mat[i][j] = oi_mat[i][j] / T;

      for (j = 1; j < NMarks; j++)      /* add E(delta(delta))        */
      {
         set_rc (j, Recounts, rcz);
         for (i = 0; i < 2; i++)
         {
            oi_mat[(2 * j - 1 + i)][(2 * j - 1 + i)] -=
               (rcz[i][0] / pow (mle[i][j], 2) + rcz[i][1] /
                pow ((1 - mle[i][j]), 2)) / T;
            oi_mat[(2 * j - 1 + i)][0] =
               (rcz[i][0] / mle[i][j] - rcz[i][1] / (1 - mle[i][j])) / T;
         }
      }
      if (R == RedoEst)
      {
         printf ("\n\n ESTIMATED GRADIENT LOG-LIKE: ");
      }
      for (j = 1; j < (nLoci - 1); j++) /* subtract [E(delta)]^2  */
      {
         if (R == RedoEst)
            printf (" %6.4f ", oi_mat[j][0]);
         for (i = 1; i < (nLoci - 1); i++)
            oi_mat[j][i] -= oi_mat[j][0] * oi_mat[i][0];
      }
      for (i = 1; i < (nLoci - 1); i++) /* negate obs_info_mat  */
         for (j = 1; j < (nLoci - 1); j++)
            oi_mat[i][j] = -oi_mat[i][j]; /*observed information done */
      if (R == RedoEst)
      {
         printf ("\n\n Partial derivatives taken in the order:"
                 " [F1, M1, F2, M2, ...]");
      }
   }
   else                                 /* more sex-averaged compnoents of obs_info */
   {
      for (i = 1; i < NMarks; i++)
         for (j = 1; j < NMarks; j++)
            oi_mat[i][j] = oi_mat[i][j] / T; /* E((delta)^2) done */

      for (j = 1; j < NMarks; j++)      /* add E(delta(delta))    */
      {
         set_rc (j, Recounts, rcz);
         oi_mat[j][j] -= ((rcz[0][0] + rcz[1][0]) / pow (mle[0][j], 2)
                          + (rcz[0][1] + rcz[1][1]) / pow ((1 - mle[0][j]),
                                                           2)) / T;

         oi_mat[j][0] = ((rcz[0][0] + rcz[1][0]) / mle[0][j] -
                         (rcz[0][1] + rcz[1][1]) / (1 - mle[0][j])) / T;
      }
      if (R == RedoEst)
         printf ("\n\n ESTIMATED GRADIENT of LOG-LIKE: ");
      for (j = 1; j < NMarks; j++)      /* subtract [E(delta)]^2  */
      {
         if (R == RedoEst)
            printf (" %6.4f ", oi_mat[j][0]);
         for (i = 1; i < NMarks; i++)
            oi_mat[j][i] -= oi_mat[j][0] * oi_mat[i][0];
      }
      for (i = 1; i < NMarks; i++)      /* negate obs_info_mat */
         for (j = 1; j < NMarks; j++)
            oi_mat[i][j] = -oi_mat[i][j]; /* observed information done  */
   }
   for (i = 0; i < 2; i++)
      free (rcz[i]);
   free (rcz);
}

/* Inverts a matrix of doubles (created dynamically) [nnr x nnr],
the input and the result are both one indexed.                      */
void d_invert (double **oi_mat, double **ioi_mat, int rows)
{
   /* ioi_mat holds inv_obs_info */
   int   nLoci, jnr, inr, *indxnr, cols = rows;
   double d = 0, *bnr;

   nLoci = 2 * NMarks;

   indxnr = malloc (nLoci * sizeof (int));
   alloc_chk (WHERE, indxnr);
   bnr = malloc (nLoci * sizeof (double));
   alloc_chk (WHERE, bnr);

   ludcmp (oi_mat, rows, indxnr, &d);
   for (jnr = 1; jnr <= cols; jnr++)
   {
      for (inr = 1; inr <= rows; inr++)
         bnr[inr] = 0.0;
      bnr[jnr] = 1.0;
      lubksb (oi_mat, rows, indxnr, bnr);
      for (inr = 1; inr <= rows; inr++)
         ioi_mat[inr][jnr] = bnr[inr];
   }

   free (indxnr);
   free (bnr);
}

/* Part of matrix inversion routine  */
void ludcmp (double **anr, int nnr, int *indxnr, double *d)
{
   int   i, imax, j, k;
   double big, dum, sum, temp;
   double *vv;

   vv = malloc ((nnr + 2) * sizeof (double));
   alloc_chk (WHERE, vv);
   for (i = 0; i <= (nnr + 1); i++)
      vv[i] = 0;

   d[0] = 1.0;
   for (i = 1; i <= nnr; i++)
   {
      big = 0.0;
      for (j = 1; j <= nnr; j++)
         if ((temp = fabs (anr[i][j])) > big)
            big = temp;
      if (big == 0.0)
         printf ("\n Singular matrix in routine LUDCMP");
      vv[i] = 1.0 / big;
   }
   for (j = 1; j <= nnr; j++)
   {
      for (i = 1; i < j; i++)
      {
         sum = anr[i][j];
         for (k = 1; k < i; k++)
            sum -= anr[i][k] * anr[k][j];
         anr[i][j] = sum;
      }
      big = 0.0;
      for (i = j; i <= nnr; i++)
      {
         sum = anr[i][j];
         for (k = 1; k < j; k++)
            sum -= anr[i][k] * anr[k][j];
         anr[i][j] = sum;
         if ((dum = vv[i] * fabs (sum)) >= big)
         {
            big = dum;
            imax = i;
         }
      }
      if (j != imax)
      {
         for (k = 1; k <= nnr; k++)
         {
            dum = anr[imax][k];
            anr[imax][k] = anr[j][k];
            anr[j][k] = dum;
         }
         d[0] = -(d[0]);
         vv[imax] = vv[j];
      }
      indxnr[j] = imax;
      if (anr[j][j] == 0.0)
         anr[j][j] = TINY;
      if (j != nnr)
      {
         dum = 1.0 / (anr[j][j]);
         for (i = j + 1; i <= nnr; i++)
            anr[i][j] *= dum;
      }
   }
   free (vv);
}

/* Part of matrix inversion routine   */
void lubksb (double **anr, int nnr, int *indxnr, double *bnr)
{
   int   i, ii = 0, ip, j;
   double sum;

   for (i = 1; i <= nnr; i++)
   {
      ip = indxnr[i];
      sum = bnr[ip];
      bnr[ip] = bnr[i];
      if (ii)
         for (j = ii; j <= i - 1; j++)
            sum -= anr[i][j] * bnr[j];
      else if (sum)
         ii = i;
      bnr[i] = sum;
   }
   for (i = nnr; i >= 1; i--)
   {
      sum = bnr[i];
      for (j = i + 1; j <= nnr; j++)
         sum -= anr[i][j] * bnr[j];
      bnr[i] = sum / anr[i][i];
   }
}

/* Calculates the left and right comp.data log likelihood-ratios
   from a single realized matrix S.                                  */
void lr_cdllr (int nloc, double *cdllr, double **se, double **mt, double **tt,
               int ***scount, int ***recounts)
{
   int   c, j, jj;

   double f_r = 0.0;                    /* stores total F recombs at a locus  */
   double m_r = 0.0;                    /* stores total M recombs at a locus  */
   double f_nr = 0.0;                   /* stores total F non-reco at a locus */
   double m_nr = 0.0;                   /* stores total M non-reco at a locus */

   cdllr[0] = 0;
   cdllr[1] = 0;
   for (j = 1; j < nloc; j++)
      for (c = 0; c < n_compnts; c++)
         for (jj = 0; jj < 6; jj++)
            scount[c][j][jj] = recounts[c][j][jj] - scount[c][j][jj];
   for (j = 1; j < nloc; j++)
   {
      m_r = 0;
      f_r = 0;
      m_nr = 0;
      f_nr = 0;
      for (c = 0; c < n_compnts; c++)
      {
         m_r += scount[c][j][1];
         f_r += scount[c][j][4];
         m_nr += scount[c][j][2];
         f_nr += scount[c][j][5];
      }
      cdllr[0] += (f_r * log (se[0][j] / mt[0][j])
                   + m_r * log (se[1][j] / mt[1][j])
                   + f_nr * log ((1 - se[0][j]) / (1 - mt[0][j]))
                   + m_nr * log ((1 - se[1][j]) / (1 - mt[1][j])));

      cdllr[1] += (f_r * log (tt[0][j] / mt[0][j])
                   + m_r * log (tt[1][j] / mt[1][j])
                   + f_nr * log ((1 - tt[0][j]) / (1 - mt[0][j]))
                   + m_nr * log ((1 - tt[1][j]) / (1 - mt[1][j])));
   }
   for (j = 1; j < nloc; j++)
      for (c = 0; c < n_compnts; c++)
         for (jj = 0; jj < 6; jj++)
            scount[c][j][jj] = recounts[c][j][jj];
}

/* Calculate a stochastic approx. iteration                         */
void sait (int sxs_flg, int vv, int ***Recounts, int nsij,
           double **probs, double **Temp, double **ioi_mat, double **mle)
{
   int   i, j, jj, c, nLoci, t;
   double **rcz, *temp, gama;

   nLoci = 2 * NMarks;

   temp = malloc (nLoci * sizeof (double));
   rcz = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, rcz);
   for (i = 0; i < 2; i++)
   {
      rcz[i] = malloc (3 * sizeof (double));
      alloc_chk (WHERE, rcz[i]);
      for (j = 0; j < 3; j++)
         rcz[i][j] = 0;
   }
   for (j = 0; j < NMarks; j++)
      for (c = 0; c < n_compnts; c++)
         for (jj = 0; jj < 6; jj++)
            Recounts[c][j][jj] = 0;

   for (t = 0; t < NSAGradIt; ++t)
   {
      sampling (0, 1, NMarks, nsij, nFGL, probs);
      for (c = 0; c < n_compnts; ++c)
         full_seg_prob (NMarks, c, Recounts, Temp);
   }
   if (sxs_flg)                         /* generate SA sex-specific est. */
   {
      for (j = 1; j < NMarks; j++)
      {
         set_rc (j, Recounts, rcz);
         for (i = 0; i < 2; i++)
         {
            gama = (rcz[i][0] / Marek[i][j] - rcz[i][1] /
                    (1 - Marek[i][j])) / NSAGradIt;
            temp[2 * j - 1 + i] = gama / (vv + 1);
         }
      }
      for (j = 1; j < (nLoci - 1); j++)
      {
         gama = 0;
         for (i = 1; i < (nLoci - 1); i++)
            gama += (temp[i] * ioi_mat[j][i]);
         if (j % 2)
         {
            i = (j + 1) / 2;
            mle[0][i] += gama;
            if (mle[0][i] < RF_bound)
               mle[0][i] = RF_bound;
            if (mle[0][i] > (.5 - RF_bound))
               mle[0][i] = .5 - RF_bound;
         }
         else
         {
            i = j / 2;
            mle[1][i] += gama;
            if (mle[1][i] < RF_bound)
               mle[1][i] = RF_bound;
            if (mle[1][i] > (.5 - RF_bound))
               mle[1][i] = .5 - RF_bound;
         }
      }
   }
   else                                 /* generate SA sex-averaged est. */
   {
      for (j = 1; j < NMarks; j++)
      {
         set_rc (j, Recounts, rcz);
         gama = ((rcz[0][0] + rcz[1][0]) / Marek[0][j] -
                 (rcz[0][1] + rcz[1][1]) / (1 - Marek[0][j])) / NSAGradIt;
         temp[j] = gama / (vv + 1);
      }
      for (j = 1; j < NMarks; j++)
      {
         gama = 0;
         for (i = 1; i < NMarks; i++)
            gama += (temp[i] * ioi_mat[j][i]);
         mle[0][j] += gama;
         if (mle[0][j] < RF_bound)
            mle[0][j] = RF_bound;
         if (mle[0][j] > (.5 - RF_bound))
            mle[0][j] = .5 - RF_bound;
         mle[1][j] = mle[0][j];
      }
   }
   free (temp);
   for (i = 0; i < 2; i++)
      free (rcz[i]);
   free (rcz);
}

/* Computes the log likelihood ratio of map2 over map1.         */
double logLR (double **map1, double **map2, int nsij,
              int ***Scount, int ***Recounts, double **probs,
              double **Temp, int *alf, int ss)
{
   int   c, i, j, jj, t;
   double **tt, **mt, sum_cdlr[2], cdllr[2];

   tt = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, tt);
   mt = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, mt);
   for (i = 0; i < 2; i++)
   {
      tt[i] = malloc ((NMarks + 1) * sizeof (double));
      alloc_chk (WHERE, tt[i]);
      mt[i] = malloc ((NMarks + 1) * sizeof (double));
      alloc_chk (WHERE, mt[i]);
   }

   for (i = 0; i < 2; i++)              /* compute tt, and mt  */
      for (j = 1; j < NMarks; j++)
      {
         tt[i][j] = find_intv (map1[i][j], map2[i][j], alf);
         mt[i][j] = rf ((0.5 * (cmH (tt[i][j]) + cmH (map1[i][j]))));
      }
   reset_map (1, nsij, mt, Marek, probs);
   for (i = 0; i < n_compnts; i++)
      for (j = 0; j < NMarks; j++)
         for (jj = 0; jj < 6; jj++)
            Scount[i][j][jj] = Recounts[i][j][jj] = 0;
   sum_cdlr[0] = sum_cdlr[1] = 0;
   for (t = 0; t < ss; ++t)
   {
      sampling (0, 1, NMarks, nsij, nFGL, probs);
      for (c = 0; c < n_compnts; ++c)
         full_seg_prob (NMarks, c, Recounts, Temp);

      lr_cdllr (NMarks, cdllr, map1, mt, tt, Scount, Recounts);

      sum_cdlr[0] += exp (cdllr[0]);
      sum_cdlr[1] += exp (cdllr[1]);
   }
   cpy_map (tt, map1);

   for (i = 0; i < 2; i++)
   {
      free (tt[i]);
      free (mt[i]);
   }
   free (tt);
   free (mt);
   return (log (sum_cdlr[1] / sum_cdlr[0]));
}

/* Resets the values of some global variables based on nscans */
void reset_globs (int sxs_flg, int nscans, int *g, int *oia, int *ois)
{
   if (sxs_flg)
   {
      if (NHybCurvIt < (4 * nscans))
         NHybCurvIt = 4 * nscans;
      if (NHybAscentIt < (nscans / 2))
         NHybAscentIt = nscans / 2;
      if (NSAGradIt < (nscans / 2))
         NSAGradIt = nscans / 2;
      if (NChiSqIt < (4 * nscans))
         NChiSqIt = 4 * nscans;

      ois[0] = 8 * nscans;
      g[0] = 4 * nscans;
   }
   else
   {
      if (NHybCurvIt < (2 * nscans))
         NHybCurvIt = 2 * nscans;
      if (NHybAscentIt < (nscans / 4))
         NHybAscentIt = nscans / 4;
      if (NSAGradIt < (nscans / 4))
         NSAGradIt = nscans / 4;
      if (NChiSqIt < (2 * nscans))
         NChiSqIt = 2 * nscans;

      oia[0] = 4 * nscans;
      g[0] = 2 * nscans;
   }
}
