/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#define MAX_CHAR_IN_LINE 10000
#define  ITEM_SEPERATOR "    "

typedef struct node_tag_int
{
   int   value;
   struct node_tag_int *next;
}
nodeInt;

typedef struct node_tag_double
{
   double value;
   struct node_tag_double *next;
}
nodeDouble;

typedef struct freq_dist_tag
{
   double value;
   double frequency;
   double cum_frequency;
   struct freq_dist_tag *next;
}
freq_dist;

typedef struct line_info_tag
{
   int   line_number;
   freq_dist *freq_table;
   struct line_info_tag *next;
}
line_info_struct;

void  readDataFromFile (FILE * stream, line_info_struct ** linesHeadPtr,
                        int Total_Iter);
void  ReadInputFileData (line_info_struct ** linesListPtr, int TotalIter);

void  freeLineList (line_info_struct ** HeadPtr);
void  freeDoubleList (nodeDouble ** nodePtr);

void  calculate_p_value_new (line_info_struct * linesListPtr, int locus,
                             int stat, int NoOfLocus, int iteration_No,
                             int NoOfStat, int Total_Iter,
                             int TestFlag, FILE * write_fp, int STAT_FLAG,
                             double *QUANT_OPT);

int   pairwise_comparison_withList (int locus, line_info_struct * HeadPtr,
                                    int a, int b);
int   getDistTableSize (line_info_struct * HeadPtr);
freq_dist *getDistEntry (int pos, line_info_struct * HeadPtr);
line_info_struct *getLineInfo (int line_no, line_info_struct * HeadPtr);
void  Average_Dist_withList (int locus1, int stat, int NoOfLocus,
                             int iteration_No, int NoOfStat,
                             line_info_struct * HeadPtr);
double Anderson_Dist_withList (line_info_struct * HeadPtr, int number);
double p_val_double_withList (nodeDouble * headPtr);
double p_val_int_withList (nodeInt * headPtr);
double Quantile_withList (line_info_struct * HeadPtr, int Row, double Prop);
double Mean_withList (line_info_struct * HeadPtr, int Row, int Total_Iter,
                      int NoOfLocus, int NoOfStat, int locus);
double Var_withList (line_info_struct * HeadPtr, int Row, int Total_Iter,
                     int NoOfLocus, int NoOfStat);
freq_dist *g_ModifiedArray = NULL;
int   g_NoOfDistinctElement = 0;

void freeLineList (line_info_struct ** HeadPtr)
{
   if (HeadPtr == NULL)
   {
      return;
   }
   else
   {
      line_info_struct *tempPtr = *HeadPtr;
      while (tempPtr != NULL)
      {
         line_info_struct *temp = NULL;
         freq_dist *distPtr = NULL;
         distPtr = tempPtr->freq_table;
         while (distPtr != NULL)
         {
            freq_dist *distPtr1 = NULL;
            distPtr1 = distPtr->next;
            free (distPtr);
            distPtr = distPtr1;
         }
         temp = tempPtr->next;
         free (tempPtr);
         tempPtr = temp;
      }
      *HeadPtr = NULL;
   }
}

void readDataFromFile (FILE * stream, line_info_struct ** linesHeadPtr,
                       int Total_Iter)
{
   int   seperator_string_len = strlen (ITEM_SEPERATOR);
   char  buffer[MAX_CHAR_IN_LINE];
   int   lineno = 0;

   line_info_struct *LineInfoPtr = NULL;

   if (stream == NULL)
   {
      printf
         ("Invalid File Pointer passed in readDataFromFile(). Exiting... \n");
      exit (EXIT_FAILURE);
   }
   while (fgets (buffer, MAX_CHAR_IN_LINE, stream) != NULL)
   {
      freq_dist *freqHeadPtr = NULL;
      freq_dist *currentNodePtr = NULL;
      int   line_length = strlen (buffer);
      int   offset = 0;
      double cum_total_upto = 0;
      int   nNoOfVal = 0;
      line_info_struct *LinePtr;

      LinePtr = (line_info_struct *) calloc (1, sizeof (line_info_struct));
      alloc_chk (WHERE, LinePtr);

      LinePtr->line_number = lineno++;

      while (offset < line_length)
      {
         char *tempPtr1 = strstr (buffer + offset, ITEM_SEPERATOR);
         double val;
         double frequency;

         if (freqHeadPtr == NULL)
         {
            freqHeadPtr = (freq_dist *) calloc (1, sizeof (freq_dist));
            alloc_chk (WHERE, freqHeadPtr);

            freqHeadPtr->next = NULL;
         }
         if (tempPtr1)
         {
            char  dest_buffer1[100];
            char *tempPtr2 = NULL;
            int   length1 = 0, length2 = 0;
            length1 = tempPtr1 - buffer - offset;
            strncpy (dest_buffer1, buffer + offset, length1);
            dest_buffer1[length1] = 0;
            val = atof (dest_buffer1);

            tempPtr2 = strstr (tempPtr1 + seperator_string_len, ITEM_SEPERATOR);
            if (tempPtr2)
            {
               char  dest_buffer2[100];
               freq_dist *FreqPtr = NULL;
               freq_dist *it = NULL;
               freq_dist *it1 = NULL;
               length2 = tempPtr2 - tempPtr1 - seperator_string_len;
               strncpy (dest_buffer2, tempPtr1 + seperator_string_len, length2);
               dest_buffer2[length2] = 0;
               frequency = atof (dest_buffer2);

               FreqPtr = (freq_dist *) calloc (1, sizeof (freq_dist));
               alloc_chk (WHERE, FreqPtr);

               FreqPtr->frequency = frequency;
               FreqPtr->value = val;
               FreqPtr->cum_frequency = 0.0;
               it = freqHeadPtr;
               it1 = freqHeadPtr->next;
               while ((it1 != NULL) && (val > (it1->value)))
               {
                  it = it->next;
                  it1 = it1->next;
               }
               it->next = FreqPtr;
               FreqPtr->next = it1;
            }
            offset += length1 + length2 + (2 * seperator_string_len);
            nNoOfVal++;
         }
         else
         {
            break;
         }
      }
      currentNodePtr = freqHeadPtr->next;

      while (currentNodePtr != NULL)
      {
         cum_total_upto = cum_total_upto + currentNodePtr->frequency;
         currentNodePtr->cum_frequency = (double) cum_total_upto / Total_Iter;
         currentNodePtr = currentNodePtr->next;
      }
      LinePtr->freq_table = freqHeadPtr;

      LinePtr->next = LineInfoPtr;
      LineInfoPtr = LinePtr;
   }
   *linesHeadPtr = LineInfoPtr;
}

void freeDoubleList (nodeDouble ** nodePtr)
{
   if (nodePtr == NULL)
      return;
   else
   {
      nodeDouble *tempPtr = *nodePtr;

      while (tempPtr != NULL)
      {
         nodeDouble *temp = tempPtr->next;
         free (tempPtr);
         tempPtr = temp;
      }
      *nodePtr = NULL;
   }
}

void ReadInputFileData (line_info_struct ** linesListPtr, int TotalIter)
{
   FILE *stream = NULL;

   stream = open_chkq (WHERE, "Histogram.txt", "r");

   readDataFromFile (stream, linesListPtr, TotalIter);
}

void calculate_p_value_new (line_info_struct * linesListPtr, int locus,
                            int stat, int NO_LOCI, int Iter_No, int NO_STAT,
                            int Total_Iter, int TestFlag,
                            FILE * write_fp, int STAT_FLAG, double *QUANT_OPT)
{
   int   Repeat = 0;
   int   i, j, number;
   double and_dist = 0;
   double mean_value = 0;
   double *quant_value = NULL;
   double *quant_comp = NULL;
   double val2 = 0;
   int   len;
   char  buffer[256];
   int   ret_count = 0;
   nodeDouble *meanHeadPtr = NULL;
   nodeDouble *AnddarHeadPtr = NULL;
   nodeDouble **quantvalHeadptr = NULL;

   if (linesListPtr == NULL)
   {
      printf ("Empty Data passed for p_value_calculation. Exiting...\n");
      exit (EXIT_FAILURE);
   }
   if (g_ModifiedArray)
   {
      free (g_ModifiedArray);
      g_ModifiedArray = NULL;
   }

   if (STAT_FLAG == 2)
   {
      if (QUANT_OPT == NULL)
      {
         printf ("There is error in Quantile Calculation\n");
         goto ERROR_LEVEL_1;
      }
      number = QUANT_OPT[0];
      quant_comp = (double *) calloc (number, sizeof (double));
      alloc_chk (WHERE, quant_comp);

      quantvalHeadptr = (nodeDouble **) calloc (number, sizeof (nodeDouble *));
      alloc_chk (WHERE, quantvalHeadptr);

      for (j = 0; j < number; j++)
      {
         quant_comp[j] = QUANT_OPT[j + 1];
      }
   }
   else if (STAT_FLAG == 1)
   {
      Average_Dist_withList (locus, stat, NO_LOCI, Iter_No, NO_STAT,
                             linesListPtr);
   }
   if ((TestFlag == 1) && (Repeat == 0))
   {
      int   ret_count = 0;
      while (ret_count < Iter_No)
      {
         nullmean[ret_count] = 0;
         nullvariance[ret_count] = 0;

         nullmean[ret_count] =
            Mean_withList (linesListPtr,
                           ret_count * NO_LOCI * NO_STAT + stat - 1,
                           Total_Iter, NO_LOCI, NO_STAT, locus);

         nullvariance[ret_count] =
            Var_withList (linesListPtr,
                          ret_count * NO_LOCI * NO_STAT + stat - 1,
                          Total_Iter, NO_LOCI, NO_STAT);

         ret_count = ret_count + 1;
      }
      Repeat = 1;
   }

   if (write_fp != NULL)
   {
      len = sprintf (buffer, "\n");
      fwrite (buffer, len, 1, write_fp);
   }
   for (i = (locus - 1) * NO_STAT + stat - 1;
        i <= ((locus - 1) * NO_STAT + stat - 1) +
        (Iter_No - 1) * NO_LOCI * NO_STAT; i = i + (NO_LOCI * NO_STAT))
   {
      nodeDouble *nodeptr1 = NULL;
      nodeDouble *nodeptr2 = NULL;
      and_dist = 0;
      mean_value = 0;

      if (STAT_FLAG == 0)
      {
         mean_value =
            Mean_withList (linesListPtr, i, Total_Iter, NO_LOCI,
                           NO_STAT, locus);
      }
      if (TestFlag == 1)
      {
         if (write_fp == NULL)
         {
            printf ("Invalid file pointer passed in P-Value calculation (%d)\n",
                    __LINE__);
            goto ERROR_LEVEL_1;
         }
         for (ret_count = 0; ret_count < Iter_No; ret_count++)
         {
            if ((i >= (ret_count * NO_LOCI * NO_STAT))
                && (i < (ret_count * NO_LOCI * NO_STAT + NO_LOCI * NO_STAT)))
            {
               if (nullvariance[ret_count] > 0)
               {
                  mean_value =
                     (mean_value -
                      nullmean[ret_count]) / sqrt (nullvariance[ret_count]);
               }
               len = sprintf (buffer, "% 4.7f", mean_value);
               fwrite (buffer, 1, len, write_fp);
               break;
            }
         }
         if (i == ((locus - 1) * NO_STAT + stat - 1)
             + (Iter_No - 1) * NO_LOCI * NO_STAT)
         {
            len = sprintf (buffer, "\n");
            fwrite (buffer, 1, len, write_fp);
         }
      }
      if (STAT_FLAG == 0)
      {
         nodeptr2 = (nodeDouble *) calloc (1, sizeof (nodeDouble));
         alloc_chk (WHERE, nodeptr2);

         nodeptr2->value = mean_value;
         nodeptr2->next = meanHeadPtr;
         meanHeadPtr = nodeptr2;
      }
      if (STAT_FLAG == 2)
      {
         quant_value = (double *) calloc (number, sizeof (double));
         alloc_chk (WHERE, quant_value);

         for (j = 0; j < number; j++)
         {
            nodeDouble *nodeptr;
            quant_value[j] = 0;
            quant_value[j] = Quantile_withList (linesListPtr, i, quant_comp[j]);
            nodeptr = (nodeDouble *) calloc (1, sizeof (nodeDouble));
            alloc_chk (WHERE, nodeptr);

            nodeptr->value = quant_value[j];
            nodeptr->next = quantvalHeadptr[j];
            quantvalHeadptr[j] = nodeptr;
         }
      }
      else if (STAT_FLAG == 1)
      {
         and_dist = Anderson_Dist_withList (linesListPtr, i);
         nodeptr1 = (nodeDouble *) calloc (1, sizeof (nodeDouble));
         alloc_chk (WHERE, nodeptr1);

         nodeptr1->value = mean_value;
         nodeptr1->next = AnddarHeadPtr;
         AnddarHeadPtr = nodeptr1;
      }
   }
   if (STAT_FLAG == 0)
   {
      if (stat == 1)
         prntIBDTLocPos (locus);

      val2 = p_val_double_withList (meanHeadPtr);
      printf ("  %7.4f", val2);
   }
   else if (STAT_FLAG == 1)
   {
      if (stat == 1)
         prntIBDTLocPos (locus);

      val2 = p_val_double_withList (AnddarHeadPtr);
      printf ("  %7.4f", val2);
   }
   else if (STAT_FLAG == 2)
   {
      if (stat == 1)
         prntIBDTLocPos (locus);

      for (i = 0; i < number; i++)
      {
         double val = 0;

         val = p_val_double_withList (quantvalHeadptr[i]);
         printf (" %7.4f ", val);
      }
   }
   if (STAT_FLAG == 0)
   {
      if (meanHeadPtr != NULL)
      {
         freeDoubleList (&meanHeadPtr);
      }
   }
   if (STAT_FLAG == 1)
   {
      if (AnddarHeadPtr != NULL)
      {
         freeDoubleList (&AnddarHeadPtr);
      }
   }
   if (STAT_FLAG == 2)
   {
      if (quantvalHeadptr != NULL)
      {
         for (i = 0; i < number; i++)
         {
            if (quantvalHeadptr[i] != NULL)
            {
               freeDoubleList (&quantvalHeadptr[i]);
            }
         }
         free (quantvalHeadptr);
         quantvalHeadptr = NULL;
      }
      if (quant_comp != NULL)
      {
         free (quant_comp);
      }
   }
   if (g_ModifiedArray)
   {
      free (g_ModifiedArray);
      g_ModifiedArray = NULL;
   }

   return;

 ERROR_LEVEL_1:
   {
      if (STAT_FLAG == 0)
      {
         if (meanHeadPtr != NULL)
         {
            freeDoubleList (&meanHeadPtr);
         }
      }
      if (STAT_FLAG == 1)
      {
         if (AnddarHeadPtr != NULL)
         {
            freeDoubleList (&AnddarHeadPtr);
         }
      }
      if (STAT_FLAG == 2)
      {
         if (quantvalHeadptr != NULL)
         {
            for (i = 0; i < number; i++)
            {
               if (quantvalHeadptr[i] != NULL)
               {
                  freeDoubleList (&quantvalHeadptr[i]);
               }
            }
            free (quantvalHeadptr);
            quantvalHeadptr = NULL;
         }
         if (quant_comp != NULL)
         {
            free (quant_comp);
         }
      }
      if (g_ModifiedArray)
      {
         free (g_ModifiedArray);
         g_ModifiedArray = NULL;
      }
      if (linesListPtr != NULL)
      {
         freeLineList (&linesListPtr);
      }
      exit (EXIT_FAILURE);
   }
}

int pairwise_comparison_withList (int locus, line_info_struct * HeadPtr,
                                  int a, int b)
{
   double number1, number2;
   int   count = 0, count1 = 0;
   int   j = 1;
   line_info_struct *tempPtr = HeadPtr;
   int   pairwiseResult = 0;

   while (j <= 1000)
   {
      double v = (double) rand () / RAND_MAX;
      double v1 = (double) rand () / RAND_MAX;
      int   bFirst = 0;
      int   bSecond = 0;

      while (tempPtr != NULL)
      {
         if (tempPtr->line_number == a)
         {
            freq_dist *distPtr = NULL;
            bFirst = 1;
            distPtr = tempPtr->freq_table;
            if (distPtr != NULL)
            {
               distPtr = distPtr->next;
               if (distPtr != NULL)
               {
                  if (v < distPtr->cum_frequency)
                  {
                     number1 = distPtr->value;
                  }
                  else
                  {
                     freq_dist *distPtr1 = distPtr->next;
                     while (distPtr1 != NULL)
                     {
                        if ((v > distPtr->cum_frequency)
                            && (v <= distPtr1->cum_frequency))
                        {
                           number1 = distPtr1->value;
                           break;
                        }
                        distPtr = distPtr->next;
                        distPtr1 = distPtr->next;
                     }
                  }
               }
            }
            if ((bFirst) && (bSecond))
            {
               break;
            }
         }
         else if (tempPtr->line_number == b)
         {
            freq_dist *distPtr = NULL;
            bSecond = 1;
            distPtr = tempPtr->freq_table;
            if (distPtr != NULL)
            {
               distPtr = distPtr->next;
               if (distPtr != NULL)
               {
                  if (v1 < distPtr->cum_frequency)
                  {
                     number2 = distPtr->value;
                  }
                  else
                  {
                     freq_dist *distPtr1 = distPtr->next;
                     while (distPtr1 != NULL)
                     {
                        if ((v1 > distPtr->cum_frequency)
                            && (v1 <= distPtr1->cum_frequency))
                        {
                           number2 = distPtr1->value;
                           break;
                        }
                        distPtr = distPtr->next;
                        distPtr1 = distPtr->next;
                     }
                  }
               }
            }
            if ((bFirst) && (bSecond))
            {
               break;
            }
         }
         tempPtr = tempPtr->next;
      }
      if (number1 > number2)
      {
         count = count + 1;
      }
      else if (number1 == number2)
      {
         count1 = count1 + 1;
      }
      j = j + 1;
   }
   pairwiseResult = count + (int) count1 / 2;
   return pairwiseResult;
}

int getDistTableSize (line_info_struct * HeadPtr)
{
   int   count = 0;
   line_info_struct *tempPtr = HeadPtr;
   if (tempPtr != NULL)
   {
      freq_dist *distPtr = NULL;
      distPtr = tempPtr->freq_table;
      if (distPtr)
      {
         distPtr = distPtr->next;
         while (distPtr != NULL)
         {
            count++;
            distPtr = distPtr->next;
         }
      }
   }
   return count;
}

freq_dist *getDistEntry (int pos, line_info_struct * HeadPtr)
{
   int   count = 0;
   line_info_struct *tempPtr = HeadPtr;
   if (tempPtr != NULL)
   {
      freq_dist *distPtr = NULL;
      distPtr = tempPtr->freq_table;
      if (distPtr)
      {
         distPtr = distPtr->next;
         while (distPtr != NULL)
         {
            if (count == pos)
            {
               return distPtr;
            }
            count++;
            distPtr = distPtr->next;
         }
      }
   }
   return NULL;
}

line_info_struct *getLineInfo (int line_no, line_info_struct * HeadPtr)
{
   line_info_struct *tempPtr = HeadPtr;
   while (tempPtr != NULL)
   {
      if (tempPtr->line_number == line_no)
      {
         return tempPtr;
      }
      tempPtr = tempPtr->next;
   }
   return NULL;
}

void Average_Dist_withList (int locus1, int stat, int NO_LOCI, int Iter_No,
                            int NO_STAT, line_info_struct * HeadPtr)
{
   int   locus = locus1;
   int   i, it, j, k;
   double *array;
   int   distinct = 0;
   int   num = 0;
   int   t = 0, a;
   int   tag = 0;
   line_info_struct *linePtr2 = NULL;

   line_info_struct *tempPtr = HeadPtr;
   int   Row = 0;
   int   rowSize = 0;

   Row = (locus - 1) * NO_STAT + stat - 1;

   if (tempPtr == NULL)
   {
      printf ("Sorry ! No line info available\n");
      return;
   }
   linePtr2 = getLineInfo (Row, tempPtr);

   rowSize = getDistTableSize (linePtr2);

   distinct = rowSize;

   for (a = (Row + (NO_LOCI * NO_STAT));
        a <= (Row + (Iter_No - 1) * NO_LOCI * NO_STAT);
        a = a + (NO_LOCI * NO_STAT))
   {
      line_info_struct *linePtr = NULL;
      int   distSize = 0;
      linePtr = getLineInfo (a, tempPtr);
      distSize = getDistTableSize (linePtr);

      for (j = 0; j < distSize; j++)
      {
         num = 0;
         for (k = (Row); k < a; k = k + (NO_LOCI * NO_STAT))
         {
            line_info_struct *linePtr1 = NULL;
            int   distSize1 = 0;
            linePtr1 = getLineInfo (k, tempPtr);
            distSize1 = getDistTableSize (linePtr1);

            for (i = 0; i < distSize1; i++)
            {
               freq_dist *ptr1 = getDistEntry (j, linePtr);
               freq_dist *ptr2 = getDistEntry (i, linePtr1);
               if ((ptr1 == NULL) || (ptr2 == NULL))
               {
                  printf ("Could not be able to find the entry in"
                          " those position at line (%d)\n", __LINE__);
                  return;
               }
               if (ptr1->value == ptr2->value)
               {
                  num = 1;
                  break;
               }
            }
            if (num == 1)
            {
               break;
            }
         }
         if (num == 0)
         {
            distinct = distinct + 1;
         }
      }
   }
   array = (double *) calloc (distinct, sizeof (double));
   alloc_chk (WHERE, array);

   for (i = 0; i < rowSize; i++)
   {
      freq_dist *ptr1 = getDistEntry (i, linePtr2);
      if (ptr1 == NULL)
      {
         printf ("Could not be able to find the entry in those position"
                 " at line (%d)\n", __LINE__);
         return;
      }
      array[i] = ptr1->value;
   }
   t = rowSize;
   g_ModifiedArray = (freq_dist *) calloc (distinct, sizeof (freq_dist));
   alloc_chk (WHERE, g_ModifiedArray);

   for (a = (Row + (NO_LOCI * NO_STAT));
        a <= (Row + (Iter_No - 1) * NO_LOCI * NO_STAT);
        a = a + (NO_LOCI * NO_STAT))
   {
      line_info_struct *linePtr = NULL;
      int   distSize = 0;
      linePtr = getLineInfo (a, tempPtr);
      distSize = getDistTableSize (linePtr);

      for (j = 0; j < distSize; j++)
      {
         freq_dist *ptr1 = getDistEntry (j, linePtr);
         tag = 0;
         for (k = 0; k < t; k++)
         {
            if (ptr1 == NULL)
            {
               printf ("Could not be able to find the entry in those"
                       " position at line (%d)\n", __LINE__);
               return;
            }
            if (ptr1->value == array[k])
            {
               tag = 1;
               break;
            }
         }
         if (tag == 0)
         {
            if (ptr1 == NULL)
            {
               printf ("Could not be able to find the entry in those"
                       " position at line (%d)\n", __LINE__);
               return;
            }
            array[t] = ptr1->value;

            t = t + 1;
         }
      }
   }
   k = 0;

   for (j = 0; j < distinct; j++)
   {
      for (a = ((Row) + (NO_LOCI * NO_STAT));
           a <= ((Row) + (Iter_No - 1) * NO_LOCI * NO_STAT);
           a = a + (NO_LOCI * NO_STAT))
      {
         line_info_struct *linePtr = NULL;
         int   distSize = 0;
         freq_dist *LastPtr = NULL;
         freq_dist *FirstPtr = NULL;
         linePtr = getLineInfo (a, tempPtr);
         FirstPtr = getDistEntry (0, linePtr);

         distSize = getDistTableSize (linePtr);

         LastPtr = getDistEntry (distSize - 1, linePtr);

         if ((FirstPtr == NULL) || (LastPtr == NULL))
         {
            printf ("line no (%d), size(%d),Row(%d)\n", linePtr->line_number,
                    distSize, a);
            printf ("Could not be able to find the entry in those"
                    " position at line (%d)\n", __LINE__);
            return;
         }

         for (it = 0; it < (distSize - 1); it++)
         {
            freq_dist *ptr1 = getDistEntry (it, linePtr);
            freq_dist *ptr2 = getDistEntry (it + 1, linePtr);
            if ((ptr1 == NULL) || (ptr2 == NULL))
            {
               printf ("Could not be able to find the entry in those"
                       " position at line (%d)\n", __LINE__);
               return;
            }
            if (ptr1->value == array[j])
            {
               g_ModifiedArray[k].value = array[j];
               g_ModifiedArray[k].frequency =
                  g_ModifiedArray[k].frequency + ptr1->frequency;
               g_ModifiedArray[k].cum_frequency =
                  g_ModifiedArray[k].cum_frequency + ptr1->cum_frequency;
               break;
            }
            else if ((ptr1->value < array[j]) && (ptr2->value > array[j]))
            {
               g_ModifiedArray[k].value = array[j];
               g_ModifiedArray[k].cum_frequency =
                  g_ModifiedArray[k].cum_frequency +
                  ((array[j] - ptr1->value) * ptr2->cum_frequency +
                   (ptr2->value -
                    array[j]) * ptr1->cum_frequency) / (ptr2->value -
                                                        ptr1->value);
               g_ModifiedArray[k].frequency = g_ModifiedArray[k].frequency + 0;
            }
         }
         if (FirstPtr->value > array[j])
         {
            g_ModifiedArray[k].value = array[j];
            g_ModifiedArray[k].frequency = g_ModifiedArray[k].frequency + 0;
            g_ModifiedArray[k].cum_frequency =
               g_ModifiedArray[k].cum_frequency + 0;
         }
         else if (LastPtr->value <= array[j])
         {
            g_ModifiedArray[k].value = array[j];
            g_ModifiedArray[k].frequency = g_ModifiedArray[k].frequency + 0;
            g_ModifiedArray[k].cum_frequency =
               g_ModifiedArray[k].cum_frequency + 1;
         }
      }
      g_ModifiedArray[k].cum_frequency =
         (double) (g_ModifiedArray[k].cum_frequency / Iter_No);
      if (g_ModifiedArray[k].cum_frequency > 1)
      {
         printf ("Error");
      }
      g_ModifiedArray[k].frequency =
         (double) (g_ModifiedArray[k].frequency / Iter_No);
      k = k + 1;
   }

   g_NoOfDistinctElement = distinct;
   if (array)
      free (array);
}

double Anderson_Dist_withList (line_info_struct * HeadPtr, int number)
{
   int   i;
   double distnew = 0;
   double cal = 0, cal1 = 0, cal2 = 0;
   line_info_struct *tempPtr = NULL;
   tempPtr = HeadPtr;

   while (tempPtr != NULL)
   {
      if (tempPtr->line_number == number)
      {
         for (i = 0; i < g_NoOfDistinctElement; i++)
         {
            int   bFound = 0;
            cal = 0;
            cal1 = 0;
            cal2 = 0;
            if ((g_ModifiedArray[i].cum_frequency != 0)
                && (g_ModifiedArray[i].cum_frequency != 1))
            {
               freq_dist *distPtr = NULL;
               double dFirstEntryVal = 0;
               double dLastEntryVal = 0;
               distPtr = tempPtr->freq_table;
               if (distPtr != NULL)
               {
                  freq_dist *distPtr1 = NULL;
                  distPtr = distPtr->next;
                  if (distPtr != NULL)
                  {
                     dFirstEntryVal = distPtr->value;
                     distPtr1 = distPtr->next;
                  }
                  while (distPtr1 != NULL)
                  {
                     dLastEntryVal = distPtr1->value;
                     if ((distPtr->value < g_ModifiedArray[i].value)
                         && (distPtr1->value > g_ModifiedArray[i].value))
                     {
                        bFound = 1;
                        cal = ((g_ModifiedArray[i].value
                                - distPtr->value) * distPtr1->cum_frequency
                               + (distPtr1->value - g_ModifiedArray[i].value)
                               * distPtr->cum_frequency)
                           / (distPtr1->value - distPtr->value);
                        distnew = distnew +
                           ((cal - g_ModifiedArray[i].cum_frequency)
                            * (cal - g_ModifiedArray[i].cum_frequency))
                           / (g_ModifiedArray[i].cum_frequency *
                              (1 - g_ModifiedArray[i].cum_frequency));
                        break;
                     }
                     else if (distPtr->value == g_ModifiedArray[i].value)
                     {
                        bFound = 1;
                        cal = distPtr->cum_frequency;
                        distnew = distnew +
                           ((cal - g_ModifiedArray[i].cum_frequency)
                            * (cal - g_ModifiedArray[i].cum_frequency))
                           / (g_ModifiedArray[i].cum_frequency
                              * (1 - g_ModifiedArray[i].cum_frequency));
                        break;
                     }
                     distPtr = distPtr->next;
                     distPtr1 = distPtr->next;
                  }
                  if (bFound == 0)
                  {
                     if (dFirstEntryVal >= g_ModifiedArray[i].value)
                     {
                        cal1 = 0;
                        distnew = distnew
                           + ((cal1 - g_ModifiedArray[i].cum_frequency)
                              * (cal1 - g_ModifiedArray[i].cum_frequency))
                           / (g_ModifiedArray[i].cum_frequency
                              * (1 - g_ModifiedArray[i].cum_frequency));
                     }
                     else if (dLastEntryVal <= g_ModifiedArray[i].value)
                     {
                        cal2 = 1;
                        distnew = distnew +
                           ((cal2 - g_ModifiedArray[i].cum_frequency)
                            * (cal2 - g_ModifiedArray[i].cum_frequency))
                           / (g_ModifiedArray[i].cum_frequency
                              * (1 - g_ModifiedArray[i].cum_frequency));
                     }
                  }
               }
            }
         }
         break;
      }
      tempPtr = tempPtr->next;
   }
   return distnew;
}

double p_val_double_withList (nodeDouble * headPtr)
{
   int   count = 0;
   double p_val = 0;
   double value = 0;
   int   nSize = 0;
   nodeDouble *tempPtr = NULL;
   nodeDouble *tempPtr2 = NULL;
   tempPtr = headPtr;
   tempPtr2 = headPtr;
   while (tempPtr2 != NULL)
   {
      value = tempPtr2->value;
      tempPtr2 = tempPtr2->next;
   }
   while (tempPtr != NULL)
   {
      if (tempPtr->value >= value)
      {
         count = count + 1;
      }
      tempPtr = tempPtr->next;
      nSize++;
   }
   if (nSize > 0)
      p_val = (double) count / nSize;
   return p_val;
}

double p_val_int_withList (nodeInt * headPtr)
{
   int   count = 0;
   double p_val = 0;
   int   value = 0;
   int   nSize = 0;
   int   bFrist = 1;
   nodeInt *tempPtr = NULL;
   tempPtr = headPtr;

   while (tempPtr != NULL)
   {
      if (bFrist)
      {
         value = tempPtr->value;
         bFrist = 0;
      }
      if (tempPtr->value >= value)
      {
         count = count + 1;
      }
      tempPtr = tempPtr->next;
      nSize++;
   }
   if (nSize > 0)
      p_val = (double) count / nSize;
   return p_val;
}

double Quantile_withList (line_info_struct * HeadPtr, int Row, double Prop)
{
   double quantile = 0;
   line_info_struct *tempPtr = HeadPtr;
   while (tempPtr != NULL)
   {
      if (tempPtr->line_number == Row)
      {
         freq_dist *distPtr = NULL;
         distPtr = tempPtr->freq_table;
         if (distPtr)
         {
            distPtr = distPtr->next;
            if (distPtr != NULL)
            {
               if (distPtr->cum_frequency > Prop)
               {
                  quantile = distPtr->value;
                  return quantile;
               }
               else
               {
                  freq_dist *distPtr1 = distPtr->next;
                  while (distPtr1 != NULL)
                  {
                     if ((distPtr->cum_frequency <= Prop)
                         && (distPtr1->cum_frequency > Prop))
                     {
                        quantile = distPtr1->value;
                        return quantile;
                     }
                     distPtr = distPtr->next;
                     distPtr1 = distPtr->next;
                  }
                  quantile = distPtr->value;
                  return quantile;
               }
            }
         }
         break;
      }
      tempPtr = tempPtr->next;
   }
   return quantile;
}

double Mean_withList (line_info_struct * HeadPtr, int Row, int Total_Iter,
                      int NoOfLocus, int NoOfStat, int locus)
{
   double mean = 0;
   line_info_struct *tempPtr = HeadPtr;

   while (tempPtr != NULL)
   {
      if (tempPtr->line_number == Row)
      {
         freq_dist *distPtr = NULL;
         /* printLineInfo(tempPtr); */
         distPtr = tempPtr->freq_table;
         if (distPtr)
         {
            distPtr = distPtr->next;
            while (distPtr != NULL)
            {
               mean = mean + (double) (distPtr->frequency * distPtr->value);
               distPtr = distPtr->next;
            }
         }
         break;
      }
      tempPtr = tempPtr->next;
   }

   mean = mean / Total_Iter;
#if 0
   if ((Mean_Tag == 1) && ((Row >= 0) && (Row < (NoOfLocus * NoOfStat))))
   {
      mean_array[locus - 1] = mean;
   }
#endif
   return mean;
}

double Var_withList (line_info_struct * HeadPtr, int Row, int Total_Iter,
                     int NoOfLocus, int NoOfStat)
{
   double var = 0;
   double mean = 0;
   line_info_struct *tempPtr = HeadPtr;
   while (tempPtr != NULL)
   {
      if (tempPtr->line_number == Row)
      {
         freq_dist *distPtr = NULL;
         distPtr = tempPtr->freq_table;
         if (distPtr)
         {
            distPtr = distPtr->next;
            while (distPtr != NULL)
            {
               mean = mean + (double) (distPtr->frequency * distPtr->value);
               var =
                  var +
                  (double) ((distPtr->frequency * distPtr->value *
                             distPtr->value));
               distPtr = distPtr->next;
            }
         }
         break;
      }
      tempPtr = tempPtr->next;
   }
   var = var / Total_Iter - ((mean / Total_Iter) * (mean / Total_Iter));
   return var;
}
