/* This file contains doxygen comments. */
/** @file */

/* MORGAN: Copyright 1991-2009: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*****************************************************************************
 program name: civil.c  (running as a version of lm_markers).
 ******************************************************************************/

/**
 * @defgroup lm_auto lm_auto
 * @ingroup autozyg
 * @{
 */

#define PNAM "lm_markers"
#define MARKERS
#define stdmsg stdout

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "lodscore.h"
#include "rans.h"

/* #include "autozyg.h" */
/* #include "auto_chk.c" */
/* #define AUTO_CKPNTS 1000 */

#include "Utils/civil_datastruct.c"
#include "Utils/civil_utils.c"
#include "Utils/civil_readpar_MORGAN.c"
#include "Utils/civil_vc.c"
#include "Utils/civil_w.c"

static int nsw;

int main (int argc, char **argv)
{
   /* File names */
   file_names_t fns;

   /* Test parameters */
   data_pars_t data_pars;
   trait_model_t trt_mod;
   miv_pars_t miv_pars;
   civ_pars_t civ_pars;

   /* Pedigree structures */
   ped_list_t ped_list;
   adg_list_t adg_list;
   iv_list_t iv_list, iv_unlinked;

   FILE *fp;
   int   k;

   /* Call setupAutozyg to:
    *   1. process the MORGAN format parameter file,
    *   2. setup structures for pedigrees, meioses, markers and trait values.
    */
   nsw = setupAutozyg (argc, argv, PNAM, 0);

   /*
    *  list_switches(nsw);
    *  list_nghds();
    *  printf("\n done lists ");
    *  fflush(stdout);
    */

   /* open the xtra data file */
   if (!(fp = open_fil (XTRA)))
      give_up (ERR_qit);

   /* Parsing the XTRA parameter file */
   fprintf (stdmsg,
            "Parsing the XTRA parameter file and populating local data structures ...\n");

   /* Directory and file names */
   read_file_names (fp, &fns);

   /* Info on the oscor file (inheritance vectors) */
   read_data_pars (fp, &data_pars);

   /* If the oscor file is by component, convert it to by MCMC run. */
   if (data_pars.by_component)
   {
      fprintf (stdmsg,
               "The meiosis indicators are by component, rearrange the rows so the order is by MCMC run.\n");
      rearrange_iv_file (&data_pars, &fns);
   }

   /* Populate pedigree data structures using info from MORGAN global variables */
   read_peds_MORGAN (fns.name_ped, &ped_list);

   /*
      for (k = 0; k < ped_list.n_peds; k++) {
      print_ped(&ped_list.peds[k]);
      }
    */

   /* Read trait model parameters */
   /* Assign default mean trait and sd trait value to be sample mean and sample sd.
    * Mean trait value will be overwritten if user provides other value.
    * trt_mod.sd will be used to estimate residual se if user prefers
    */
   trt_mod.sd = ped_list.sd;
   trt_mod.mu = ped_list.mu;
   read_trait_model_pars (fp, &trt_mod);

   /* Marginal test parameters */
   read_miv_pars (fp, &miv_pars, &data_pars);
   sprintf (fns.file_miv_t, "%s/%s.miv.%s.t.out", fns.dir_out, fns.name_ped,
            miv_pars.test);
   sprintf (fns.file_miv_p, "%s/%s.miv.%s.p.out", fns.dir_out, fns.name_ped,
            miv_pars.test);

   /* Conditional test parameters */
   read_civ_pars (fp, &civ_pars, &data_pars);
   sprintf (fns.file_civ_t, "%s/%s.civ.%s.t.out", fns.dir_out, fns.name_ped,
            civ_pars.test);
   sprintf (fns.file_civ_p, "%s/%s.civ.%s.p.out", fns.dir_out, fns.name_ped,
            civ_pars.test);

   /* close the XTRA par file */
   fclose (fp);

   /**
   * Record the set of all possible test and conditioning positions (from MORGAN).
   */
   record_all_test_pos (&data_pars, &miv_pars, &civ_pars);
   /* fclose(fp); */

   /* Verify that we have read things in correctly. */
   print_pars (fns, data_pars, trt_mod, miv_pars, civ_pars);

   /* exit(1); */

   /* Construct allele descent graphs (adgs) for pedigrees */
   fprintf (stdmsg, "Constructing allele descent graphs ...\n");

   adg_list.n_peds = ped_list.n_peds;
   adg_list.mu = ped_list.mu;
   adg_list.sd = ped_list.sd;
   adg_list.peds = (adg_t *) malloc (sizeof (adg_t) * adg_list.n_peds);

   for (k = 0; k < ped_list.n_peds; k++)
   {
      /*
         if (PrintLevel == 2)
         fprintf(stdmsg, "Constructing FGL-graph for pedigree %d ...\n", k);
       */
      query_ped_info (&ped_list.peds[k], &adg_list.peds[k]);
      /* if (PrintLevel > 2) */
      print_adg (&adg_list.peds[k]);
   }
   free_ped_list (&ped_list);

   if (strcmp (civ_pars.test, "vc") == 0 || strcmp (miv_pars.test, "vc") == 0)
   {
      fprintf (stdmsg, "Computing vc tables ...\n");
      /* adg_list.n_peds=60; // For debugging */
      compute_vc_tables (&adg_list, &trt_mod, fns.dir_out);
      fprintf (stdmsg, "Done computing vc tables ...\n\n");
   }

   if (strcmp (civ_pars.test, "w") == 0 || strcmp (miv_pars.test, "w") == 0)
   {
      fprintf (stdmsg, "Computing log w-score tables ...\n");
      /* adg_list.n_peds=60; // For debugging */
      compute_lgw_tables (&adg_list, &trt_mod, fns.dir_out);
      fprintf (stdmsg, "Done computing log w-score tables. \n\n");
   }

   /* exit(1); */

   fprintf (stdmsg, "Reading inheritance vectors ...\n");
   if (data_pars.by_component)
   {
      /*   read_ivs(file_ivs, &dat_pars, &adg_list, &iv_list_new, &iv_unlinked_new); */
      read_ivs (fns.file_mis_rearranged, &data_pars, &adg_list, &iv_list,
                &iv_unlinked);
   }
   else
   {
      read_ivs (fns.file_mis, &data_pars, &adg_list, &iv_list, &iv_unlinked);
   }

   /*     print_iv_list2(&iv_list, &adg_list); */
   /*     print_iv_list2(&iv_unlinked, &adg_list); */
   /*     print_iv_list_3(&iv_list); */

   /* exit(1); */

   fprintf (stdmsg, "Performing marginal tests ...\n");
   if (strcmp (miv_pars.test, "vc") == 0)
   {
      miv_vc (&adg_list, &iv_list, &miv_pars, fns.dir_out);
   }
   else if (strcmp (miv_pars.test, "w") == 0)
   {
      miv_w (&adg_list, &iv_list, &trt_mod, &miv_pars, fns.dir_out);
   }

   print_miv_pars (&miv_pars);
   printf_miv_pars (&miv_pars, fns.file_miv_t, fns.file_miv_p);
   free_miv_pars (&miv_pars);
   fprintf (stdmsg, "Done marginal tests.\n");
   fprintf (stdmsg, "\n");

   /*     miv_pars.n_mc= 9999; */
   /*     miv_pars.n_pos = iv_unlinked.n_loci; */
   /*     miv_pars.pos = (int *) malloc(sizeof(int) * miv_pars.n_pos); */
   /*     for (k=0; k<miv_pars.n_pos; k++) { */
   /*         miv_pars.pos[k] = k; */
   /*     } */
   /*     miv_w(&adg_list, &iv_unlinked, &trt_mod, &miv_pars, dir); */
   /*     print_miv_pars(&miv_pars); */
   /*     free_miv_pars(&miv_pars); */

   fprintf (stdmsg, "Performing conditional tests ...\n");
   if (strcmp (civ_pars.test, "vc") == 0)
   {
      civ_vc (&adg_list, &iv_list, &civ_pars, fns.dir_out);
   }
   else if (strcmp (civ_pars.test, "w") == 0)
   {
      civ_w (&adg_list, &iv_list, &trt_mod, &civ_pars, fns.dir_out);
   }
   print_civ_pars (&civ_pars);
   printf_civ_pars (&civ_pars, fns.file_civ_t, fns.file_civ_p);
   free_civ_pars (&civ_pars);
   /*    free_mi_mat(&s_mat); */
   /*    free_mi_mat(&s_unlinked); */
   fprintf (stdmsg, "Done conditional tests. \n");

   /* Clean up */

   free_adg_list (&adg_list);
   free_iv_list (&iv_list);
   free_iv_list (&iv_unlinked);
   /*
    *    free_iv_list(&iv_list_new);
    *    free_iv_list(&iv_unlinked_new);
    *
    */

   free_trait_model (&trt_mod);
   free_data_pars (&data_pars);

   /* TODO
    * Free vc tables?
    */
   return 0;
}
