/*******************************************************************************
  program name: lm_map (see Autozyg/README_userdoc) for overview.
********************************************************************************/
/**
 * @defgroup lm_auto lm_auto
 * @ingroup autozyg
 * @{
 */

#define PNAM  "lm_map"

#include "autozyg.h"
#include "./Utils/map.h"
#include "./Utils/setup_peel.h"
#include "./Utils/locus_peel.h"

#include "./Utils/small_map_fncts.c"
#include "./Utils/large_map_fncts.c"
#include "./Utils/setup_peel.c"
#include "./Utils/locus_peel.c"

int main (int argc, char *argv[])
{
   int   i, j, jj, k, nt, t, ta, c, pgmidy = 0;
   int ***recounts;                     /* [nc][nl][6] counts of recombs */
   double reject_rate = 0, total = 0;
   double **tot_log_prob_seg;
   double **tot_log_prob_pen;
   double **temp_log_prob_seg;
   double **prob;
   FILE *oSeedF;
#include "./Utils/variable.defs"

#if 0
   ChkIt = 2;                           /* parfile statement: check progress
                                           MC iteration K (not implemented
                                           for lm_map)                      */
#endif

   NSw = setupAutozyg (argc, argv, PNAM, 0);
   setupSampling ();

   /* NMarks is the cardinal number of markers for which phenotypes or
      genotypes are provided                                                  */

   nLocy = 2 * NMarks;

   alloc_loglik (&recounts, &tot_log_prob_seg, &temp_log_prob_seg,
                 &tot_log_prob_pen);

#include "./Utils/alloc.mem"            /* use vars from par file to alloc memory */

   alloc_logprob ();
   t = 0;

   if (!MapEstMod)                /*============= ESTIMATION W/0 =============*/
   {
      prob = setupInherit (1);
      if (MapTyp == SEX_AVG || MapTyp == BOTH)   /*--------- SXA-W/O ---------*/
      {
         for (r = 0; r < RedoEst; r++)
         {
            get_init (0, simple_null, tmp_map, &epsi, alpha);
            reset_map (1, NSw, tmp_map, Marek, prob);
            pr_section (1, r);
            pr_map (0, Marek);
            printf ("\n");
            size_flag = 0;
            for (i = 1; i <= 55; i++)
               printf ("=");
            printf ("\n Run %d: Generating the MCEM "
                    "sex-averaged estimate...\n", r + 1);
            for (i = 1; i <= 55; i++)
               printf ("=");
            printf ("\n");

            emtime = time (NULL);
            n_scans = mcemits (0, size_flag, NSw, &conv, recounts,
                               temp_log_prob_seg, mle_sxa, prob);
            cpy_map (mle_sxa, tmp_map);
            pr_map (0, mle_sxa);
            if (conv)
               printf ("\n MCEM convergence criterion met. \n");
            else
               printf ("\n MCEM terminated by user defined limit. \n");
            emtime = time (NULL) - emtime;
            if (!emtime)
               emtime = 1;

            reset_globs (0, n_scans, &Glob_Max, &Obs_Info_Avg, &Obs_Info_Spc);

            abort = 0;
            if (MaximAlgo == HYBRID)
            {
               printf ("\n");
               for (i = 1; i <= 55; i++)
                  printf ("=");
               printf ("\n Run %d: Generating the SA "
                       "sex-averaged estimate... \n", r + 1);
               for (i = 1; i <= 55; i++)
                  printf ("=");
               printf ("\n");

               if (!r)
               {
                  cpy_map (mle_sxa, best);
                  fish_inv_flag = 1;
               }
               else
                  fish_inv_flag = 1 - comp_map (EM_bound, mle_sxa, best);

               satime = time (NULL);
               if (fish_inv_flag)       /*   EST. INV OBS INFO: SXA W/O       */
               {
                  fish_inv_flag = 0;
                  obs_info (0, obs_info_mat, scount, recounts, NSw,
                            prob, temp_log_prob_seg, mle_sxa, NHybCurvIt, r);
                  nnr = NMarks - 1;
                  d_invert (obs_info_mat, ymat, nnr);
                  avgeff[0] = 0;
                  j = 0;
                  for (i = 1; i <= nnr; i++)
                  {
                     if (ymat[i][i] > 0)
                     {
                        avgeff[0] +=
                           (mle_sxa[0][i] * (1 - mle_sxa[0][i]) / ymat[i][i]);
                        j++;
                     }
                  }
                  avgeff[0] = avgeff[0] / (double) j;
               }

               v = 0;                   /* PROPOSE INIT. VALUE FOR SA: MLE_SXA W/O   */
               size_flag = 0;
               llr_flag = 0;
               while (!size_flag)       /*      while not close         */
               {
                  v++;
                  sait (0, v, recounts, NSw, prob, temp_log_prob_seg,
                        ymat, mle_sxa);
                  size_flag = comp_map (SA_bound, mle_sxa, Marek);

                  if (llr_flag < 5)     /* SEARCH FOR SCALING FACTOR */
                  {
                     cpy_map (Marek, se);
                     dn = mle_sxa;
                     accum_lr_flag = 1;
                     llr = 0;
                     while (accum_lr_flag)
                     {
                        accum_lr_flag = 0;
                        llr += logLR (se, dn, NSw, scount, recounts,
                                      prob, temp_log_prob_seg,
                                      &accum_lr_flag, NHybAscentIt);
                     }
                     cpy_map (Marek, mle_sxa); /* SA is not initiated until   */
                     /* the scaling factor is found */
                     if (llr > 0)
                        llr_flag++;
                     else
                        v = v + (int) avgeff[0];

                     if ((time (NULL) - satime) > (TIME * emtime))
                     {
                        printf ("\n SA aborted! Scaling factor NOT found. \n");
                        llr_flag = 5;
                        size_flag = 1;
                        abort = 1;
                        reset_map (1, NSw, tmp_map, mle_sxa, prob);
                     }
                  }
                  else                  /* SCALING FACTOR FOUND; INITIATE SA */
                  {
                     reset_map (0, NSw, mle_sxa, Marek, prob);
                  }
                  if ((time (NULL) - satime) > (TIME * emtime)
                      && size_flag == 0)
                  {
                     printf ("\n SA convergence criterion NOT met. \n");
                     size_flag = 1;
                  }
               }
            }

            if (!abort)
            {                           /* COMPARE. CANDIDATE MLE to CURR.MLE: SXA W/O */
               cpy_map (mle_sxa, se);
               dn = best;
               accum_lr_flag = 1;
               llr = 0;
               if (MaximAlgo == HYBRID)
               {
                  pr_map (0, se);
                  printf ("\n SA convergence criterion met. \n\n");
               }
               if (MaximAlgo == HYBRID || r)
               {
                  for (i = 1; i <= 55; i++)
                     printf ("=");
                  printf ("\n Run %d: Comparing current and "
                          "previous ML estimates...\n", r + 1);
                  for (i = 1; i <= 55; i++)
                     printf ("=");
                  printf ("\n");

                  while (accum_lr_flag)
                  {
                     accum_lr_flag = 0;
                     llr += logLR (se, dn, NSw, scount, recounts,
                                   prob, temp_log_prob_seg, &accum_lr_flag,
                                   Glob_Max);
                  }
                  if (llr < 0)
                     for (i = 0; i < 2; i++)
                        for (j = 1; j < NMarks; j++)
                           best[i][j] = mle_sxa[i][j];
                  else
                     reset_map (1, NSw, best, mle_sxa, prob);
                  printf ("\n Run %d: Best sex-averaged map so far... \n",
                          r + 1);
                  pr_map (0, best);
               }
            }
         }
         printf ("\n");
         for (i = 1; i <= 55; i++)
            printf ("=");
         printf ("\n LM_MAP output for sex-averaged map estimation: \n");
         for (i = 1; i <= 55; i++)
            printf ("=");

         printf ("\n\n Estimating the variance of the sex-averaged mle...");
         obs_info (0, obs_info_mat, scount, recounts, NSw,
                   prob, temp_log_prob_seg, mle_sxa, Obs_Info_Avg, r);

         nnr = NMarks - 1;
         d_invert (obs_info_mat, ymat, nnr); /* ymat <-- invs. obs. infor. */
#if 0
         printf ("\n\n");
         for (c = 0; c < n_compnts; c++)
            pr_reco_counts (0, NMarks, recounts[c]);
#endif
         printf ("\n\n MAXIMUM LIKELIHOOD ESTIMATES\n");
         printf ("\n  Interval        Sex-Averaged (RF)");
         printf ("\n  --------        ----------------\n");
         for (j = 1; j < NMarks; j++)
            printf ("    %d         %15.4f \n", j, mle_sxa[0][j]);
         printf ("\n ESTIMATED VARIANCE OF SEX-AVERAGED MAP \n ");
         for (inr = 1; inr <= nnr; inr++)
         {
            printf ("\n  ");
            for (jnr = 1; jnr <= nnr; jnr++)
            {
               if (inr == jnr && ymat[inr][jnr] <= 0)
                  ymat[inr][jnr] =
                     mle_sxa[0][inr] * (1 - mle_sxa[0][inr]) / avgeff[0];
               if (ymat[inr][jnr] <= 0)
                  printf (" %2.6f ", ymat[inr][jnr]);
               else
                  printf ("  %2.6f ", ymat[inr][jnr]);
            }
         }
         printf ("\n\n EFFECTIVE NUMBER OF MEIOSES\n\n");
         for (i = 1; i < NMarks; i++)
            printf ("     Interval %d: %1.0f \n", i,
                    ceil (mle_sxa[0][i] * (1 - mle_sxa[0][i]) / ymat[i][i]));
         printf ("\n");
      }

      if (MapTyp == SEX_SPEC || MapTyp == BOTH)   /*------- SXS-W/O ----------*/
      {
         for (r = 0; r < RedoEst; r++)
         {
            get_init (1, simple_null, tmp_map, &epsi, alpha);
            reset_map (1, NSw, tmp_map, Marek, prob);
            pr_section (2, r);
            pr_map (1, Marek);
            printf ("\n");
            size_flag = 0;
            for (i = 1; i <= 55; i++)
               printf ("=");
            printf ("\n Run %d: Generating the MCEM "
                    "sex-specific estimate... \n", r + 1);
            for (i = 1; i <= 55; i++)
               printf ("=");
            printf ("\n");

            emtime = time (NULL);
            n_scans = mcemits (1, size_flag, NSw, &conv, recounts,
                               temp_log_prob_seg, mle_sxs, prob);
            cpy_map (mle_sxs, tmp_map);
            pr_map (1, mle_sxs);
            if (conv)
               printf ("\n MCEM convergence criterion met. \n");
            else
               printf ("\n MCEM terminated by user defined limit. \n");
            emtime = time (NULL) - emtime;
            if (!emtime)
               emtime = 1;

            reset_globs (1, n_scans, &Glob_Max, &Obs_Info_Avg, &Obs_Info_Spc);

            abort = 0;
            if (MaximAlgo == HYBRID)
            {
               printf ("\n");
               for (i = 1; i <= 55; i++)
                  printf ("=");
               printf ("\n Run %d: Generating the SA "
                       "sex-specific estimate... \n", r + 1);
               for (i = 1; i <= 55; i++)
                  printf ("=");
               printf ("\n");

               if (!r)                  /*  BEGIN CRUDE ESTIMATE OF INV OBS INFO SXS-W/O */
               {
                  cpy_map (mle_sxs, best);
                  fish_inv_flag = 1;
               }
               else
                  fish_inv_flag = 1 - comp_map (EM_bound, mle_sxs, best);
               satime = time (NULL);
               if (fish_inv_flag)       /*   EST. INV OBS INFO: SXS-W/O       */
               {
                  obs_info (1, obs_info_mat, scount, recounts, NSw,
                            prob, temp_log_prob_seg, mle_sxs, NHybCurvIt, r);
                  nnr = 2 * NMarks - 2;
                  d_invert (obs_info_mat, ymat, nnr);
                  avgeff[0] = avgeff[1] = j = v = 0;
                  for (i = 1; i <= nnr; i++)
                  {
                     if (ymat[i][i] > 0 && (i % 2))
                     {
                        avgeff[0] += mle_sxs[0][((i + 1) / 2)] *
                           (1 - mle_sxs[0][((i + 1) / 2)]) / ymat[i][i];
                        j++;
                     }
                     if (ymat[i][i] > 0 && !(i % 2))
                     {
                        avgeff[1] += (mle_sxs[1][i / 2] *
                                      (1 - mle_sxs[1][i / 2]) / ymat[i][i]);
                        v++;
                     }
                  }
                  avgeff[0] = avgeff[0] / (double) j;
                  avgeff[1] = avgeff[1] / (double) v;
               }
               v = llr_flag = 0;        /* PROP. INIT. VALUE FOR SA */
               size_flag = 0;
               while (!size_flag)
               {
                  v++;
                  sait (1, v, recounts, NSw, prob, temp_log_prob_seg,
                        ymat, mle_sxs);
                  size_flag = comp_map (SA_bound, mle_sxs, Marek);

                  if (llr_flag < 10)    /* SEARCH FOR SCALING FACTOR */
                  {
                     cpy_map (Marek, se);
                     dn = mle_sxs;
                     accum_lr_flag = 1;
                     llr = 0;
                     while (accum_lr_flag)
                     {
                        accum_lr_flag = 0;
                        llr += logLR (se, dn, NSw, scount, recounts,
                                      prob, temp_log_prob_seg,
                                      &accum_lr_flag, NHybAscentIt);
                     }
                     cpy_map (Marek, mle_sxs); /*  SA is not initiated until  */
                     /* the scaling factor is found */
                     if (llr > 0)
                        llr_flag++;
                     else
                     {
                        i = (int) avgeff[0];
                        if (avgeff[1] > i)
                           i = (int) avgeff[1];
                        v = v + (int) i;
                     }
                     if ((time (NULL) - satime) > (TIME * emtime))
                     {
                        printf ("\n SA aborted! Scaling factor NOT found. \n");
                        llr_flag = 10;
                        size_flag = 1;
                        abort = 1;
                     }
                  }
                  else                  /* SCALING FACTOR FOUND; INITIATE SA */
                  {
                     reset_map (0, NSw, mle_sxs, Marek, prob);
                  }

                  if ((time (NULL) - satime) > (TIME * emtime)
                      && size_flag == 0)
                  {
                     printf ("\n SA convergence criterion NOT met. \n");
                     size_flag = 1;
                  }
               }
            }

            if (!abort)                 /* COMPARE CANDIDATE MLE to CURR.MLE: SXS-W/O */
            {
               cpy_map (mle_sxs, se);
               dn = best;
               accum_lr_flag = 1;
               llr = 0;
               if (MaximAlgo == HYBRID)
               {
                  pr_map (1, se);
                  printf ("\n SA convergence criterion met. \n\n");
               }
               if (MaximAlgo == HYBRID || r)
               {
                  for (i = 1; i <= 55; i++)
                     printf ("=");
                  printf ("\n Run %d: Comparing current and "
                          "previous ML estimates...\n", r + 1);
                  for (i = 1; i <= 55; i++)
                     printf ("=");
                  printf ("\n");

                  while (accum_lr_flag)
                  {
                     accum_lr_flag = 0;
                     llr += logLR (se, dn, NSw, scount, recounts,
                                   prob, temp_log_prob_seg, &accum_lr_flag,
                                   Glob_Max);
                  }
                  if (llr < 0 || r == 0)
                     for (i = 0; i < 2; i++)
                        for (j = 1; j < NMarks; j++)
                           best[i][j] = mle_sxs[i][j];
                  else
                     reset_map (1, NSw, best, mle_sxs, prob);
                  printf ("\n Run %d: Best sex-specific map so far... \n",
                          r + 1);
                  pr_map (1, best);
               }
            }
         }
         printf ("\n");
         for (i = 1; i <= 55; i++)
            printf ("=");
         printf ("\n");
         printf (" LM_MAP output for sex-specific map estimation: \n");
         for (i = 1; i <= 55; i++)
            printf ("=");

         printf ("\n\n Estimating the variance of the sex-specific mle...");
         obs_info (1, obs_info_mat, scount, recounts, NSw,
                   prob, temp_log_prob_seg, mle_sxs, Obs_Info_Spc, r);

         nnr = 2 * NMarks - 2;
         d_invert (obs_info_mat, ymat, nnr); /* ymat <-- invs. obs. infor. */
#if 0
         printf ("\n\n");
         for (c = 0; c < n_compnts; c++)
            pr_reco_counts (0, NMarks, recounts[c]);
#endif
         printf ("\n\n MAXIMUM LIKELIHOOD ESTIMATES\n");
         printf ("\n  Interval    Female (RF)         Male (RF)");
         printf ("\n  --------    -----------         ---------\n");
         for (j = 1; j < NMarks; j++)
            printf ("     %d     %12.4f   %15.4f \n", j, mle_sxs[0][j],
                    mle_sxs[1][j]);
         printf ("\n");
         printf (" ESTIMATED VARIANCE OF SEX-SPECIFIC MAP ");
         printf ("[F1,M1,F2,M2,... x F1,M1,F2,M2,...]\n");
         for (inr = 1; inr <= nnr; inr++)
         {
            printf ("\n  ");
            for (jnr = 1; jnr <= nnr; jnr++)
            {
               if ((inr % 2) && inr == jnr && ymat[inr][jnr] <= 0)
                  ymat[inr][jnr] = mle_sxs[0][(inr + 1) / 2] *
                     (1 - mle_sxs[0][(inr + 1) / 2]) / avgeff[0];
               if (!(inr % 2) && inr == jnr && ymat[inr][jnr] <= 0)
                  ymat[inr][jnr] = mle_sxs[1][inr / 2] *
                     (1 - mle_sxs[1][inr / 2]) / avgeff[1];

               if (ymat[inr][jnr] <= 0)
                  printf (" %2.6f ", ymat[inr][jnr]);
               else
                  printf ("  %2.6f ", ymat[inr][jnr]);
            }
         }
         printf ("\n\n EFFECTIVE NUMBER OF MEIOSES");
         printf ("\n\n Interval  Female      Male");
         printf ("\n--------  --------    -------\n");
         for (i = 1; i < NMarks; i++)
         {
            j = 2 * i - 1;
            printf ("     %d: %7.0f  ", i,
                    ceil (mle_sxs[0][i] * (1 - mle_sxs[0][i]) / ymat[j][j]));
            printf ("  %7.0f \n",
                    ceil (mle_sxs[1][i] * (1 - mle_sxs[1][i]) /
                          ymat[j + 1][j + 1]));
         }
      }
      if (LikRatStats == 1)             /* ---------- Compute LRTs ---------- */
      {
         printf ("\n");
         for (i = 1; i <= 55; i++)
            printf ("=");
         printf ("\n");
         printf (" LM_MAP output for hypothesis testing: \n");
         for (i = 1; i <= 55; i++)
            printf ("=");

         printf ("\n\n Estimating the likelihood ratio statistic(s)...");
         tests = 1;
         accum_lr_flag = 1;
         llr = 0;
         for (j = 1; j < NMarks; j++)
         {
            Marek[0][j] = rf (.5 * (cmH (simple_null[0][j]) +
                                    cmH (simple_null[1][j])));
            Marek[1][j] = Marek[0][j];
         }
         cpy_map (Marek, se);
         dn = mle_sxa;
         if (MapTyp == BOTH)
            tests = 2;
         for (lrt = 0; lrt < tests; lrt++)
         {
            if (MapTyp == SEX_SPEC && lrt == 0)
            {
               dn = mle_sxs;
               accum_lr_flag = 1;
               llr = 0;
               cpy_map (simple_null, se);
            }
            if (lrt)
            {
               accum_lr_flag = 1;
               llr = 0;
               cpy_map (mle_sxa, se);
               dn = mle_sxs;
            }
            while (accum_lr_flag)
            {
               accum_lr_flag = 0;
               llr += logLR (se, dn, NSw, scount, recounts,
                             prob, temp_log_prob_seg, &accum_lr_flag, NChiSqIt);
            }
            chi_square[lrt] = 2 * llr;
         }
         switch (MapTyp)
         {
             case SEX_AVG:
                printf
                   ("\n\n Estimate of twice the log likelihood ratio of the\n");
                printf (" sex-averaged mle to the averaged given map: ");
                printf ("%6.4f\n", chi_square[0]);
                break;

             case SEX_SPEC:
                printf ("\n\n Estimate of twice the log likelihood ratio \n");
                printf (" between the sex-specific mle and the given map: ");
                printf ("%6.4f\n", chi_square[0]);
                break;

             case BOTH:
                printf
                   ("\n\n Estimate of twice the log likelihood ratio of the\n");
                printf (" sex-averaged mle and the averaged given map: ");
                printf ("%6.4f\n", chi_square[0]);

                printf
                   ("\n Estimate of twice the log likelihood ratio between \n");
                printf
                   (" the sex-specific and the sex-averaged ML estimates: ");
                printf ("%7.4f\n", (chi_square[1]));
                break;

             default:
                printf ("\n An ERROR occurred in estimating: 2*log(L(.,.))\n");
                break;
         }
      }
   }

   if (MapEstMod)                /*======== BEGIN ESTIMATION W/ ERROR ========*/
   {
      if (MapTyp == BOTH)
      {
         printf
            ("\n When using an error model, sex-averaged and sex-specific ");
         printf ("\n maps must be estimated from separate runs of lm_map. \n");
         exit (8);
      }
      epsi = GenoErrRat;
      LSamplProb = 1;                   /* The M-sampler is not implemented w/ error */

      /* NOTE: For that which follows in this LONG comment, i indexes the
         indivs. who provide marker phenotypes i = 0,1,...,(NObs-1), j
         indexes the MARKER LOCI ONLY, j = 1, ..., NMarks, and k indexes the
         indivs in the pedigree k = 0,1,...,(pedsize-1):

         The pair ( nghd[k].g_dat[2], nghd[k].g_dat[3] ) is a geno at the
         first locus of the kth individual, nghd[k].index = k, and
         nghd[k].self is the (possibly character valued) name of the kth
         individual.

         ATTENTION: some members within the nghds structure are NULL.         */

      /*   Allocate space for M-H MCMC variables.                             */
      /*   Note that the locus indexed of xx0[][] and xx1[][] goes
         from 1, ..., NMarks (the zero position is for the trait), and
         indiv. index goes from 0, ..., (pedsize-1).                        */
      /*   PGIF := (NMarks x 4 x no. families x ragged_family_size).          */

      /* SETUP */

      fam_list = malloc ((n_compnts + 1) * sizeof (int));
      alloc_chk (WHERE, fam_list);
      peel_seq = malloc ((n_compnts + 1) * sizeof (int *));
      alloc_chk (WHERE, peel_seq);
      for (i = 1; i <= n_compnts; i++)
      {
         fam_list[i] = 0;
         peel_seq[i] = malloc ((MXPSQ + 1) * sizeof (int));
         alloc_chk (WHERE, peel_seq[i]);
      }
      S0_curr = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, S0_curr);
      S1_curr = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, S1_curr);

      for (i = 0; i < pedsize; i++)
      {
         j = nghd[i].component;
         fam_list[j]++;
      }

      get_peelseq (peel_seq);

      curr_genos = malloc ((n_compnts + 1) * sizeof (int *));
      alloc_chk (WHERE, curr_genos);
      for (i = 1; i <= n_compnts; i++)
      {
         curr_genos[i] = malloc ((fam_list[i] + 1) * sizeof (int));
         alloc_chk (WHERE, curr_genos[i]);
         curr_genos[i][0] = fam_list[i];
      }
      PGIF = malloc ((NMarks + 1) * sizeof (int ***));
      alloc_chk (WHERE, PGIF);
      for (k = 1; k <= NMarks; k++)
      {
         PGIF[k] = malloc (5 * sizeof (int **));
         alloc_chk (WHERE, PGIF[k]);
         for (i = 0; i <= 4; i++)
         {
            PGIF[k][i] = malloc ((n_compnts + 1) * sizeof (int *));
            alloc_chk (WHERE, PGIF[k][i]);
            for (jj = 1; jj <= n_compnts; jj++)
            {
               PGIF[k][i][jj] = malloc ((fam_list[jj] + 1) * sizeof (int));
               alloc_chk (WHERE, PGIF[k][i][jj]);
               PGIF[k][i][jj][0] = fam_list[jj];
               for (j = 1; j <= PGIF[k][i][jj][0]; j++)
                  PGIF[k][i][jj][j] = 0;
            }
         }
      }

      if (SIM)                          /* By default, this is set to zero in Utils/setup_peel.h */
      {
         printf ("\n Simulated Marker Phenotypes:");
         for (i = 0; i < pedsize; i++)
         {
            for (j = 1; j <= NMarks; j++)
            {
               if (nghd[i].g_dat != NULL)
               {
                  if (nghd[i].g_dat[2 * j] != 0)
                  {
                     if (rans () < 0.05) /* a high value for testing */
                     {
                        v = loci[j].n_all;
                        vv = code_dna (nghd[i].g_dat[2 * j],
                                       nghd[i].g_dat[2 * j + 1], v);
                        obs = 0;
                        v = .5 * v * (v + 1);
                        while (obs == 0)
                        {
                           k = intran (1, v);
                           if (k != vv)
                              obs = 1;
                        }
                        v = loci[j].n_all;
                        uncode_dna (k, &nghd[i].g_dat[2 * j],
                                    &nghd[i].g_dat[2 * j + 1], v);
                        printf ("\n  INDIV %s  ", nghd[i].self);
                        printf (" %d %d   %d", nghd[i].g_dat[2 * j],
                                nghd[i].g_dat[2 * j + 1], j);
                     }
                  }
               }
            }
         }
         printf ("\n The marker phenotypic data");
         for (i = 0; i < pedsize; i++)
         {
            printf ("\n %s ", nghd[i].self);
            if (nghd[i].g_dat != NULL)
               for (j = 1; j <= NMarks; j++)
                  printf (" %d %d  ", nghd[i].g_dat[2 * j],
                          nghd[i].g_dat[2 * j + 1]);
         }
         printf ("\n");
      }

      for (jjj = 1; jjj <= NMarks; jjj++) /* Set the P, I, and F of PGIF */
      {
         i = 0;
         for (jj = 1; jj <= n_compnts; jj++)
         {
            for (j = 1; j <= fam_list[jj]; j++)
            {
               k = 2 * jjj;
               if (nghd[i].g_dat != NULL)
               {
                  PGIF[jjj][1][jj][j] = code_dna (nghd[i].g_dat[k],
                                                  nghd[i].g_dat[k + 1],
                                                  loci[jjj].n_all);
               }
               PGIF[jjj][3][jj][j] = i;
               if (data[i][2] == 0)
                  PGIF[jjj][4][jj][j] = 1;
               i++;
            }
         }
      }
      printf ("\n\n Constructing the initial state...");
      for (j = 1; j <= NMarks; j++)     /* Generate the initial G */
      {
         k = 0;
         for (jj = 1; jj <= n_compnts; jj++)
         {
            hr_vec[0] = locus_peel (1, epsi, jj, peel_seq[jj], j, PGIF[j], &sf,
                                    pene_mat[j], sege_mat[j]);
            for (i = 1; i <= fam_list[jj]; i++)
            {
               if (PGIF[j][1][jj][i] != 0)
               {
                  uncode_dna (PGIF[j][2][jj][i], &nghd[k].g_dat[2 * j],
                              &nghd[k].g_dat[(2 * j + 1)], loci[j].n_all);
               }
               k++;
            }
         }
      }
#if 0
      printf ("\n The initial latent genos");
      for (i = 0; i < pedsize; i++)
      {
         printf ("\n %s ", nghd[i].self);
         if (nghd[i].g_dat != NULL)
            for (j = 1; j <= NMarks; j++)
               printf (" %d %d  ", nghd[i].g_dat[2 * j],
                       nghd[i].g_dat[2 * j + 1]);
      }
#endif
      if (MapTyp == SEX_AVG)
      {
         get_init (0, simple_null, tmp_map, &epsi, MapEstMod);
         printf ("\n\n***************************\n");
         printf (" The initial sex-averaged map: \n");
         for (i = 1; i < NMarks; i++)
            printf (" %3d  %6.4f \n", i, tmp_map[0][i]);
         printf ("\n The initial mistyping rate: %6.2f%%\n\n", 100 * epsi);
      }
      if (MapTyp == SEX_SPEC)
      {
         get_init (1, simple_null, tmp_map, &epsi, MapEstMod);

         printf ("\n\n***************************\n");
         printf (" The initial sex-specific map: \n");
         for (i = 1; i < NMarks; i++)
            printf (" %3d  %6.4f  %6.4f \n", i, tmp_map[0][i], tmp_map[1][i]);
         printf ("\n The initial mistyping rate: %6.2f%%\n\n", 100 * epsi);
      }
      prob = setupInherit (1);
      reset_map (1, NSw, tmp_map, Marek, prob);

      n_scans = NIt;

      /* hr_vec[0]:=indicates acceptance of proposed state           */
      /* hr_vec[1]:=logP(prop G.j | S.+/-, Y.j)                      */
      /* hr_vec[2]:=logP( curr G.j)                                  */
      /* hr_vec[3]:=logP( curr G.j | S.+/-, Y.j)                     */
      /* hr_vec[4]:=logP(prop G.j)                                   */
      /* hr_vec[5]:=cumm. numb. of diffs between phenos and genos    */
      /* hr_vec[6]:=cumm. numb. of sames ...                         */
      /* hr_vec[7]:=per state version of hr_vec[5]                   */
      /* hr_vec[8]:=per state version of hr_vec[6]                   */
      /* after much simplification...                                */
      /* hratio := exp(hr_vec[1]+hr_vec[2]-hr_vec[3]-hr_vec[4])      */

      /* initialize obs_info_mat, scounts */
      for (i = 0; i < nLocy; i++)
         for (j = 0; j < nLocy; j++)
            obs_info_mat[i][j] = 0;

      for (c = 0; c < n_compnts; c++)
         for (j = 1; j < NMarks; j++)
            for (jj = 0; jj < 6; jj++)
               scount[c][j][jj] = 0;

      for (vv = 1; vv <= ItEM; vv++)    /* loop for ItEM MCEM iterates */
      {
         hr_vec[5] = hr_vec[6] = pgmidy = 0;
         if (vv != 1)
            n_scans *= 2;
         for (j = 0; j < NMarks; j++)
            for (c = 0; c < n_compnts; c++)
               for (jj = 0; jj < 6; ++jj)
                  recounts[c][j][jj] = 0;

         printf ("\n Map estimate %d based on %d realizations \n", vv, n_scans);
         for (k = 1; k <= 43; k++)
            printf ("-");
         printf ("\n");

         for (k = 1; k <= n_scans; k++) /* Run n_scans steps of the chain   */
         {
            /* Basically, this section of code: defined from here to call of */
            /* sampling(), should form a new function: samplewerr(), so that */
            /* any Morgan program can use the (GS given Y) sampler.          */

            /* printf (" %3d", k);      Track current run time               */

            hr_vec[7] = 0;
            hr_vec[8] = 0;
            flip = NMarks + 1 - flip;
            for (j = 1; j <= NMarks; j++) /* Do one complete scan */
            {
               pj = fabs (flip - j);
               pedstart = 0;

               for (jj = 1; jj <= n_compnts; jj++)
               {
                  i = pedstart;
                  total++;
                  for (jjj = 0; jjj <= 4; jjj++)
                     hr_vec[jjj] = 0;

                  /* set curr G.pj */
                  for (jjj = 1; jjj <= fam_list[jj]; jjj++)
                  {
                     if (nghd[i].g_dat != NULL)
                        curr_genos[jj][jjj] = code_dna (nghd[i].g_dat[2 * pj],
                                                        nghd[i].g_dat[(2 * pj +
                                                                       1)],
                                                        loci[pj].n_all);
                     else
                        curr_genos[jj][jjj] = 0;
                     PGIF[pj][2][jj][jjj] = curr_genos[jj][jjj];
                     i++;
                  }

                  /* calc logP(curr G_jj,pj) */
                  tau = log (locus_peel (0, 0, jj, peel_seq[jj],
                                         pj, PGIF[pj], &sf, pene_mat[pj],
                                         sege_mat[pj]));
                  hr_vec[2] += (tau + sf * log (SMALL));

                  swcpy (xx0[pj], xx1[pj], S0_curr, S1_curr);

                  /* sample S.pj given curr G.pj */
                  /* and calc logP(G.pj|...)     */
                  hr_vec[3] = do_lsmpl (NMarks, pj, prob, nFGL);

                  /* sample prop G.pj given Y.pj */
                  i = pedstart;
                  tau = locus_peel (1, epsi, jj, peel_seq[jj], pj, PGIF[pj],
                                    &sf, pene_mat[pj], sege_mat[pj]);

                  /* calc logP(prop G_jj,pj)     */
                  tau = log (locus_peel (0, 0, jj, peel_seq[jj], pj, PGIF[pj],
                                         &sf, pene_mat[pj], sege_mat[pj]));
                  hr_vec[4] += (tau + sf * log (SMALL));

                  /* reset nghd and prob using prop G_jj,pj */
                  for (jjj = 1; jjj <= fam_list[jj]; jjj++)
                  {
                     if (nghd[i].g_dat != NULL)
                        uncode_dna (PGIF[pj][2][jj][jjj],
                                    &nghd[i].g_dat[2 * pj],
                                    &nghd[i].g_dat[(2 * pj + 1)],
                                    loci[pj].n_all);
                     i++;
                  }
                  locus_peel_free ();
                  locus_peel_init (0, 1);

                  /* sample S.pj given prop G.pj   */
                  /* and calc logP(prop G.,pj|...) */
                  hr_vec[1] = do_lsmpl (NMarks, pj, prob, nFGL);

                  decide (hr_vec);
                  if (!hr_vec[0])       /* if (reject) */
                  {
                     reject_rate++;
                     swcpy (S0_curr, S1_curr, xx0[pj], xx1[pj]);
                     i = pedstart;
                     for (jjj = 1; jjj <= fam_list[jj]; jjj++)
                     {
                        if (nghd[i].g_dat != NULL)
                        {
                           uncode_dna (curr_genos[jj][jjj],
                                       &nghd[i].g_dat[2 * pj],
                                       &nghd[i].g_dat[(2 * pj + 1)],
                                       loci[pj].n_all);
                           PGIF[pj][2][jj][jjj] = curr_genos[jj][jjj];
                        }
                        i++;
                     }
                     locus_peel_free ();
                     locus_peel_init (0, 1);
                  }
                  pedstart += fam_list[jj];
               }
            }

            sampling (0, 1, NMarks, NSw, nFGL, prob); /* move in S space */

            /* Finished sampling (G,S | Y) */

            for (c = 0; c < n_compnts; ++c) /* calc stats for this (G,S) */
            {
               full_seg_prob (NMarks, c, recounts, temp_log_prob_seg);
               for (j = 1; j <= NMarks; j++)
                  for (jj = 1; jj <= fam_list[c + 1]; jj++)
                  {
                     if (PGIF[j][1][c + 1][jj] != 0 &&
                         PGIF[j][1][c + 1][jj] != PGIF[j][2][c + 1][jj])
                     {
                        hr_vec[7]++;
                        if (vv == ItEM) /* if last step, calc post probs */
                           PGIF[j][0][c + 1][jj]++;
                     }
                     else
                        if ((PGIF[j][1][c + 1][jj] + PGIF[j][2][c + 1][jj]) > 0)
                        hr_vec[8]++;
                  }
            }
            if (hr_vec[7] == 0)
               pgmidy++;
            hr_vec[5] += hr_vec[7];
            hr_vec[6] += hr_vec[8];

            /* if(last step), get contributions to E[ (nabla)^2 ] */
            if (vv == ItEM && MapTyp == SEX_AVG)
            {
               for (c = 0; c < n_compnts; c++)
                  for (j = 1; j < NMarks; j++)
                     for (jj = 0; jj < 6; jj++)
                        scount[c][j][jj] =
                           recounts[c][j][jj] - scount[c][j][jj];
               for (j = 1; j < NMarks; j++)
               {
                  set_rc (j, scount, rc);
                  obs_info_mat[j][0] = (rc[0][0] + rc[1][0]) / mle_sxa[0][j];
                  obs_info_mat[j][0] -=
                     (rc[0][1] + rc[1][1]) / (1 - mle_sxa[0][j]);
               }
               obs_info_mat[NMarks][0] =
                  hr_vec[7] / epsi - hr_vec[8] / (1 - epsi);
               for (j = 1; j <= NMarks; j++)
                  for (i = 1; i <= NMarks; i++)
                     obs_info_mat[j][i] +=
                        obs_info_mat[j][0] * obs_info_mat[i][0];
               for (c = 0; c < n_compnts; c++)
                  for (j = 1; j < NMarks; j++)
                     for (jj = 0; jj < 6; jj++)
                        scount[c][j][jj] = recounts[c][j][jj];
            }
         }
         tau = epsi;                    /* store previous value of epsi at last step */
         epsi = hr_vec[5] / (hr_vec[5] + hr_vec[6]); /* update epsi */
         if (epsi == 0)
            epsi = .001;
         if (MapTyp == SEX_AVG)
         {
            for (j = 1; j < NMarks; j++)
            {
               set_rc (j, recounts, rc);
               mle_sxa[0][j] = (rc[0][0] + rc[1][0]) / (rc[0][2] + rc[1][2]);
               if (fabs (mle_sxa[0][j] - 0.25) >= (0.25 - RF_bound))
                  mle_sxa[0][j] = Marek[0][j];
               if (vv != ItEM)
                  mle_sxa[1][j] = mle_sxa[0][j];
               /* on the last step, don't copy so mle_sxa[1] can be used later */
            }
            reset_map (0, NSw, mle_sxa, Marek, prob);
            pr_map (0, Marek);
            printf ("\n Estimated mistyping rate: %6.2f%%\n\n", 100 * epsi);
         }
         if (MapTyp == SEX_SPEC)
         {
            for (j = 1; j < NMarks; j++)
            {
               set_rc (j, recounts, rc);
               mle_sxs[0][j] = rc[0][0] / rc[0][2];
               mle_sxs[1][j] = rc[1][0] / rc[1][2];
               if (fabs (mle_sxs[0][j] - 0.25) >= (0.25 - RF_bound))
                  mle_sxs[0][j] = Marek[0][j];
               if (fabs (mle_sxs[1][j] - 0.25) >= (0.25 - RF_bound))
                  mle_sxs[1][j] = Marek[1][j];
            }
            reset_map (0, NSw, mle_sxs, Marek, prob);
            pr_map (1, Marek);
            printf ("\n Estimated mistyping rate: %6.2f%%\n\n", 100 * epsi);
         }
      }
      if (MapTyp == SEX_AVG)
      {
         /* normalize the sums in obs_info_mat */
         for (i = 1; i <= NMarks; i++)
            for (j = 1; j <= NMarks; j++)
               obs_info_mat[i][j] /= n_scans;

         /* finished computing E[ (nabla)^2 ]                    */
         /* get contributions to E[nabla(nabla)] and E[nabla]    */
         /* subtract off ( E[nabla] )^2                          */
         printf ("\n\n");
         for (j = 1; j < NMarks; j++)   /* add E[ nabla(nabla) ] */
         {
            set_rc (j, recounts, rc);
            obs_info_mat[j][j] -=
               ((rc[0][0] + rc[1][0]) / pow (mle_sxa[1][j],
                                             2) + (rc[0][1] +
                                                   rc[1][1]) / pow ((1 -
                                                                     mle_sxa[1]
                                                                     [j]),
                                                                    2)) /
               n_scans;
            obs_info_mat[j][0] =
               ((rc[0][0] + rc[1][0]) / mle_sxa[1][j] -
                (rc[0][1] + rc[1][1]) / (1 - mle_sxa[1][j])) / n_scans;
         }
         obs_info_mat[NMarks][NMarks] -= (hr_vec[5] / (tau * tau) +
                                          hr_vec[6] / ((1 - tau) * (1 - tau))) /
            n_scans;
         obs_info_mat[NMarks][0] =
            (hr_vec[5] / tau - hr_vec[6] / (1 - tau)) / n_scans;
         for (j = 1; j <= NMarks; j++)
            for (i = 1; i <= NMarks; i++)
               obs_info_mat[j][i] -= obs_info_mat[j][0] * obs_info_mat[i][0];
         for (i = 1; i <= NMarks; i++)
            for (j = 1; j <= NMarks; j++)
               anr[i][j] = -obs_info_mat[i][j];
         /* observed information now in anr[.][.] */

         printf ("\n ===============================================");
         nnr = NMarks;
         ludcmp (anr, nnr, indxnr, d);
         for (jnr = 1; jnr <= nnr; jnr++)
         {
            for (inr = 1; inr <= nnr; inr++)
               bnr[inr] = 0.0;
            bnr[jnr] = 1.0;
            lubksb (anr, nnr, indxnr, bnr);
            for (inr = 1; inr <= nnr; inr++)
               ymat[inr][jnr] = bnr[inr];
         }
         printf ("\n\n Estimated gradient of logL(mle): ");
         for (j = 1; j <= NMarks; j++)
         {
            printf (" %6.4f ", obs_info_mat[j][0]);
            for (i = 1; i < NMarks; i++)
               obs_info_mat[j][i] -= obs_info_mat[j][0] * obs_info_mat[i][0];
         }
         printf ("\n\n\n       MAXIMUM LIKELIHOOD ESTIMATES\n");
         printf ("\n  Interval        Sex-Averaged (RF)");
         printf ("\n  --------        ----------------\n");
         for (j = 1; j < NMarks; j++)
            printf ("    %d         %15.4f \n", j, mle_sxa[0][j]);
         printf ("\n       MISTYPING RATE: %3.6f%%\n", epsi * 100);
         printf
            ("\n\n ESTIMATED VARIANCE OF (MAP,MISTYPING RATE) SEX-AVERAGED");
         printf ("\n------------------------------------------------------");
         for (inr = 1; inr <= nnr; inr++)
         {
            printf ("\n  ");
            for (jnr = 1; jnr <= nnr; jnr++)
            {
               if (ymat[inr][jnr] < 0)
                  printf (" %2.6f ", ymat[inr][jnr]);
               else
                  printf ("  %2.6f ", ymat[inr][jnr]);
            }
         }
         printf ("\n\n\n EFFECTIVE NUMBER OF MEIOSES\n");
         printf ("------------------------------\n");
         for (i = 1; i < NMarks; i++)
            printf ("     Interval %d: %1.0f \n", i,
                    ceil (mle_sxa[0][i] * (1 - mle_sxa[0][i]) / ymat[i][i]));

      }
      if (MapTyp == SEX_SPEC)
      {
         printf ("\n       MAXIMUM LIKELIHOOD ESTIMATES\n");
         printf ("\n  Interval        Female (RF)        Male (RF)");
         printf ("\n  --------        ------------       ---------\n");
         for (j = 1; j < NMarks; j++)
            printf ("    %d         %15.4f %15.4f \n", j, mle_sxs[0][j],
                    mle_sxs[1][j]);
         printf ("\n       MISTYPING RATE: %3.6f%%\n", epsi * 100);
      }

      printf ("\n       LATENT GENOS = OBS PHENOS ");
      printf ("%3d out of %3d \n", pgmidy, n_scans);
      printf ("\n");
      printf ("\n POSTERIOR PROBS OF MISTYPING: \n");
      for (i = 1; i <= n_compnts; i++)
      {
         for (j = 1; j <= fam_list[i]; j++)
         {
            printf ("\n %s ", nghd[PGIF[1][3][i][j]].self);
            for (k = 1; k <= NMarks; k++)
               printf (" %1.2f ", (double) PGIF[k][0][i][j] / (double) n_scans);
         }
      }
      printf ("\n\n");
#if 0
      for (c = 0; c < n_compnts; ++c)
         pr_reco_counts (0, NMarks, recounts[c]);
#endif
      printf (" The M-H rejection rate was %6.4f \n", reject_rate / total);

         /*---------- COMPUTE CHI-SQUARE STATISTICS W/ ERROR --------------*/
      if (LikRatStats == 1)
      {
         printf ("\n When using an error model, lm_map does not test ");
         printf ("\n hypotheses about the map.\n");
      }
   }

   printf ("\n");
   for (i = 1; i <= 55; i++)
      printf ("=");
   printf ("\n LM_MAP execution is complete. \n");
   for (i = 1; i <= 55; i++)
      printf ("=");
   printf ("\n");

   fflush (stdout);

   if (SamplSeedsOut)
   {
      getSeeds (SAMPL);
      if (!(oSeedF = open_fil (OSEED)))
         give_up (ERR_qit);
      writSeeds (oSeedF);
      fclose (oSeedF);
   }

#include "Utils/free.mem"               /* frees all allocated memory */

   exit (EXIT_SUCCESS);
}
