/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#ifndef GL_LODS_H
#define GL_LODS_H

#include "IBDgraph.h"

#ifdef  DEFINE_DATA

/******************************************************************************/
/*                                                                            */
/* Storage for all the lodscore_elements.  This is a number of IBDgraphs by   */
/* number of selected markers matrix, where each entry in the matrix is a     */
/* lodscore_element (LSCORE_ELEM).  Only one lodscore_element in each         */
/* equivalence class will actually be used for computing the contribution to  */
/* the log likelihood.  The others will not be used.                          */
/*                                                                            */
/******************************************************************************/

GLOBAL LSCORE_ELEM **lses_storage;      /* Storage for all lodscore_elements. */

/******************************************************************************/
/*                                                                            */
/* Vector of equivalence structures.  This is a number of selected markers    */
/* long vector of marker_classes structures (MARK_CLASS).                     */
/*                                                                            */
/******************************************************************************/

GLOBAL MARK_CLASS *equivs_vector;       /* Vector of equivalence structures.  */

/******************************************************************************/
/*                                                                            */
/* Assigned lodscore_element pointers matrix.  This is a number of IBDgraphs  */
/* by number of selected markers matrix, where each entry in the matrix is a  */
/* lodscore_element pointer (LSE_PTR).  The pointers in this matrix point to  */
/* the lodscore_elements in lses_storage.  It is by reassigning the pointers  */
/* in this matrix that a single lodscore_element will be used for all the     */
/* graphs in a single equivalence class.                                      */
/*                                                                            */
/******************************************************************************/

GLOBAL LSE_PTR **p_lse_index;           /* Lodscore_element pointers matrix.  */

/******************************************************************************/
/*                                                                            */
/* Graph class subscripts matrix.  This is a number of IBDgraphs by number of */
/* selected markers matrix, where each entry in the matrix is a long integer, */
/* which is the class_index for the corresponding IBDgraph and selected       */
/* marker.                                                                    */
/*                                                                            */
/******************************************************************************/

GLOBAL long **class_index;              /* Graph class subscripts matrix.     */

GLOBAL long *comp_first_line_num;       /* The first line number for each of  */
                                        /* the components being input.        */
GLOBAL long input_num_graphs;           /* Number of IBDgraphs read in from   */
                                        /* the unprocessed IBD graphs file.   */
                                        /* Set to the number of IBDgraphs     */
                                        /* read in for the first component.   */
GLOBAL long use_num_graphs;             /* Number of IBDgraphs to use.        */
GLOBAL long lines_per_graph;            /* Lines in each IBDgraph realization.*/
GLOBAL int *comp_first_indiv_index;     /* DGL_index subscript for the first  */
                                        /* individual in each component.      */
GLOBAL int *comp_all_indivs;            /* Number of all individuals in each  */
                                        /* component.                         */
GLOBAL int *comp_included;              /* Vector of binary flags indicating  */
                                        /* which components are included in   */
                                        /* the unprocessed IBDgraphs file.    */
                                        /* If an entry is a 1, the component  */
                                        /* was included.  Otherwise it wasn't.*/
GLOBAL int  num_comp_input;             /* Number of components read in from  */
                                        /* the unprocessed IBD graphs file.   */
GLOBAL int  tot_num_indivs;             /* Total number of individuals found  */
                                        /* in the unprocessed IBDgraphs file. */
GLOBAL char **DGL_names;
GLOBAL int *DGL_index;

#else
GLOBAL LSCORE_ELEM **lses_storage;      /* Storage for all lodscore_elements. */
GLOBAL MARK_CLASS *equivs_vector;       /* Vector of equivalence structures.  */
GLOBAL LSE_PTR **p_lse_index;           /* Lodscore_element pointers matrix.  */
GLOBAL long **class_index;              /* Graph class subscripts matrix.     */

GLOBAL long *comp_first_line_num;       /* Each component's first line number.*/
GLOBAL long input_num_graphs;           /* Number of IBDgraphs input.         */
GLOBAL long use_num_graphs;             /* Number of IBDgraphs to use.        */
GLOBAL long lines_per_graph;            /* Lines in each IBDgraph realization.*/
GLOBAL int *comp_first_indiv_index;     /* Subscript of each first individual.*/
GLOBAL int *comp_all_indivs;            /* Number individuals in each compon. */
GLOBAL int *comp_included;              /* Flag the components in ibd_graphs. */
GLOBAL int  num_comp_input;             /* Number of components from input.   */
GLOBAL int  tot_num_indivs;             /* Total number of individuals.       */
GLOBAL char **DGL_names;
GLOBAL int *DGL_index;
#endif  /*DEFINE_DATA*/

/* Subroutines currently in the gl_lods_rtnes.c file.                         */

int      dgl_index (void);
int      dgl_strip (int *, int);
int      copy_line (FILE *, FILE *, int, int);

void     fgl_set (FILE *, int);
void     set_line (FILE *, int **);
void     check_fgl (void);

void     find_num_indivs (int *);
int      compute_num_indivs (long, int);

void     find_max_fgl (int *);
void     find_comp_no (long, int *);

int      get_component_graphs (int, int *);
int      end_of_realization (char *, char *, char *, int);

void     gl_lods_alloc (long);
void     gl_lods_free (void);

void     det_equivalence_classes (IBDGraphList *);
void     initialize_lodscore_element (LSE_PTR);

void     collapse_lodscore_elements (IBDGraphLocationEquivalences *, long);
LSE_PTR  do_single_range (LSE_PTR, long, long, long);

void     prnt_contributions (FILE *, long, int, int);
void     do_single_lse (int, long, long);

long     comp_mark_ndx (long);
double   comp_loglike (int, long);

void     gl_lods_loglikes (int);
void     prnt_gl_lods (int, double *);

void     free_DGL_vectors (void);

/* Debugging subroutines in the lods_debug.c file.                            */

void     lods_prnt_trait_data (void);

#endif  /* GL_LODS_H */
