/******************************************************************************/
/*                                                                            */
/* Header file for the ibd_haplo program -- taking stuff out of ibd_haplo.c.  */
/* Program to do a basic analysis of IBD inference with no LD.                */
/*                                                                            */
/******************************************************************************/

struct ss_size_pointers_structure
{
   int   **state_matrix;          /* The current IBD state matrix.            */
   int   num_orbits;              /* Number of nontrivial orbits.             */
   int   *orbitSizes;             /* The vector of sizes of each orbit.       */
   int   **orbits;                /* The matrix of the orbit members.         */
   double *marg_probs;            /* The marginal ibd probabilities vector.   */
   double *tmat_rate;             /* Rate vector of leaving each state.       */
   double **tmat;                 /* The transition rate matrix which then    */
                                  /* becomes the jump chain matrix.           */
   double **transmat;             /* The locus-to-locus transition matrix.    */
};
typedef struct ss_size_pointers_structure SS_SIZE_STRUCTURE;

#define IS_SET(x,n) ((x) & (1 << (n)))

#define CONCORDANT   1            /* Concordant phasing at current marker.    */
#define DISCORDANT  -1            /* Discordant phasing at current marker.    */

/* Vector of score_set_size_pointers structures.                              */
SS_SIZE_STRUCTURE *score_set_size_pointers;

int   num_states;                 /* Number of possible IBD states.           */

double *penmat;                   /* nph pheno, num_states states.            */
double **transmat;              /* num_states x num_states transition matrix. */
                                  /* penmat and transmat now reset each locus */
                                  /*  and penmat only for given data.         */

double *marg_probs;               /* Marginal ibd probabilities.              */
double *tmat_rate;                /* Rate vector of leaving each state.       */
double **tmat;                    /* The transition rate matrix which then    */
                                  /* becomes the jump chain matrix; used to   */
                                  /* determine locus-to-locus transmat.       */

double *allefreq;                 /* Allele frequencies at each marker.       */
double *snppos;                   /* Selected marker positions.               */

int   **phasdatindiv;             /* Phasing data for individuals.            */
int   *phasdat;                   /* Phasing data for pairs of individuals.   */

int   *duplicated_rows;           /* A vector of 0's and 1's reflecting where */
                                  /* there are duplicate rows.                */
int   **state_matrix;             /* The IBD state matrix with all duplicated */
                                  /* rows removed.                            */

                                  /* Stuff for computing genotypic orbits.    */
int   num_orbits;                 /* Number of nontrivial orbits.             */
int   *orbitSizes;                /* The vector of sizes of each orbit.       */
int   **orbits;                   /* The matrix of the orbit members.         */

/* Subroutines currently in the ibd_haplo.c file:                             */
/* The debugging printing functions are declared below along with all the     */
/* other debugging printing functions.                                        */

void set_ProbIx (void);
void init_ibd_haplo (void);

/* Subroutines currently in the compute_ibd_probs.c file:                     */
/* The debugging printing functions are declared below along with all the     */
/* other debugging printing functions.                                        */

void init_ibd_lengths (void);
int  compute_ibd_label (int, int *);
void compute_ibd_probs (int, int, FILE *, FILE *);

/* Subroutines currently in the haplo_process.c file:                         */

void make_pair_data (int *, int **, int, int, int, int *);
void make_data (int *, int **, int, int, int, int, int);
void print_haplo (FILE *, int *, int);

/* Subroutines currently in the transitions.c file:                           */
/* The debugging printing functions are declared below along with all the     */
/* other debugging printing functions.                                        */

char *compute_ibd_state (int, int);
void unique_int (int, int *, int *, int **);
void rplc (int, int *, int *);
void get_ibd_state_mat (int, int *, int **, int ***);
double **transitions (int, int *, double);
void set_tmat_2011 (int, double, double ***);

/* Subroutines currently in the hap_fwd_bk_compute.c file:                    */

void avg_orbits (double *);
void reorder_tmat (void);
void set_tmat_2009 (void);
void hap_set_penmat (double, double *, int, int, int, double, int, int, int **);
void set_marg_ibd (int, int, double *, int **);
void set_transmat (int, int, double, double *, double **);
void hap_fwd_compute (int *, int, int, double *, double *, double *, double **,
                      int, int *, int);
void hap_bck_compute (int *, int, int, double *, double *, double *, double **,
                      int, int *, int, double **);
void hap_sample (int, int, int *, double *, double *, double **, int, int);

/* Subroutines currently in the orbits.c file:                                */

void swap (int *, int *);
int  power (int, int);
int  stateEncode (int, int *);
void makeOrbits (int, int, int **);
int  compute_number_orbits (void);
void collapse_to_distinct (double **, int, double ***, int, char ***);

/* Subroutines currently in the haplo_debug.c file:                           */

void print_one_line (char *, double, int, int, double **, int, FILE *);
void haplo_debug_phased (int, int, double **, FILE *);
void haplo_debug_unphased (int, int, double **, char **, FILE *);
void haplo_debug_initial_stderr (int, double **);
void haplo_debug_qq_ibd (int, int, double **, FILE *);
void print_phase_data (int, int);

