/******************************************************************************/
/*                                                                            */
/* Header file for declarations shared by more than one program found in the  */
/* IBD_Haplo directory.  As of November 26, 2014, that means declarations     */
/* shared by the ibd_haplo and ibd_create programs.                           */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* Define four integer variables where various lengths needed by the          */
/* compute_ibd_probs and the gen_fgl2ibd_labels subroutines can be kept.      */
/*                                                                            */
/******************************************************************************/

GLOBAL int nc;                          /* Character length of longest name.  */
GLOBAL int nm;                          /* Character length of highest        */
/*                                         numbered selected marker.          */
GLOBAL int nsn;                         /* Character length of highest        */
/*                                         numbered score set.                */
GLOBAL int nst;                         /* Character length of largest number */
/*                                         of possible IBD states.            */

/******************************************************************************/
/*                                                                            */
/* The following two variables are needed to handle the DoingAllPairs case.   */
/* By making these variables available here it is not necessary to pass them  */
/* as arguments to the several subroutines required to handle this case.      */
/*                                                                            */
/******************************************************************************/

GLOBAL char ** indiv_names;             /* Vector of individual names, which  */
/*                                         when running ibd_haplo is ObsNam,  */
/*                                         but for ibd_create must be set up  */
/*                                         by the make_names_vector routine.  */
GLOBAL int num_names;                   /* Number of individuals named in the */
/*                                         vector of individual names.        */

/******************************************************************************/
/*                                                                            */
/* If the user has requested "output four-gamete state order jacquard", so    */
/* that Lexico is set to FALSE, and if the score set size is 4, then the      */
/* following three integer vectors can be used to either get the jacquard     */
/* order subscript (jacquard_index) or to output from lexicographically       */
/* ordered data into genetic order (lex2gen_ph and lex2gen_un).  Note that    */
/* the first two vectors are in some sense the reverse of each other.         */
/*                                                                            */
/* The jacquard_index vector is used by the reorder_tmat subroutine in        */
/* ibd_haplo and the fgl2ibd subprogram in ibd_create.  The lex2gen_ph and    */
/* lex2gen_un vectors are used by the ibd_haplo main program.                 */
/*                                                                            */
/******************************************************************************/

#ifdef  DEFINE_DATA
GLOBAL int jacquard_index[] = {0, 2, 3, 1, 4, 5, 8, 10, 9, 6, 12, 11, 13, 7, 14};
GLOBAL int lex2gen_ph[] = {0, 3, 1, 2, 4, 5, 9, 13, 6, 8, 7, 11, 10, 12, 14};
GLOBAL int lex2gen_un[] = {0, 2, 1, 3, 4, 7, 5, 6, 8};
#else
GLOBAL int jacquard_index[];
GLOBAL int lex2gen_ph[];
GLOBAL int lex2gen_un[];
#endif  /*DEFINE_DATA*/

/* The following subroutines are needed by both the ibd_haplo and the         */
/* ibd_create programs.  These subroutines were moved from the ibd_haplo.c    */
/* file to the ibd_create_rtnes.c file on November 6, 2014.            SML    */

void alloc_for_pairs (void);
void init_ibd_lengths (void);
void set_for_pair (int, int);

