/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/******************************************************************************/
/*                                                                            */
/*       markers.h:  Header file for the Markers library.                     */
/*                -  It is included by Markers functions and by main programs */
/*                   calling these functions.                                 */
/*                                                                            */
/******************************************************************************/

#ifndef MARKERS_H
#define MARKERS_H

#include "incs.h"
#include "nghds.h"

/**
 * @addtogroup markers libmarkers
 * @ingroup Libraries
 * @{
 */

typedef double (*Measure) (int, int, int * , double * );
     /* ibd measure function: needs to move to traits eventually !! */
typedef struct affcount
{ /* A new structure for counting number of affected and unaffected at each generation of a pedigree */
        int count_aff;
        int count_unaff;
        int level;
        struct affcount* next;
}Affcount;


GLOBAL int ***lpflag;         /* [2][nh+1][NMarks]  flags set by set_lpflag
                               * to indicate when log-likelihoods may differ
                               * and must be included in probability ratio
                               * (note: NLoci is NMarks + NTraits)           */

GLOBAL int **xx0, **xx1;      /* arrays of maternal and paternal segregation
                                 indicators at all the loci                  */
GLOBAL int **gg0, **gg1;      /* arrays of all labelled founder genes in
                                 the pedigree                                */

                              /* (note: NLoci is NMarks + NTraits)           */
GLOBAL int ** chromo;         /* [nh+1][NLoci]  indices of loci in ordered
                               * chromosome, for each set of recomb values   */
GLOBAL int ** revchro;        /* [nh+1][NLoci]  positions of loci in ordered
                               * chromosome --  "inverse" of chromo[][]:
                               * (chromo[][pos] = nam, revchro[][nam] = pos) */
GLOBAL double *** xecomo;     /* [2][nh+1][NLoci]  sex-valued recombinations
                               * in ordered chromosome                       */

void  setObsDat (void);
void  setMaps (int);
void  setLocTyp (void);

int   read_gen_dat (FILE *, int **, int **, int *);

void  set_switches (int, int, int *, int *);
void  read_segs (FILE *, int, int, int);
int   only_kid (Ind *, Ind *);
void  set_chromosomes (int, double *, int, int *, double *, int);
void  set_lpflag (int, int, int);
void  invert_label (int, int, int *);
int   next_state (int, int *, int *);
int   label_state (int, int *);
int   index_ibd (int, int *, int *, int *);
void  get_genes (int, int, int *, int *, int *);
int   score_ibd (int, int *);

void  print_set_up (int);
void  print_obs_genes (int, int ***);
void  print_scores (FILE *, int, int, int);
void  prnt_xxx (FILE *, int, int **, int);
void  prnt_fgl (FILE *, int **, int);
void  pr_pvals (int, double **, int, int);
void  pr_log_prob_seg (int, int, double, double *, double);
void  pr_log_prob_segl(int, int, double, double *, double *, double *, double);
void  pr_log_prob_pen (int, int, double *, int);
void  pr_reco_counts (int, int, int **);
void  pr_histo (int, int ***);
void  prntLodScores (int, double *, double *);
void  prntLodSHead (char *);
void  prntLodSJHead (void);
void  prntLodSColHd (void);
void  prntLodSPos0 (void);
int   prntLodSPos (int);
int   score_either (int, int, int);
int   score_whichever (int, int, int, int);
void  free_allefrq (void);
void  print_chromosome (int, int, int);

  /* aloc_affhist allocates and initializes the histogram               */
  /* scoraff calls ibd Measure, such as max_fgl, and sets the histogram  */
  /* make_quantiles converts the histogram to 0.1 % quantiles of the dsn */
  /*   non-histogram versions implemented July 2005, EAT                  */

int ***alloc_affhist (void);
double **alloc_hdsn (int,int);
void  score_aff (Measure , int ***);
void  setup_ibdmeas (Measure);
int   setupAutozyg (int, char **, char *, int);
void  setupwithIndivList (int, char **, char *, int);
double **setupInherit (int);
void  multi_score (Measure , int ***);
void  multi_gen_score (int, Measure , int, double **);
void score_all (Measure , double **, int , int , double **, double *);
void scor_pval (double , int , double *, double *);
void  make_quantiles(int , int, double *, double *);
void compress_marginals (int, int, int, double **);
void  pr_quantiles (int, int , int, int, double **);

  /* the following are the current actual scoring routines              */
double max_fgl (int, int, int *, double *);
double spair (int, int, int *, double *);
double robdom (int, int, int *, double *);
double Saffunaff (int, int, int *, double *);
double Calculate_Likelihood_Lambda (int, int , int * , double * );
void   free_ibd_measures ();

void  norcum (int , int ,  double *);

int score_genes (int, int, int *, int *, int *,
        int *, int *, double *, int, int *);
void score_windows(double **, int , int *, int );
void score_xact(int, int *, int *, int *,          /* added EAT 01/2009 */
        int *, int *, double *, double);

int   ibd_pattern (int, int *, int *);
int   bin_pattern (int, int, int *);
void  binary (int, int, int *);

/* While updating the frees in Morgan, I found a need to call free_chromo and */
/* free_lpflag using the atexit capabilities in C.  However, subroutines used */
/* by atexit are not permitted to have arguments.  So the free_chromo_atexit  */
/* and the free_lpflag_atexit subroutines were added to Morgan.  A variable,  */
/* NDisPos, was added to make this possible.  Eventually the free_chromo and  */
/* free_lpflag subroutines should be modified to not need any arguments, so   */
/* they may be used by atexit; the two new subroutines may then be deleted    */
/* from Morgan.                                     Steven Lewis,  Dec, 2013. */
void  free_chromo_atexit (void),      free_lpflag_atexit (void);

void  alloc_chromo (int, int),        free_chromo (int);
void  alloc_lpflag (int, int),        free_lpflag (int);
void  alloc_logprob (void),           free_logprob (void);
void  ali_nc      (int **),           ali_nl      (int **),
      ali_n_nl    (int ***, int),     ali_nc_n    (int ***, int),
      ali_nl_n    (int ***, int),     ali_nc_nl   (int ***),
      ali_nc_nl_n (int ****, int);
void  ald_nl      (double **),        ald_nc_n    (double ***, int),
      ald_nl_n    (double ***, int),  ald_nc_nl   (double ***),
      ald_nc_nl_n (double ****, int), ald_nc_nl_ns(double ****);
void  fre_n       (void **),          fre_n_n     (void ***, int),
      fre_nc_n    (void ***),         fre_nl_n    (void ***),
      fre_nc_nl_n (void ****);

/* following moved from Autozyg (mostly)  EAT  feb 2002 */
void  alloc_ibd   (int **, int ***, int ***, int ***, int **,
                   int ***, double *****, double **);
void  alloc_wndw  (int **, int **, double *****, int, int);
void  alloc_loglik(int ****, double ***, double ***, double ***);
void  alloc_lod   (int **, int ***, double ***,
                   double ***, double ***, double ***, double ***);
void  alloc_lf_rt (int, double ***, double ***, int **, int **);
void  alloc_rb    (int, int, double **, double **, double ***, double **);
void  realloc_lf_rt (int, double ***, double ***, int **, int **);

void  free_ibd    (int **, int ***, int ***, int ***, int **,
                   int ***, double *****, double **);
void  free_cst    (void);
void  free_wndw   (int **, int **, double *****, int, int);
void  free_loglik (int ****, double ***, double ***, double ***);
void  free_lod    (int **, int ***, double ***, double ***, double ***,
                   double ***, double ***);
void  set_traitdat(int, int, int);
void  prnt_traitdat(int, int);
int   set_probgam (int);
void  set_affind  (int, int);

/* Print list of affected individuals, by component */
void  print_affind (void);
/* print functions for lm_auto and ibddrop */
void  pr_npatsets (int *);
void  pr_pgamset  (int, int);
void  pr_wndwpat  (int, int);
void  pr_cmpnthead(int);
void  pr_probhead (void);
void  pr_prbihead (void);
void  pr_prbwhead (int, int);
void  pr_lochead  (int **);
void  pr_locprob  (int, int *, double *, int);
void  pr_wndwhead (int);
void  pr_wndwprob (int, int, int, int, int *, double ****);

void    alloc_xxgg  (int, int);
void    free_xxgg   (void);
int  ** alloc_genes (int, int);
void    free_genes  (int ***);

/* Following routines in set_markdat.c:  EAT March 2002.                      */

void unset_recval   (double **, double **, double **, double **);
void reset_recval   (int, double **, double **, double **, double **);
void set_markmap    (int , int , int *);
void reset_traitmap (int, int, int*);    /* reset Loc0, Rec0 in lm_auto */
void set_markdat    (int, int);

/* transferred from Quant: EAT June 2002                     */
double comput_lods (double **, double *, int, int, double *);

double lik_eigen(double **, double *, int, double *);

/** @} */

#endif
