/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/**
 * @file   globs.h
 *
 * @brief header file defining global variables.
 *
 *   -  global in scope
 *   -  included by main programs and function libraries
 *   -  contains modelling or simulation parameters
 *      derived from the input statements
 *   -  values are stored by Pars/ functions
 *
 * $Id: parseprog_opts.h,v 1.1.2.17 2014/10/01 19:40:17 slewis Exp $
 **/

#ifndef  PARSEPROG_OPTS_H
#define  PARSEPROG_OPTS_H

/**
 * @addtogroup Libraries
 * @{
 */

#define UNIVAR      (psx->pcod == 'u')
#define UNIBIG      (psx->pcod == 'g')
#define BIVAR       (psx->pcod == 'b')
#define MULTIVAR    (psx->pcod == 'm')
#define PEDCHECK    (psx->pcod == 'p')
#define KIN         (psx->pcod == 'k')
#define GENEDROP    (psx->pcod == 'd')
#define IBDDROP     (psx->pcod == 'i')
#define MARKERDROP  (psx->pcod == 'r')
#define LM_MAP      (psx->pcod == 'q')
#define LM_PVAL     (psx->pcod == 'v')
#define LM_IBDTESTS (psx->pcod == 't')
#define LM_AUTO     (psx->pcod == 'a')
#define LM_LODS     (psx->pcod == 'l')
#define LM_BAYES    (psx->pcod == 'y')
#define LM_MARKERS  (psx->pcod == 'e')
#define LM_SCHNELL  (psx->pcod == 's')
#define LM_TWOQTL   (psx->pcod == 'w')
#define IBD_HAPLO   (psx->pcod == 'h')
#define IBD_CREATE  (psx->pcod == 'c')
#define GL_LODS     (psx->pcod == 'n')
#define PHOENIX     (psx->pcod == 'x')

#include "progSimSel.h"

/* To change limits, revise in limdefs.h:
 *     #define MAXCHR     23
 *     #define MAXMK   99999
 *     #define MAXTL     999
 *     #define MAXTR     999
 *     #define MINPED      3
 */

struct _progSimSel ProgSimSel[] = {
/*   number of:       chrom     markers    traits     tloc    test  fixd eff*/
/*                      max    min  max   min  max   min max  pos?   max   */
 {"univar"     , 'u',     -1,  -1,   -1,   1,    1,   1,    1,   0,    5 },
 {"unibig"     , 'g',     -1,  -1,   -1,   1,    1,   1,    1,   0,    5 },
 {"bivar"      , 'b',     -1,  -1,   -1,   2,    2,   2,    2,   0,    5 },
 {"multivar"   , 'm',     -1,  -1,   -1,   1,MAXQT,   1,MAXQT,   0,    5 },
 {"pedcheck"   , 'p',     -1,  -1,   -1,  -1,   -1,  -1,   -1,   0,    0 },
 {"kin"        , 'k',     -1,  -1,   -1,  -1,   -1,  -1,   -1,   0,    0 },
 {"genedrop"   , 'd', MAXCHR,   0,MAXMK,   0,MAXTR,   0,MAXTL,   0,   -1 },
 {"ibddrop"    , 'i', MAXCHR,   1,MAXMK,   0,    0,   1,    1,   0,   -1 },
 {"markerdrop" , 'r', MAXCHR,   1,MAXMK,   0,    1,   1,    1,   0,   -1 },
 {"lm_map"     , 'q', MAXCHR,   1,MAXMK,   0,    0,   0,    0,   0,   -1 },
 {"lm_pval"    , 'v', MAXCHR,   1,MAXMK,   1,    1,   0,    0,   0,   -1 },
 {"lm_ibdtests", 't', MAXCHR,   1,MAXMK,   1,    1,   0,    0,   0,   -1 },
 {"lm_auto"    , 'a', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   0,   -1 },
 {"lm_lods"    , 'l', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   1,   -1 },
 {"lm_bayes"   , 'y', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   1,   -1 },
 {"lm_markers" , 'e', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   1,   -1 },
 {"lm_schnell" , 's', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   1,   -1 },
 {"lm_twoqtl"  , 'w', MAXCHR,   1,MAXMK,   1,    1,   1,    2,   1,   -1 },
 {"ibd_haplo"  , 'h', MAXCHR,   1,MAXMK,  -1,   -1,  -1,   -1,   0,   -1 },
 {"ibd_create" , 'c', MAXCHR,   1,MAXMK,  -1,   -1,  -1,   -1,   0,   -1 },
 {"gl_lods"    , 'n', MAXCHR,   1,MAXMK,   1,    1,   1,    1,   1,   -1 },
 {"phoenix"    , 'x', MAXCHR,   0,MAXMK,   0,MAXTR,   0,MAXTL,   1,    1 },
 {""           ,'\0',      0,   0,    0,   0,    0,   0,    0,   0,    0 }};

/**  For each program, limits for the data.                                   */

struct _progDat  {
  char pcod;                       /* one character program code              */
  int  max_mk_all;                 /* maximum number of marker alleles        */
  int  min_ped;                    /* minimum pedigree size                   */
  int  max_ped;                    /* maximum pedigree size                   */
  int  max_obs;                    /* maximum number of observed individuals  */
  int  max_cmp;                    /* maximum component size                  */
  int  max_n_cmp;                  /* maximum number of components            */
               } *pmx;

struct _progDat ProgDat[] = {
/*  Maximum number of:                                                        */
/*       marker   minimum   maximum   number    comp   number                 */
/*      alleles  ped size  ped size   of obs    size  of comps                */
 {'u',     -1,    MINPED,      400,     400,      0,    200 },
 {'g',     -1,    MINPED,    20000,   20000,    400,    200 },
 {'b',     -1,    MINPED,    20000,   20000,    400,    200 },
 {'m',     -1,    MINPED,    20000,   20000,    400,    200 },
 {'p',     -1,    MINPED,    20000,   20000,  20000,   6666 },
 {'k',     -1,    MINPED,    20000,   20000,  20000,   6666 },
 {'d',    128,    MINPED,    20000,   20000,  20000,   6666 },
 {'i',     -1,    MINPED,    20000,      -1,    400,    200 },
 {'r',    128,    MINPED,    20000,      -1,    400,    200 },
 {'q',    128,    MINPED,    20000,      -1,    400,    200 },
 {'v',    128,    MINPED,    20000,      -1,    400,    200 },
 {'t',    128,    MINPED,    20000,      -1,    400,    200 },
 {'a',    128,    MINPED,    20000,      -1,    400,   2000 },
 {'l',    128,    MINPED,    20000,      -1,    400,    200 },
 {'y',    128,    MINPED,    20000,      -1,    400,    200 },
 {'e',    128,    MINPED,    20000,      -1,    400,    200 },
 {'s',    128,    MINPED,    20000,      -1,    400,    200 },
 {'w',    128,    MINPED,    20000,      -1,    400,    200 },
 {'h',    128,         0,        0,      -1,    400,    200 },
 {'c',    128,         0,        0,      -1,    400,    200 },
 {'n',    128,    MINPED,    20000,      -1,    400,    200 },
 {'x',    128,         0,    20000,      -1,    400,    200 },
 {'\0',     0,         0,        0,       0,      0,      0 }};

struct _progOpt1{
  char pcod;                       /* one character program code              */
  int sel_mark;                    /* marker selection                        */
  int sel_tloc;                    /* tloc selection                          */
  int sel_inher;                   /* inheritance selection                   */
  int sel_trait;                   /* trait selection                         */
               } *pox1;

struct _progOpt1 ProgOpt1[] = {
/* pcod  SEL_MARK SEL_TLOC SEL_INHER SEL_TRAIT  */
 {'u',         0,       0,        0,       1}, /* univar      */
 {'g',         0,       0,        0,       1}, /* unibig      */
 {'b',         0,       0,        0,       1}, /* bivar       */
 {'m',         0,       0,        0,       1}, /* multivar    */
 {'p',         0,       0,        0,       0}, /* pedcheck    */
 {'k',         0,       0,        0,       0}, /* kin         */
 {'d',         1,       0,        0,       1}, /* genedrop    */
 {'i',         1,       1,        0,       0}, /* ibddrop     */
 {'r',         1,       0,        1,       1}, /* markerdrop  */
 {'q',         1,       0,        0,       0}, /* lm_map      */
 {'v',         1,       0,        0,       1}, /* lm_pval     */
 {'t',         1,       0,        0,       1}, /* lm_ibdtests */
 {'a',         1,       0,        0,       1}, /* lm_auto     */
 {'l',         1,       0,        0,       1}, /* lm_lods     */
 {'y',         1,       0,        0,       1}, /* lm_bayes    */
 {'e',         1,       0,        0,       1}, /* lm_markers  */
 {'s',         1,       0,        0,       1}, /* lm_schnell  */
 {'w',         1,       0,        0,       1}, /* lm_twoqtl   */
 {'h',         1,       0,        0,       0}, /* ibd_haplo   */
 {'c',         1,       0,        0,       0}, /* ibd_create  */
 {'n',         0,       0,        0,       1}, /* gl_lods     */
 {'x',         1,       1,        1,       1}, /* phoenix     */
 {'\0',        0,       0,        0,       0}};

#define SEL_MARK        pox1->sel_mark
#define SEL_TLOC        pox1->sel_tloc
#define SEL_INHER       pox1->sel_inher
#define SEL_TRAIT       pox1->sel_trait

struct _progOpt2{
  char pcod;                       /* one character program code              */
  int mark_map;                    /* marker maps                             */
  int mark_allele;                 /* marker allele frequencies               */
  int mark_min_dist;               /* minimum dist between markers            */
  int mark_dat;                    /* marker data                             */
  int mark_nam;                    /* marker name                             */
               } *pox2;

struct _progOpt2 ProgOpt2[] = {
/* pcod  MARK_MAP MARK_ALLELE MARK_MIN_DIST MARK_DATA MARK_NAME */
 {'u',         0,          0,            0,        0,        0}, /* univar      */
 {'g',         0,          0,            0,        0,        0}, /* unibig      */
 {'b',         0,          0,            0,        0,        0}, /* bivar       */
 {'m',         0,          0,            0,        0,        0}, /* multivar    */
 {'p',         0,          0,            0,        0,        0}, /* pedcheck    */
 {'k',         0,          0,            0,        0,        0}, /* kin         */
 {'d',         1,          1,            0,        0,        0}, /* genedrop    */
 {'i',         1,          0,            0,        0,        1}, /* ibddrop     */
 {'r',         1,          1,            0,        1,        1}, /* markerdrop  */
 {'q',         1,          1,            1,        1,        1}, /* lm_map      */
 {'v',         1,          1,            1,        1,        1}, /* lm_pval     */
 {'t',         1,          1,            1,        1,        1}, /* lm_ibdtests */
 {'a',         1,          1,            1,        1,        1}, /* lm_auto     */
 {'l',         1,          1,            1,        1,        1}, /* lm_lods     */
 {'y',         1,          1,            1,        1,        1}, /* lm_bayes    */
 {'e',         1,          1,            1,        1,        1}, /* lm_markers  */
 {'s',         1,          1,            1,        1,        1}, /* lm_schnell  */
 {'w',         1,          1,            1,        1,        1}, /* lm_twoqtl   */
 {'h',         1,          1,            1,        1,        1}, /* ibd_haplo   */
 {'c',         1,          0,            1,        1,        1}, /* ibd_create  */
 {'n',         0,          0,            0,        0,        1}, /* gl_lods     */
 {'x',         1,          1,            1,        1,        0}, /* phoenix     */
 {'\0',        0,          0,            0,        0,        0}};

#define MARK_MAP        pox2->mark_map
#define MARK_ALLELE     pox2->mark_allele
#define MARK_MIN_DIST   pox2->mark_min_dist
#define MARK_DATA       pox2->mark_dat
#define MARK_NAME       pox2->mark_nam

struct _progOpt3{
  char pcod;                       /* one character program code              */
  int tloc_map;                    /* tloc maps                               */
  int tloc_vmap;                   /* tloc test positions                     */
  int tloc_allele;                 /* tloc allele frequencies                 */
  int inher_tloc;                  /* tlocs for inheritances                  */
  int trait_tloc;                  /* tlocs for traits                        */
  int trait_tloc_geno;             /* genotype means for traits at tlocs      */
               } *pox3;

struct _progOpt3 ProgOpt3[] = {
/* pcod  TLOC_MAP  TLOC_VMAP TLOC_ALLELE INHER_TLOC TRAIT_TLOC TRT_TLC_GENO  */
 {'u',         0,         0,          0,         0,         0,          0}, /* univar      */
 {'g',         0,         0,          0,         0,         0,          0}, /* unibig      */
 {'b',         0,         0,          0,         0,         0,          0}, /* bivar       */
 {'m',         0,         0,          0,         0,         0,          0}, /* multivar    */
 {'p',         0,         0,          0,         0,         0,          0}, /* pedcheck    */
 {'k',         0,         0,          0,         0,         0,          0}, /* kin         */
 {'d',         1,         0,          1,         0,         1,          1}, /* genedrop    */
 {'i',         1,         0,          0,         0,         0,          0}, /* ibddrop     */
 {'r',         1,         0,          1,         1,         1,          0}, /* markerdrop  */
 {'q',         0,         0,          0,         0,         0,          0}, /* lm_map      */
 {'v',         0,         0,          0,         0,         0,          0}, /* lm_pval     */
 {'t',         0,         0,          0,         0,         0,          0}, /* lm_ibdtests */
 {'a',         1,         0,          1,         0,         1,          0}, /* lm_auto     */
 {'l',         0,         1,          1,         0,         1,          0}, /* lm_lods     */
 {'y',         0,         1,          1,         0,         1,          0}, /* lm_bayes    */
 {'e',         0,         1,          1,         0,         1,          1}, /* lm_markers  */
 {'s',         0,         1,          1,         0,         1,          1}, /* lm_schnell  */
 {'w',         0,         2,          1,         0,         1,          1}, /* lm_twoqtl   */
 {'h',         0,         0,          0,         0,         0,          0}, /* ibd_haplo   */
 {'c',         0,         0,          0,         0,         0,          0}, /* ibd_create  */
 {'n',         0,         0,          1,         0,         1,          1}, /* gl_lods     */
 {'x',         0,         0,          0,         0,         0,          0}, /* phoenix     */
 {'\0',        0,         0,          0,         0,         0,          0}};

#define TLOC_MAP        pox3->tloc_map
#define TLOC_VMAP       pox3->tloc_vmap
#define TLOC_ALLELE     pox3->tloc_allele
#define INHER_TLOC      pox3->inher_tloc
#define TRAIT_TLOC      pox3->trait_tloc
#define TRT_TLC_GENO    pox3->trait_tloc_geno

struct _progOpt4{
  char pcod;                       /* one character program code              */
  int inher_dat;                   /* inheritance data (in ped file)          */
  int trait_data;                  /* trait data                              */
  int trait_vari;                  /* trait additive and residual variances   */
  int penetrances;                 /* trait penetrances                       */
  int do_fix_eff;                  /* program works with fixed effects        */
               } *pox4;

struct _progOpt4 ProgOpt4[] = {
/* pcod  INHER_DATA TRAIT_DATA TRAIT_VARI TRAIT_PENE DO_FIX_EFF */
 {'u',           0,         0,         0,         0,        1}, /* univar      */
 {'g',           0,         0,         0,         0,        1}, /* unibig      */
 {'b',           0,         0,         0,         0,        1}, /* bivar       */
 {'m',           0,         0,         0,         0,        1}, /* multivar    */
 {'p',           0,         0,         0,         0,        0}, /* pedcheck    */
 {'k',           0,         0,         0,         0,        0}, /* kin         */
 {'d',           0,         1,         1,         0,        0}, /* genedrop    */
 {'i',           0,         0,         0,         0,        0}, /* ibddrop     */
 {'r',           1,         1,         0,         1,        0}, /* markerdrop  */
 {'q',           0,         0,         0,         0,        0}, /* lm_map      */
 {'v',           0,         1,         0,         0,        0}, /* lm_pval     */
 {'t',           0,         1,         0,         0,        0}, /* lm_ibdtests */
 {'a',           0,         1,         0,         1,        0}, /* lm_auto     */
 {'l',           0,         1,         0,         1,        0}, /* lm_lods     */
 {'y',           0,         1,         0,         1,        0}, /* lm_bayes    */
 {'e',           0,         1,         1,         1,        0}, /* lm_markers  */
 {'s',           0,         1,         1,         0,        0}, /* lm_schnell  */
 {'w',           0,         1,         1,         0,        0}, /* lm_twoqtl   */
 {'h',           0,         0,         0,         0,        0}, /* ibd_haplo   */
 {'c',           0,         0,         0,         0,        0}, /* ibd_create  */
 {'n',           0,         1,         1,         1,        0}, /* gl_lods     */
 {'x',           0,         0,         0,         0,        0}, /* phoenix     */
 {'\0',          0,         0,         0,         0,        0}};

#define INHER_DATA        pox4->inher_data
#define TRAIT_DATA        pox4->trait_data
#define TRAIT_VARI        pox4->trait_vari
#define TRAIT_PENE        pox4->penetrances
#define DO_FIX_EFF        pox4->do_fix_eff

struct _progOpt5{
  char pcod;                       /* one character program code              */
  int proband_gam;                 /* proband gametes                         */
  int prob_gam_loc;                /* proband gamete scoring at individ loci  */
  int prob_gam_window;             /* p. g. scoring jointly in moving window  */
  int prob_gam_pattern;            /* p. g. scoring pattern in moving window  */
               } *pox5;

struct _progOpt5 ProgOpt5[] = {
/* pcod  PROBAND_GAM PROB_GAM_LOC PROB_GAM_WNDW PROB_GAM_PTRN */
 {'u',            0,           0,            0,            0}, /* univar      */
 {'g',            0,           0,            0,            0}, /* unibig      */
 {'b',            0,           0,            0,            0}, /* bivar       */
 {'m',            0,           0,            0,            0}, /* multivar    */
 {'p',            0,           0,            0,            0}, /* pedcheck    */
 {'k',            0,           0,            0,            0}, /* kin         */
 {'d',            0,           0,            0,            0}, /* genedrop    */
 {'i',            1,           1,            1,            0}, /* ibddrop     */
 {'r',            0,           0,            0,            0}, /* markerdrop  */
 {'q',            0,           0,            0,            0}, /* lm_map      */
 {'v',            0,           0,            0,            0}, /* lm_pval     */
 {'t',            0,           0,            0,            0}, /* lm_ibdtests */
 {'a',            1,           0,            0,            1}, /* lm_auto     */
 {'l',            0,           0,            0,            0}, /* lm_lods     */
 {'y',            0,           0,            0,            0}, /* lm_bayes    */
 {'e',            0,           0,            0,            0}, /* lm_markers  */
 {'s',            0,           0,            0,            0}, /* lm_schnell  */
 {'w',            0,           0,            0,            0}, /* lm_twoqtl   */
 {'h',            1,           0,            0,            0}, /* ibd_haplo   */
 {'c',            1,           0,            0,            0}, /* ibd_create  */
 {'n',            0,           0,            0,            0}, /* gl_lods     */
 {'x',            0,           0,            0,            0}, /* phoenix     */
 {'\0',           0,           0,            0,            0}};

#define PROBAND_GAM       pox5->proband_gam
#define PROB_GAM_LOC      pox5->prob_gam_loc
#define PROB_GAM_WNDW     pox5->prob_gam_window
#define PROB_GAM_PTRN     pox5->prob_gam_pattern

/** @} */

#endif  /*PARSEPROG_OPTS_H*/
