/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#ifndef TWOQTL_H
#define TWOQTL_H

#include <assert.h>
#include "markers.h"
#include "rans.h"
#include "quant.h"
#include "peelie.h"

/* Parameters for S_trait in each marker interval.  Note: for the left end,   */
/* index and recomb are reversed in order (reflecting the order of sampling). */

struct S_parm
{
   int  *index;                         /* Trait positions in S_trait.        */
   int   end;                           /* 1 for outside, 0 for inside.       */
   double *recomb;                      /* Recombinations between trait       */
                                        /* positions.                         */
   double ****prob;                     /* P(S_i, S_(i+1) | S_left = 0 or 1). */
};
typedef struct S_parm S_parm_t;

/* Batch means at each position under the alternative and the null.  Instead  */
/* of storing the likelihood, store foo and max for stable computation.       */

struct Eval {
      double **foo_alt;                 /* mean (exp (log_likelihood - max))  */
      double **max_alt;                 /* max (log_likelihood)               */

      double foo_null;
      double max_null;

      int n_loci_1;
      int n_loci_2;
};
typedef struct Eval mc_eval_t;

#ifdef  DEFINE_DATA
GLOBAL double sim_freq_1 = -1.0;        /* Sim. allele freq. of A in QTL 1.   */
GLOBAL double sim_freq_2 = -1.0;        /* Sim. allele freq. of A in QTL 2.   */
GLOBAL double *geno_means_1 = NULL;     /* Genotypic means AA, Aa, aa QTL 1.  */
GLOBAL double *geno_means_2 = NULL;     /* Genotypic means AA, Aa, aa QTL 2.  */
GLOBAL double *q_trait = NULL;          /* The quantitative trait value from  */
/*                                         the pedigree file for each of the  */
/*                                         observed individuals.              */
GLOBAL int mc_eval_length = UNDEF;      /* Number of elements allocated for   */
/*                                         the MC_eval array.                 */
GLOBAL int n_loci_1 = UNDEF;            /* Total number of marker and test    */
/*                                         tloc positions for QTL 1.          */
GLOBAL int n_loci_2 = UNDEF;            /* Total number of marker and test    */
/*                                         tloc positions for QTL 2.          */
GLOBAL int *ibase = NULL;               /* Index to first individual within   */
/*                                         each component.                    */
GLOBAL int *mbase = NULL;               /* Index to first meiosis within      */
/*                                         each component.                    */
GLOBAL int *n_meiosis = NULL;           /* Number of meioses within each      */
/*                                         component.                         */
GLOBAL int *nFGL_c = NULL;              /* Number of FGL's within each        */
/*                                         component.                         */
GLOBAL int *size_c = NULL;              /* Number of individuals within each  */
/*                                         component.                         */
GLOBAL int *obs_size_c = NULL;          /* Number of observed individuals     */
/*                                         within each component.             */
GLOBAL int *obs_ibase = NULL;           /* Index to first observed individual */
/*                                         within each component.             */
GLOBAL int **obs_index = NULL;          /* Index to each observed individual  */
/*                                         within each component.             */
GLOBAL int *mark_index = NULL;          /* Index to markers within S_trait.   */
GLOBAL int gxg_interaction = FALSE;     /* MC(1)/exact(0) interaction.        */
#else
GLOBAL double sim_freq_1;
GLOBAL double sim_freq_2;
GLOBAL double *geno_means_1;
GLOBAL double *geno_means_2;
GLOBAL double *q_trait;
GLOBAL int mc_eval_length;
GLOBAL int n_loci_1;
GLOBAL int n_loci_2;
GLOBAL int *ibase;
GLOBAL int *mbase;
GLOBAL int *n_meiosis;
GLOBAL int *nFGL_c;
GLOBAL int *size_c;
GLOBAL int *obs_size_c;
GLOBAL int *obs_ibase;
GLOBAL int **obs_index;
GLOBAL int *mark_index;
GLOBAL int gxg_interaction;
#endif  /*DEFINE_DATA*/

/* Subroutines currently in the update_S_trait.c file.                        */
/* The debugging printing functions are declared below along with all the     */
/* other debugging printing functions.                                        */

void update_S_trait (int, int, char ***, char ***);
void fill_S_markers (char **, char **, int);
void update_S_out (char **, char **, int, int);
void update_S_in (char **, char **, int, int);
void setup_S_parm (void);
void setup_prob (double *****, int, double *);
char ***alloc_S_trait (int, int *);
void free_S_trait (char ***);
void free_update_S_trait (void);

/* Subroutines currently in the mc_twoqtl.c file.                             */
/* The debugging printing functions are declared below along with all the     */
/* other debugging printing functions.                                        */

void log_prob_alt (double *, int, char ***, char ***, int, int, int, mc_eval_t *);
void log_prob_null (double *, int, int, int, int, mc_eval_t *);
int *setup_map (char **, char **, int, int *);
char **pack_S (char **, int *, int, int);
int **set_FGL_c (char *, char *, int);
int *obs_FGL (int *, int);
char **update_S_null (int);

/* Subroutines currently in the evaluate.c file.                              */

double evaluate (int **, int **, int);
double enumerate (int **, int **, int *, int *, int, int, double *);
double simulate (int **, int **, int *, int *, int, int, double *);
double log_prob_jnt (double *, int *, int *, int, int, int **, int **,
                     unsigned long int, unsigned long int);
char *next_allele (int, unsigned long int, int *);
char *random_allele (int, double, int *);
void adj_qtl_effect (double *, int **, int *, char *, int);
void adj_qtl_joint (double *, int **, int **, int *, int *, char *, char *);
double log_prob_allele (int, int, double);
int *set_obs_FGL_idx (int **, int *);
void set_evaluate (void);
void free_evaluate (void);

/* Subroutines currently in the batch_means.c file.                           */

void write_lod_score (mc_eval_t *);
double mean_batch (mc_eval_t *, int, int, int);
double var_batch (mc_eval_t *, int, int, int, double);
void print_top (int, int);
void print_null (double, double);
void print_lodsd (double **, double **);
void write_batch_means (mc_eval_t *, int);
void update_batch (double, int, double *, double *);
mc_eval_t *alloc_mc_eval (int);
mc_eval_t setup_mc_eval (void);
void free_mc_eval (mc_eval_t *);

/* Printing functions for intermediate results, if DebugLevel > 0.            */

void print_S_parm (int);
void print_single_marker (int, int, char **, char **);
void print_S_trait (char **, char **, int, int);
void print_step_1 (int *, int, char **, char **, char **, char **, int);
void print_step_2 (int **, int **, int, int);
void print_step_2_null (char **, int **, int *, int *, int, int);
void print_step_3 (double **, int);
void print_S (char **, char **, int, int);
void print_S_marker (int);
void print_mc_eval (mc_eval_t);

/* Functions for getting variables for componentwise computation.             */

int *get_ibase (void);
int *get_obs_ibase (void);
int **get_obs_index (void);
int *get_size_c (void);
int *get_obs_size_c (void);
int *construct_size_c (int *, int);
int *get_mbase (int);
int *get_n_meiosis (int);
int *get_nFGL_c (void);
int nghd_index (int, int);
int comp_index (int, int);
void free_component (void);

/* Subroutines currently located in the tq_gauss_quant.c file.  Some of these */
/* subroutines really should be moved to the Peel library.                    */

void peel_poly (double *, double *);
void sample_poly (double *, double *);
void sample_poly_comp (double *, double *, int);
double tq_gauss_quant (double *, int, double, double, int);
void realloc_BMatrix (int);
void free_BMatrix (void);
void print_BMatrix (int);
struct sp_element *tq_calc_nrm (int, int);
void tq_gauss_setup (int);
void cmp_gauss_setup (int, int);
double envonly_comp (double *, int);
void free_tq_gauss (void);

/* This is a separate subroutine created for testing tq_gauss_quant.          */

void test_tq_gauss_quant (double *, double *);

/* Another rans, using a different multiplier.                                */

double rans_alt (void);
int intran_alt (int, int);
void setsd_alt (uint32, uint32);
void getsd_alt (uint32 *, uint32 *);
int getranscount_alt (void);

#ifndef __GNUC__
void die(const char *format, ...);
#else
void die(const char *format, ...) __attribute__ ((__noreturn__));
#endif /* __GNUC__ */

#endif  /* TWOQTL_H */
