/******************************************************************************/
/*                                                                            */
/* Subroutines used to help debug any of the ibd_create.c subprograms.        */
/*                                                                            */
/******************************************************************************/

#include "ibd_create.h"

/******************************************************************************/
/*                                                                            */
/* The testrng subroutine was written by Chris Glazner for testing the random */
/* number generator(s) he wrote inside beaglesim.                             */
/*                                                                            */
/******************************************************************************/

void testrng (void)
{
   rngstate *st1;

   unsigned int k;
   int   i;

   st1 = initstream (1);

   /* Note: these values are taken from the Marsaglia post. The 1,000,257th   */
   /* call to kiss() from starting with this state should give 1372460312.    */
   /* Another 1,000,001 calls should produce the listed subgenerator values.  */

   st1->z = 12345;
   st1->w = 65435;
   st1->jsr = 34221;
   st1->jcong = 12345;

   for (i = 1; i < 1000001 + 256; ++i)
      k = kiss (st1);

   printf (" RNG test: These should be zero: %d ", k - 1372460312U);
   for (i = 1; i < 1000001; ++i)
      kiss (st1);

   printf ("%d %d %d\n", st1->jcong - 1529210297U, st1->jsr - 2642725982U,
           (st1->z << 16) + (st1->w) - 904977562U);
   free (st1);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The progbar (ie, progress bar) subroutine was written by Chris Glazner for */
/* generating a progress bar from within beagelsim.                           */
/*                                                                            */
/******************************************************************************/

void progbar (int percent)
{
   int   prog;
   int   i;

   /* If beaglesim is running as part of the IBD_create gold standards, the   */
   /* progress bar should not be produced because if it is included in the    */
   /* console output file it really messes up the diff command check used in  */
   /* the Makefile.                                     SML - June 27, 2014.  */

   if (DebugLevel == 4096)
      return;

   prog = ceil (percent * 50 / (double) 100);

   printf ("\r [");
   for (i = 0; i < prog; ++i)
      printf ("-");

   for (; i < 50; ++i)
      printf (" ");

   printf ("] %3.d%%", percent);
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The disp_dag_file subroutine can be called to output to stdout a printout  */
/* of the entire DAG file, with rows in reverse order within markers.         */
/*                                                                            */
/******************************************************************************/

void disp_dag_file (marker * mhead)
{
   marker *mcurr;
   edge *ecurr;

   mcurr = mhead;

   while (mcurr != NULL)
   {
      ecurr = mcurr->edges;

      while (ecurr != NULL)
      {
         printf ("%d %d %d %c %d %d\n", mcurr->index, ecurr->parent,
                 ecurr->child, ecurr->allele, ecurr->count, ecurr->nsamechild);

         ecurr = ecurr->nextedge;
      }

      mcurr = mcurr->nextmarker;
   }

   return;
}
