/******************************************************************************/
/*                                                                            */
/* Program: ibd_create - Front end program for calling the IBD_Create library */
/*                       subprograms.                                         */
/*                                                                            */
/* Written by Steven Lewis.                                                   */
/*                                                                            */
/******************************************************************************/

#define PNAM "ibd_create"
#define DEFINE_DATA

#include "markers.h"
#include "ibd_haplo_directory.h"
#include "ibd_create.h"

/******************************************************************************/
/*                                                                            */
/* The print_help subroutine.                                                 */
/*                                                                            */
/******************************************************************************/

void print_help (void)
{
   printf ("\n Usage:  ./ibd_create  <choice of subprogram flag> ");
   printf (" <morgan parameter statements file name>\n\n");

   printf (" Subprogram name  flag\n");
   printf (" ---------------  ----\n\n");

   printf ("  beaglesim        -s\n");
   printf ("  beagledag        -d\n");
   printf ("  simpop_fgl       -p\n");
   printf ("  fgl2haplo        -f\n");
   printf ("  fgl2ibd          -i\n\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The arg_count_error subroutine.                                            */
/*                                                                            */
/******************************************************************************/

void arg_count_error (char flag)
{
   printf ("\n");
   printf ("  ERROR: Incorrect number of arguments for the -%c flag.\n", flag);
   printf ("  Use the --help or -h flag to see usage.\n\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The unrecognized_error subroutine.                                         */
/*                                                                            */
/******************************************************************************/

void unrecognized_error (char *bad_arg)
{
   printf ("\n");
   printf ("  ERROR: Unrecognizable subprogram choice flag: %s.\n", bad_arg);
   printf ("  Use the --help or -h flag to see usage.\n\n");
   return;
}

/******************************************************************************/
/*                                                                            */
/* The main subroutine is here.                                               */
/*                                                                            */
/******************************************************************************/

int main (int argc, char **argv)
{
   char  help1[] = "--help";
   char  help2[] = "-h";

   char **morg_args = NULL;             /* Temporary copy of Morgan args.     */

   int   num_args;                      /* Number of arguments still passed.  */
   int   from_ndx;                      /* From index.                        */
   int   to_ndx;                        /* To index.                          */
   int   rc;                            /* Return code for the main routine.  */

   /***************************************************************************/
   /*                                                                         */
   /* The executable portion of the main program starts here.                 */
   /*                                                                         */
   /***************************************************************************/

   if (argc == 1 || strcmp (argv[1], help1) == 0 ||
       strcmp (argv[1], help2) == 0)
   {
      print_help ();
      rc = 0;
      exit (rc);
   }

   /* Parse the command line options.                                         */
   if ((argc > 1) && (strlen (argv[1]) == 2) && (argv[1][0] == '-'))
   {
      SubProgChoice = argv[1][1];       /* Get the subprogram choice flag.    */

      if (argc < 3)
      {
         arg_count_error (SubProgChoice);
         rc = -2;
         exit (rc);
      }

      /* Prepare a local copy of the command line argument list leaving out   */
      /* the subprogram choice flag.                                          */

      num_args = argc - 1;
      morg_args = malloc (num_args * sizeof (char *));
      morg_args[0] = argv[0];           /* Use the ibd_create program name.   */

      for (from_ndx = 2; from_ndx < argc; ++from_ndx)
      {
         to_ndx = from_ndx - 1;
         morg_args[to_ndx] = argv[from_ndx];
      }

      switch (SubProgChoice)
      {
          case 'h':
             print_help ();
             rc = 0;
             break;

          case 's':
             rc = beaglesim (num_args, morg_args);
             break;

          case 'd':
             rc = beagledag (num_args, morg_args);
             break;

          case 'p':
             rc = simpop_fgl (num_args, morg_args);
             break;

          case 'f':
             rc = fgl2haplo (num_args, morg_args);
             break;

          case 'i':
             rc = fgl2ibd (num_args, morg_args);
             break;

          default:
             {
                unrecognized_error (argv[1]);
                rc = -1;
             }
             break;
      }
   }

   else
   {
      unrecognized_error (argv[1]);
      rc = -1;
   }

   if (morg_args != NULL)
      free (morg_args);

   exit (rc);
}
