void swap (int *x, int *y)
{
   (*x) += (*y);
   (*y) = (*x) - (*y);
   (*x) -= (*y);
}

int power (int base, int exponent)
{
   if (exponent == 0)
      return 1;
   if (exponent < 0)
      return 0;
   int   result = 1;
   int   i;
   for (i = 0; i < exponent; i++)
   {
      result *= base;
   }
   return result;
}

int stateEncode (int num_haps, int *state)
{
   int   i;
   int   encoding = 0;
   for (i = 0; i < num_haps; i++)
   {
      encoding += state[i] * power (num_haps + 1, i);
   }
   return encoding;
}

void makeOrbits (int num_haps, int ms, int **statemat)
{
   int  *encodedStates, *visited;
   int   i, k, norb, nindivs;
   unsigned j;
   int   nonSingletonOrbit;
   int  *tmpState, *tmpState2;
   int   tmp;

   tmpState = alloc_int1d (num_haps);
   tmpState2 = alloc_int1d (num_haps);

   nindivs = num_haps / 2;
   norb = 0;

   encodedStates = malloc (ms * sizeof (int));
   visited = malloc (ms * sizeof (int));

   for (i = 0; i < ms; i++)
   {
      orbitSizes[i] = 0;
   }
   for (i = 0; i < ms; i++)
   {
      encodedStates[i] = stateEncode (num_haps, statemat[i]);
      visited[i] = 0;
   }
   for (i = 0; i < ms; i++)
   {
      if (visited[i])
         continue;
      visited[i] = 1;
      nonSingletonOrbit = 0;
      for (j = 1; j < power (2, nindivs); j++)
      {
         for (k = 0; k < num_haps; k++)
         {
            tmpState[k] = statemat[i][k];
         }
         for (k = 0; k < nindivs; k++)
         {
            if (IS_SET (j, k))
               swap (tmpState + k * 2, tmpState + k * 2 + 1);
         }
         rplc (num_haps, tmpState, tmpState2);
         tmp = stateEncode (num_haps, tmpState2);
         for (k = i + 1; k < ms; k++)
         {
            if (visited[k])
               continue;
            if (tmp == encodedStates[k])
            {
               visited[k] = 1;
               if (nonSingletonOrbit)
               {
                  orbits[norb - 1][orbitSizes[norb - 1]] = k;
                  orbitSizes[norb - 1]++;
               }
               else
               {
                  norb++;
                  orbits[norb - 1][0] = i;
                  orbits[norb - 1][1] = k;
                  orbitSizes[norb - 1] = 2;
                  nonSingletonOrbit = 1;
               }
            }
         }
      }
   }

   free_1d ((void *) &tmpState);
   free_1d ((void *) &tmpState2);

   free (encodedStates);
   free (visited);

   num_orbits = norb;
   return;
}

/******************************************************************************/
/*                                                                            */
/* Since the num_orbits variable set by makeOrbits is just the number of      */
/* orbits with more than one state, the compute_number_orbits subroutine is   */
/* called to compute the total number of orbits, including singleton orbits.  */
/*                                                                            */
/******************************************************************************/

int compute_number_orbits (void)
{
   int   states_in_non_singletons;      /* Number of states accounted for in  */
   /*                                      the non-singleton orbits.          */
   int   num_all_orbits;                /* Total number of orbits (including  */
   /*                                      singleton orbits.                  */

   int   os_ndx;                        /* Subscript into orbitSizes[].  In   */
   /*                                      other words, os_ndx is the current */
   /*                                      non-singleton orbit subscript.     */

   /* Find the total number of orbits.  This is done by computing the number  */
   /* of singleton orbits and adding this number to num_orbits, which was set */
   /* by the makeOrbits subroutine to the number of non-singleton orbits.     */

   states_in_non_singletons = 0;

   for (os_ndx = 0; os_ndx < num_orbits; ++os_ndx)
      states_in_non_singletons += orbitSizes[os_ndx];

   num_all_orbits = num_states - states_in_non_singletons + num_orbits;

   return num_all_orbits;
}

/******************************************************************************/
/*                                                                            */
/* For a genotypic analysis the collapse_to_distinct subroutine computes the  */
/* conditional probabilities of gene IBD states totaling the probabilities of */
/* IBD over the genotypically equivalent states.  Before collapse_to_distinct */
/* is called the makeOrbits subroutine must be called.  The results produced  */
/* by makeOrbits, setting the orbitSizes and orbits variables, are used by    */
/* collapse_to_distinct to compute the total probability of gene IBD for each */
/* orbit.                                                                     */
/*                                                                            */
/* The collapse_to_distinct subroutine also constructs the gene IBD state     */
/* identifications.  These are returned to the caller in a character array to */
/* be pointed to by the p_state_names argument.  These state labels help to   */
/* identify which haplotypic IBD states were added together to produce the    */
/* genotypic IBD state probabilities.                                         */
/*                                                                            */
/* Note that the num_orbits variable set by makeOrbits is really only the     */
/* number of orbits with more than one state.  It does not include singleton  */
/* orbits.  The num_all_orbits argument needs to have been set to the total   */
/* number of orbits (including singleton orbits).                             */
/*                                                                            */
/******************************************************************************/

void collapse_to_distinct (double **qq_ibd_input, int num_all_orbits,
                           double ***p_qq_ibd_result, int num_haps,
                           char ***p_state_names)
{
   double **qq_ibd_result;              /* Resulting probabilities matrix.    */
   double tot_prob;                     /* Current orbit total probability.   */

   char **state_ids;                    /* Local copy of state_names matrix.  */

   char  id_delimit[] = "+";            /* Delimiter used in state_ids.       */
   char *curr_state_id;                 /* Local copy of haplotypic state_id. */
   char *p_state_ids;                   /* Working pointer to within the      */
   /*                                      state_id being constructed.        */

   int  *visited;                       /* Track whether state has been used. */

   int   os_ndx;                        /* Subscript into orbitSizes[].  In   */
   /*                                      other words, os_ndx is the current */
   /*                                      non-singleton orbit subscript.     */
   int   qq_ndx;                        /* Subscript to current output orbit. */

   int   state_ndx;                     /* Subscript to input IBD gene state. */

   int   m1;                            /* Subscript to current marker.       */
   int   s1;                            /* Subscript to current state within  */
   /*                                      the orbitSizes[] vector.           */
   int   s2;                            /* Subscript to current state within  */
   /*                                      the qq_ibd_input[][] matrix.       */

   state_ids = (char **) malloc (num_all_orbits * sizeof (char *));
   alloc_chk (WHERE, state_ids);

   state_ids[0] = (char *) malloc (num_states * (num_haps + 1) * sizeof (char));
   alloc_chk (WHERE, state_ids[0]);
   p_state_ids = state_ids[0];

   qq_ibd_result = alloc_dbl2d (NMarks, num_all_orbits);
   visited = alloc_int1d (num_states);

   os_ndx = 0;                          /* Initialize orbitSizes subscript.   */
   qq_ndx = 0;                          /* Initialize output orbit subscript. */

   for (state_ndx = 0; state_ndx < num_states; ++state_ndx)
   {
      if (visited[state_ndx] == FALSE)
      {
         if (orbits[os_ndx][0] == state_ndx)
         {
            /* For each non-singleton orbit,                                  */
            for (m1 = 0; m1 < NMarks; ++m1) /* For each marker,               */
            {
               tot_prob = 0.0;
               for (s1 = 0; s1 < orbitSizes[os_ndx]; ++s1)
               {
                  s2 = orbits[os_ndx][s1];
                  tot_prob += qq_ibd_input[m1][s2];
               }
               qq_ibd_result[m1][qq_ndx] = tot_prob;
            }

            state_ids[qq_ndx] = p_state_ids;

            for (s1 = 0; s1 < orbitSizes[os_ndx]; ++s1)
            {
               /* Get the subscript to the next gene IBD state.               */
               s2 = orbits[os_ndx][s1];

               /* Update the visited vector to reflect the fact that each of  */
               /* the states in this orbit has been used by this subroutine.  */
               visited[s2] = TRUE;

               /* Add the current IBD state to the state_ids vector.          */
               curr_state_id = compute_ibd_state (s2, num_haps);
               strcpy (p_state_ids, curr_state_id);
               free_1d ((void *) &curr_state_id);
               p_state_ids += num_haps;

               /* Add a delimiter unless this is the last state in the orbit. */
               if (s1 < orbitSizes[os_ndx] - 1)
                  strcat (p_state_ids, id_delimit);

               /* In either case increment the working state_ids pointer.     */
               ++p_state_ids;
            }

            ++os_ndx;
         }
         else
         {
            /* For each singleton orbit,                                      */
            for (m1 = 0; m1 < NMarks; ++m1) /* For each marker,               */
               qq_ibd_result[m1][qq_ndx] = qq_ibd_input[m1][state_ndx];

            state_ids[qq_ndx] = p_state_ids;
            visited[state_ndx] = TRUE;

            /* Add the current IBD state to the state_ids vector.              */
            curr_state_id = compute_ibd_state (state_ndx, num_haps);
            strcpy (p_state_ids, curr_state_id);
            free_1d ((void *) &curr_state_id);
            p_state_ids += (num_haps + 1);
         }

         ++qq_ndx;
      }
   }

   free_1d ((void *) &visited);

   *p_qq_ibd_result = qq_ibd_result;
   *p_state_names = state_ids;

   return;
}
