/*  vim: sw=3
 *  $Id: print_lods.c,v 1.24.2.6 2013/10/04 19:20:16 slewis Exp $ */

static int nCol1;                       /* file global for formatting
                                         * column headings and data columns */

#include "markers.h"

void print_lods (double ***lkratios)
{
   /* Global variables used */
   extern int n_compnts;
   extern int NTPos;

   double **lods = NULL;                /* Lod scores */
   double **lods_cmb = NULL;            /* Combined Lod Scores */
   double lodsval[4];
   double eigen2 = 0.0;
   int   c, c1, i, m;

   lods = alloc_dbl2d (3, NTPos);
   lods_cmb = alloc_dbl2d (3, NTPos);

   printf (" ESTIMATED LOD SCORES\n\n");

   for (c = 0; c < n_compnts; ++c)
   {
      c1 = c + 1;
      printf ("  Component %d:\n\n", c1);

      if (cmplist[c1].nmeioses == 0)
      {
         printf ("    Component %d has no informative meioses\n\n", c1);
         printf ("\n");
         fflush (stdout);
         continue;
      }

      for (m = 0; m < 3; ++m)
      {
         lodsval[m] = comput_lods (lkratios[c], lods[m], NTPos, m, &eigen2);
         if (m == 0)
            lodsval[3] = eigen2;
      }

      printf ("    The largest eigenvalue        :%9.5f\n\n", lodsval[0]);
      printf ("    The second largest eigenvalue :%9.5f\n\n", lodsval[3]);
      printf ("    Cumulative from left          :%9.5f\n\n", lodsval[1]);
      printf ("    Cumulative from right         :%9.5f\n\n", lodsval[2]);

      printf ("    LOD scores:\n");
      printf ("  %10s %10s %10s %10s %10s %10s\n", "position",
              "female", "male", "eigen", "left", "right");

      for (i = 0; i < NTPos; ++i)
      {
         printf ("  %10d", i);
         printf (" %10.5f %10.5f",
                 map_position (i, FEM), map_position (i, MAL));

         for (m = 0; m < 3; ++m)
         {
            lods_cmb[m][i] += lods[m][i];
            printf (" %10.5f", lods[m][i]);
         }
         printf ("\n");
      }

      printf ("\n\n");
      fflush (stdout);
   }

   free_2d ((void *) &lods);

   if (n_compnts > 1)
   {
      printf ("  Combined All Components:\n");
      printf ("  %10s %10s %10s %10s %10s %10s\n", "position",
              "female", "male", "eigen", "left", "right");

      for (i = 0; i < NTPos; ++i)
      {
         printf ("  %10d", i);
         printf (" %10.5f %10.5f",
                 map_position (i, FEM), map_position (i, MAL));
         for (m = 0; m < 3; ++m)
            printf (" %10.5f", lods_cmb[m][i]);
         printf ("\n");
      }

      printf ("\n");
      fflush (stdout);
   }

   free_2d ((void *) &lods_cmb);

   return;
}

void prntLodSHead (char *title)
{
   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   printf ("\n %s:\n\n", title);
   prntLodSColHd ();

   return;
}

void prntLodSJHead (void)
{
   printf ("\n LodScores estimated jointly over all components:\n\n");

   prntLodSColHd ();

   return;
}

void prntLodSColHd (void)
{
   printf (" %*s     position (Haldane cM)", nCol1, "Trait pos #");

   if (IS_LM_BAYES)
      printf ("%11s %8s %19s", "pseudo", "freq", "LodScore    ");

   printf ("\n");

   printf (" %*s        male     female   ", nCol1, "or marker");

   if (IS_LM_MARKERS || IS_LM_TWOQTL)
      printf (" %11s%10s", "LodScore", "StdErr");

   if (IS_LM_LODS)
      printf (" %10s %10s %10s", "eigen", "left", "right");

   if (IS_LM_BAYES)
      printf ("%11s %8s %9s %9s", "prior", "visited", "crude", "R-B");

   printf ("\n\n");
   fflush (stdout);

   return;
}

/* Prints trait position number or marker name, followed by position along    */
/* the chromosome for each gender.  This subroutine returns true for a trait  */
/* position or false for a marker position.                                   */

int prntLodSPos (int lpx)
{
   static LOCPOS_PTR p_p;
   static int ntp, nmk;
   int   rc;

   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

/* If the argument to this subroutine is a 0, the subroutine is being called  */
/* for the first time, so initialize the local variables.                     */
   if (lpx == 0)
   {
      p_p = LocPos;
      ntp = 0;
      nmk = 0;
   }

   if (p_p->locus_typ == MARK_LOC)
   {

/* If lodscores are being evaluated at the selected markers, as was always    */
/* done traditionally in MORGAN, still print out the marker name and continue */
/* prntLodSPos as it used to always execute.                                  */
      if (Lodscores_at_markers)
      {
         printf (" %*s ", nCol1, MarkNam[nmk]);
         ++nmk;
         rc = 0;
      }
/* But if a "map test tlocs ... at markers ..." statement is in effect, the   */
/* selected markers were not included in the set of lodscore evaluation loci, */
/* so skip to the next locus position which is not a selected marker.         */
      else
      {
         while (p_p->locus_typ == MARK_LOC)
            ++p_p;
      }
   }

   if (p_p->locus_typ == TLOC_LOC || p_p->locus_typ == UNSELMRK_LOC)
   {
      ++ntp;
      printf (" %*d", nCol1, ntp);
      if (p_p->locus_typ == TLOC_LOC)
         printf (" ");
      else if (p_p->locus_typ == UNSELMRK_LOC)
         printf ("*");
      rc = 1;
   }
   printf ("%11.3f%11.3f   ", p_p->pos[MAL], p_p->pos[FEM]);

   ++p_p;

   return rc;
}

void prntLodSPos0 (void)
{
   printf (" %*d", nCol1, 0);
   printf (" %11s%11s   ", "unlinked", "unlinked");

   return;
}

void prntLodScores (int ntp, double *lod, double *sd)
{
   int   p1;

   for (p1 = 1; p1 < ntp; ++p1)
   {
      if ((!prntLodSPos (p1 - 1)) && NAforMarkers)
         /* The current test position is a marker position.  This test was    */
         /* reinstated by EAT in Oct 2005.  It is needed for some genotypic   */
         /* and discrete trait cases.  See the note regarding do_rblik().     */
         printf (" %11s%10s\n", "NA", "NA");
      else
         printf (" %11.4f%10.4f\n", lod[p1], sd[p1]);
   }
   printf ("\n");
   fflush (stdout);

   return;
}
