/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "nghds.h"

/* Read and store individuals file data.                                      */

void setInd (double **qdata)
{
   FILE *indF;

   /* Allocate structures for pedgree data: int data[PedSize][NColData]       */
   /*                                       Ind nghd[PedSize + 1]             */
   alloc_ped ();

   /* Open individuals file for reading.                                      */
   if (!(indF = open_ind ()))
      give_up (ERR_qit);

   /* Read in trait data from individuals file data:  data[][].               */
   input_ped (indF, qdata);

   fclose (indF);

   /* Finish initializing the nghd structure enough so that an input          */
   /* individuals file can be used.                                           */
   partial_nghds ();

   /* Set up lists of members in each component and also information about    */
   /* each component.                                                         */
   if ((n_compnts = components ()) != nCompnt)
      give_up_where (WHERE, ERN_ccp, nCompnt, n_compnts);

   return;
}

#include "pedchk.h"

/******************************************************************************/
/*                                                                            */
/* readInd(): Reads individuals data.  Checks data for valid format.          */
/*            Stores data in PedDat, IntDat and RSDat arrays.                 */
/*                                                                            */
/******************************************************************************/

int readInd (FILE * pF)
{
   PED_PTR pP = pPedDat;
   PEDI_PTR pI = pIntDat;

   char  pedRec[MAXLINLEN + 2];
   char *pStr, *pTok;

   int   namlen;
   int   num;
   int   nchr;
   int   i1;

   widPNam = 0;

   /*                                    * Reads record for each individual.  */
   while (nInd < PedSize && fgets (pedRec, MAXLINLEN + 1, pF) != NULL)
   {
      pStr = pedRec;                    /* Gets individual's name,
                                         * gender (if present) and
                                         * observed (if present).             */
      for (i1 = 0; i1 < NNams; ++i1)
      {
         switch (StdIn[i1])
         {
             case NAM:
                {
                   if (!nextNam (&pStr, &pTok, nInd + 1))
                      return OK;
                   strncpy (pP->nam, pTok, MAXNAMLEN);
                   /* Terminate the string.                                   */
                   pP->nam[MAXNAMLEN] = '\0';
                   /* Figure out the maximum length of the names.             */
                   namlen = strlen (pP->nam);
                   if (widPNam < namlen)
                      widPNam = namlen;
                   break;
                }
             case PA:
                {
                   if (!nextInt (&pStr, &pTok, &num, nInd + 1))
                      return !OK;
                   if (num < 0)
                   {
                      give_err (ERC_fin, pTok, nInd + 1);
                      return !OK;
                   }
                   pP->comp = num;
                   break;
                }
             default:
                break;
         }
         pStr = NULL;                   /* Required for strtok().             */
      }

      for (i1 = NNams; i1 < NStdIn; ++i1)
      {
         if (!nextInt (&pStr, &pTok, &num, nInd + 1))
            return !OK;
         if (num < 0)
         {
            give_err (ERC_fin, pTok, nInd + 1);
            return !OK;
         }
         switch (StdIn[i1])
         {
             case SEX:
                {
                   pP->sex = num;
                   if (pP->sex != MALE && pP->sex != FEMALE
                       && pP->sex != UNSEXED)
                   {
                      give_err (ERC_fsx, pTok, nInd + 1);
                      return !OK;
                   }
                   break;
                }
             case OBS:
                {
                   pP->obsrvd = num;
                   break;
                }
             default:
                break;
         }
         pStr = NULL;                   /* Required for strtok().             */
      }

      /*                                 * Tests for option to assign         */
      /*                                 * or ignore gender.                  */
      if (SexOp == ASSIGN)
         pP->sex = UNSEXED;
      /*                                 * Tests for option to observe        */
      if (ObsOp == ALL)                 /* all individuals.                   */
         pP->obsrvd = OBSERVED;

      /*                                 * Gets the integers.                 */
      for (i1 = 0; i1 < NNInts; ++i1)
      {
         if (!nextInt (&pStr, &pTok, &num, nInd + 1))
            return !OK;

         *(pI + i1) = num;
         if (num > maxPInt[i1])
            maxPInt[i1] = num;
         else if (num < minPInt[i1])
            minPInt[i1] = num;
      }

      /*                                 * Checks the reals                   */
      if (NReals)                       /* and stores the tokens.             */
      {
         *(ppRSDat + nInd) = *ppRSDat + curRSDat;

         for (i1 = 0; i1 < NReals; ++i1)
         {
            if (!nextReal (&pStr, &pTok, nInd + 1))
               return !OK;
            nchr = strlen (pTok);

            while (curRSDat + nchr + 3 >= maxRSDat)
               reallocRSDat ();

            strcpy (*ppRSDat + curRSDat, "  ");
            curRSDat += 2;
            strcpy (*ppRSDat + curRSDat, pTok);
            curRSDat += nchr;
         }
         strcpy (*ppRSDat + curRSDat, "\0");
         ++curRSDat;
      }
      ++pP;
      pI += NNInts;
      ++nInd;
   }

   return OK;
}

/* Opens and reads the individuals file.  Stores the data in the pedchk       */
/* data structures.                                                           */
/* On return, leaves the keytable sorted in chronological order.              */

int chk_indfil (void)
{
   FILE *indF;

   if (!(indF = open_ind ()))           /* Opens the individuals file.        */
      return !OK;

   allocPed ();                         /* Allocates pedchk arrays.           */

   if (!readInd (indF))                 /* Reads the individuals file.        */
      return !OK;

   if (!checkPed (indF))                /* Reports pedigree size.             */
      return !OK;
   fclose (indF);

   allocKey ();                         /* Allocates key table.               */

   storKey ();                          /* Initializes the key table entries. */

   getComp ();                          /* Copy component numbers from PedDat */
   /*                                      to the key table.                  */

   sortNam ();                          /* Sorts key table on name.           */

   if (!checkNam ())                    /* Checks for dupicate names.         */
      return !OK;

   sortIx ();                           /* Sorts key table on record number.  */

   reassignChron ();                    /* Use index order as chronological   */
   /*                                     order and store chron in key table. */

   if (!checkChron ())                  /* Checks for chronological order     */
      return !OK;                       /* found.                             */

   sortComp ();                         /* Sorts key table by component and   */
   /*                                      chronological order.               */

   allocComp ();                        /* Allocates component table.         */

   storComp ();                         /* Stores component table info.       */

   prntComp ();                         /* Prints component information.      */

   reassignChron ();                    /* Sets chronological order by        */
   /*                                      component in key table.            */

   sortChron ();                        /* Sorts by component, then           */
   /*                                      chronological order.               */

   prntStatusCIF ();

   return OK;
}
