/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "markers.h"

/* Routine to determine if an individual is the only child of a parent.       */

int only_kid (Ind * parent, Ind * child)
{
   Marlist *mlpt;
   Indlist *ilpt;

   for (mlpt = parent->marriages; mlpt != NULL; mlpt = mlpt->next)
   {
      for (ilpt = mlpt->current->kids; ilpt != NULL; ilpt = ilpt->next)
         if (strncmp (ilpt->current->self, child->self, MAXNAMLEN))
            return (0);
   }

   return (1);
}

/* generate a list of the 2(pedsize-n_founder) switches (max), so can choose
   a random switch  -- switch is indicated by individual and mat/pat.
   Also, set up a linked list of the switches in a given component for
   later use in sampling and scoring routines -- this bit added, EAT 11/2000
   Also, move some switch variables to msample.h so do not have to be passed
   -- now moved likewise to Markers and markers.h : EAT June 2002    */

/* add component info to each meiosis (Switch)      EAT  10/2004     */
/* Replace these linked lists of meioses by component, by vector
   of meosis indices, in the Component structure    EAT  10/2004     */
/* get rid of *nns argument of make_switches -- info now in Component strct */
/* get rid of setting xx0 and xx1 to -1 for non-meioses -- do elsewhere */
/*   -- hence get rid of argument nl of make_switches    EAT  11/2004   */
/* This meiosis set-up is now moved to Nghds library     EAT 10/2004    */

int make_switches ()
{
   int   i, k, kk, ns;
   Ind  *dad, *mum;
   int  *nns;

   /* The maximum number of switches is: 2 * (PedSize - nFound) */

   meiosis = malloc ((2 * (PedSize - nFound) + 1) * sizeof (Switch));
   alloc_chk (WHERE, meiosis);

   ns = 0;

   for (k = 1; k <= nCompnt; ++k)
      cmplist[k].nmeioses = 0;

   for (i = 0; i < pedsize; ++i)
   {
      if (!isNullNode (nghd[i].pa))
      {
         dad = nghd[i].pa;
         mum = nghd[i].ma;

         if (!(isNullNode (nghd[mum->index].ma) && only_kid (mum, &nghd[i])))
         {
            ++ns;
            meiosis[ns].ind = nghd[i];
            meiosis[ns].sex = 0;
            meiosis[ns].cmpnt = nghd[i].component;
            ++cmplist[nghd[i].component].nmeioses;
         }

         if (!(isNullNode (nghd[dad->index].pa) && only_kid (dad, &nghd[i])))
         {
            ++ns;
            meiosis[ns].ind = nghd[i];
            meiosis[ns].sex = 1;
            meiosis[ns].cmpnt = nghd[i].component;
            ++cmplist[nghd[i].component].nmeioses;
         }
      }
   }

   /* Terminate with a fatal error message if there are no MCMC-scoreable     */
   /* meioses in any component anywhere in the predigree.                     */

   if (ns <= 0)
      give_up (ERR_nsm);

   /* Now set space and initialize the meiosis info in individual structures. */
   /*     added EAT 4/2005 */
   for (i = 1; i <= ns; ++i)
   {
      k = (meiosis[i].ind).index;
      if (meiosis[i].sex == 1)
      {
         nghd[k].patmeio = i;
         dad = nghd[k].pa;
         ++nghd[dad->index].nkidmeio;
      }
      if (meiosis[i].sex == 0)
      {
         nghd[k].matmeio = i;
         mum = nghd[k].ma;
         ++nghd[mum->index].nkidmeio;
      }
   }

   nghd[0].kidmeio = (int *) malloc (ns * sizeof (int));
   alloc_chk (WHERE, nghd[0].kidmeio);

   for (k = 1; k < pedsize; ++k)
      nghd[k].kidmeio = nghd[k - 1].kidmeio + nghd[k - 1].nkidmeio;

   for (k = 1; k < pedsize; ++k)
      if (nghd[k].nkidmeio == 0)
         nghd[k].kidmeio = NULL;

   nns = (int *) malloc (pedsize * sizeof (int));
   alloc_chk (WHERE, nns);

   for (k = 0; k < pedsize; ++k)
      nns[k] = 0;

   for (i = 1; i <= ns; ++i)
   {
      k = (meiosis[i].ind).index;
      if (meiosis[i].sex == 1)
      {
         dad = nghd[k].pa;
         kk = dad->index;
         nghd[kk].kidmeio[nns[kk]] = i;
         ++nns[kk];
      }
      if (meiosis[i].sex == 0)
      {
         mum = nghd[k].ma;
         kk = mum->index;
         nghd[kk].kidmeio[nns[kk]] = i;
         ++nns[kk];
      }
   }

   /* now set up the space for lists of meiosis switches by component */
   /*  note that cmplist and components are indexed from 1 (not 0)    */
   /*  note that meioses (Switches)  are indexed from 1 (not 0)    */

   cmplist[0].meioses = (int *) malloc (ns * sizeof (int));
   alloc_chk (WHERE, cmplist[0].meioses);

   cmplist[1].meioses = cmplist[0].meioses;
   for (k = 2; k <= nCompnt; ++k)
      cmplist[k].meioses = cmplist[k - 1].meioses + cmplist[k - 1].nmeioses;

   for (k = 0; k < pedsize; ++k)        /* actually will now use only
                                           indices 1,...,n_compnts */
      nns[k] = 0;

   for (i = 1; i <= ns; ++i)
   {
      k = meiosis[i].cmpnt;
      cmplist[k].meioses[nns[k]] = i;
      ++nns[k];
   }
   free (nns);

   return (ns);
}

void prnt_switches (int nsw)
{
   int   k;

   if (PrintLevel >= 3)
   {
      printf ("\n The number of meiosis switches to sample is %d\n\n", nsw);
      printf ("    compnt  switches   first    last\n");
      for (k = 1; k <= nCompnt; ++k)
         printf ("    %6d %9d %7d %7d\n",
                 cmplist[k].index, cmplist[k].nmeioses, cmplist[k].meioses[0],
                 cmplist[k].meioses[(cmplist[k].nmeioses) - 1]);
      printf ("\n\n");
      fflush (stdout);
   }

   if (DebugNghds)
      list_switches (nsw);

   return;
}

/* Print the pedigree meiosis information.                                    */
void list_nghds ()
{
   Ind  *mum, *dad;
   int   i1;

   printf (" Listing of pedigree meioses by individual:\n");
   printf ("          name      name.pa      name.ma");
   printf ("  component  pat.meio  mat.meio\n");

   for (i1 = 0; i1 < pedsize; ++i1)
   {
      if (!isNullNode (nghd[i1].pa))
      {
         dad = nghd[i1].pa;
         mum = nghd[i1].ma;
         printf ("  %12s %12s %12s",
                 nghd[i1].self, nghd[dad->index].self, nghd[mum->index].self);
      }
      else
         printf ("  %12s            0            0", nghd[i1].self);
      printf ("    %7d   %7d   %7d\n",
              nghd[i1].component, nghd[i1].patmeio, nghd[i1].matmeio);
   }
   printf ("\n");

   return;
}

/* List the details of all switches.                                          */

void list_switches (int nsw)
{
   int   i1;

   printf (" List of switches:\n");

   for (i1 = 1; i1 <= nsw; ++i1)
   {
      printf (" %d %d %d %s %d\n", i1,
              (meiosis[i1].ind).index, meiosis[i1].cmpnt,
              (meiosis[i1].ind).self, meiosis[i1].sex);
   }
   printf ("\n\n");

   return;
}

/* Free the data structures for the switches.                                 */

void free_switches (void)
{

   if (!meiosis)
      return;

   free (meiosis);                      /* frees (Switch *) meiosis */
   meiosis = NULL;

   return;
}
