/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 *   Autozgy/Lodscore setup functions:
 *     setPed(), setSwitches()
 */

#include "nghds.h"

/**
 * Read and store pedigree data.
 */
void setPed (double **qdata)
{
   int   n;
   FILE *pedF;

   /* allocate structures for pedgree data:  int data[PedSize][NColData]
    *                                        Ind nghd[PedSize + 1]  */
   alloc_ped ();

   /* open for reading */
   if (!(pedF = open_ped ()))
      give_up (ERR_qit);

   /* read in pedigree and trait data:       data[][]  */
   input_ped (pedF, qdata);

   fclose (pedF);

   /* set up neighborhood and pedigree components data structures */
   nghds ();

   /* do an extra check on no. founders */
   if ((n = count_founders ()) != nFound)
      give_up_where (WHERE, ERN_cfd, nFound, n);

   /* set up lists of members in each component and also
      information about each component */
   if ((n_compnts = components ()) != nCompnt)
      give_up_where (WHERE, ERN_ccp, nCompnt, n_compnts);

   /* If the type of the data being analyzed is discrete with liability,      */
   /* check to make sure that liability class penetrance data was provided by */
   /* the user and that user's liability class penetrance data is consistent  */
   /* with the user's input pedigree data.                                    */

   if (PedDatTyp == DISCR_LIAB)
      test_liab_classes ();

   return;
}

/**
 * set and print the switches for each component.
 */
int setSwitches (void)
{
   int   nsw;

   nsw = make_switches ();
   if (nsw)
      prnt_switches (nsw);

   return nsw;
}

/******************************************************************************/
/*                                                                            */
/* The test_liab_classes subroutine checks that there are enough liability    */
/* classes defined to include any value of an individual's liability class    */
/* found in the user's data.                                                  */
/*                                                                            */
/******************************************************************************/

void test_liab_classes (void)
{
   int   ind_ndx;                       /* Current individual subscript.      */

   for (ind_ndx = 0; ind_ndx < pedsize; ++ind_ndx)
   {
      if (data[ind_ndx][5] != 0)
      {
         if (data[ind_ndx][6] > NumbLiabClasses)
            give_up (ERC_lcm, nghd[ind_ndx].self, data[ind_ndx][6],
                     NumbLiabClasses);
      }
   }

   return;
}
