/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* These functions are called from proc_auto().                               */
/* Some are also called from proc_lods() and proc_ibd().                      */

/**
 * checks simulate/select marker/trait/tloc/inheritance statements for errors.
 */
int test_auto_sel (void)
{
   if (!first_sim_chrom_node)           /*select/simu ... markers .. required */
   {
      if (IBDDROP || MARKERDROP)
         err_typ = ERA_SIM;
      else
         err_typ = ERA_SEL;
      return FALSE;
   }

   else if (first_sim_chrom_node->p_next) /*checks for only one chromosome    */
   {
      if (IBDDROP || MARKERDROP)
         err_typ = ERA_OOM;
      else
         err_typ = ERA_OOS;
      return FALSE;
   }

   if (!sim_traits)                     /* checks for sel trait/tloc/inher    */
   {
      if (IBDDROP)                      /* ibddrop:  sim tloc                 */
         err_typ = ERA_MSL;
      else if (MARKERDROP)              /* markerdrop:  sel trait or inher    */
         err_typ = ERA_MTI;
      else if (LM_IBDTESTS || LM_PVAL || LM_AUTO) /* Autozyg/:  sel trait     */
         err_typ = ERA_MST;

      if (err_typ)
         return FALSE;
   }

   return TRUE;
}

/**
 * Checks marker data statements for errors.
 */
int test_markdat (SIM_PTR p_sel)
{
   int   i, j, k;
   MARKDAT_PTR p_mdt;
   LOCUS_PTR p_locus;
   int  *p_a;
   char *p_o;

   /* In most cases when test_markdat is called by an ibd_create subprogram   */
   /* there will be no Morgan style marker data to be checked out for errors. */
   /* The only time this first portion of test_markdat needs to be executed   */
   /* is when the fgl2haplo subprogram is using Morgan style genotypic data.  */

   if (IBD_CREATE)
   {
      if (SubProgChoice != 'f')
         return TRUE;

      /* So it must be the fgl2haplo subprogram calling test_markdat.         */

      if (!INP_GENOTYPES)
         return TRUE;
   }

   NObs = 0;
   if (!(p_mdt = first_markdat_node))   /* finds marker data for this chrom   */
   {
      err_typ = ERA_NMD;
      for_chrom (p_sel->chrom_no);
   }
   else
   {
      while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
         p_mdt = p_mdt->p_next;
      if (p_mdt->chrom_no != p_sel->chrom_no)
      {
         err_typ = ERA_NMD;
         for_chrom (p_sel->chrom_no);
      }
      else
      {                                 /* checks no. of markers              */
         if ((p_sel->no_mark_map && p_mdt->no_marks != p_sel->no_mark_map)
             || (!p_sel->no_mark_map && p_mdt->no_marks != 1))
         {
            err_typ = ERA_BND;
            for_chrom (p_mdt->chrom_no);
         }
         else
            NObs = p_mdt->no_obs;
      }
   }
   if (err_typ)
      return FALSE;

   if (IBD_CREATE)
      return TRUE;

   p_o = p_mdt->p_obs;                  /* tests marker alleles within range  */
   p_a = p_mdt->p_alls;                 /* ... for each observed individual   */
   for (i = 0; i < NObs; ++i)
   {
      p_locus = first_mark_locus_node;
      for (j = 0; j < p_sel->no_mark_map; ++j) /* ... and each marker.        */
      {
         while (((p_locus->p_next && (p_locus->chrom_no < p_sel->chrom_no))
                 || (p_locus->chrom_no == p_sel->chrom_no && p_locus->no <= j)))
            p_locus = p_locus->p_next;

         for (k = 0; k < 2; ++k)
         {
            if (*p_a < 0 || *p_a > p_locus->no_alleles)
            {
               sprintf (errnumb, "%s", p_o);
               err_txt = errnumb;
               err_typ = ERA_WMA;
            }
            ++p_a;
         }
      }
      p_o += MAXNAMLEN + 1;
   }

   if (err_typ)
      return FALSE;
   return TRUE;
}

/**
 * Checks marker names statements for errors.
 */
int test_marknam (SIM_PTR p_sel)
{
   MARKNAM_PTR p_mnm;

   if (!(p_mnm = first_marknam_node))   /* looks for mrk names for this chrom */
      return TRUE;
   else
   {
      while (p_mnm->p_next && p_mnm->chrom_no < p_sel->chrom_no)
         p_mnm = p_mnm->p_next;
      if (p_mnm->chrom_no != p_sel->chrom_no)
         return TRUE;
      else
         if ((p_sel->no_mark_map && p_mnm->no_marks != p_sel->no_mark_map)
             || (!p_sel->no_mark_map && p_mnm->no_marks != 1))
      {
         err_typ = ERA_BNM;
         for_chrom (p_mnm->chrom_no);
      }
   }
   if (err_typ)
      return FALSE;
   return TRUE;
}

/**
 * Checks program gamete statements for errors.
 */
int test_probgam (SIM_PTR p_sel)
{
   PROBGAM_PTR p_pg;
   LOCWND_PTR p_node;

   if (DoingAllPairs || (IBD_CREATE && (SubProgChoice == 'f')))
      return TRUE;

   if (!(p_pg = first_probgam_node))    /* Expects proband gametes.           */
   {
      err_typ = ERA_MPG;
      return TRUE;
   }
   else if ((p_node = first_locwind_node)) /* checks for locus window         */
   {                                    /* for this chromosome                */
      while (p_node->p_next && p_node->chrom_no < p_sel->chrom_no)
         p_node = p_node->p_next;

      if (p_node->chrom_no == p_sel->chrom_no)
      {
         NLocWndw = p_node->window;
         if (NLocWndw < 2 || NLocWndw > NLoci)
         {
            err_typ = ERA_LWO;
            sprintf (errnumb, "%d", NLoci);
            err_txt = errnumb;
         }
      }
   }

   if (err_typ)
      return FALSE;
   return TRUE;
}

int test_probgam_jnt (SIM_PTR p_sel)
{
   PROBGAM_PTR p_pg;

   if (!(p_pg = first_probgam_node))    /* expects proband gametes            */
      err_typ = ERA_MPG;

   if (IBDDROP && NLocWndw)             /* if ibddrop, and locus window,      */
      while (p_pg)                      /* tests for one pair of gametes      */
      {                                 /* in each set                        */
         if (p_pg->no_pgams != 2)
            err_typ = ERA_PIG;
         p_pg = p_pg->p_next;
      }

   if (err_typ)
      return FALSE;
   return TRUE;
}

/**
 * checks the pattern label statements for errors.
 *
 * Expects a locus window size if there are window patterns to score.
 * If not, a warning is issued.
 * Expects at least one proband gamete scoreset for each component for which
 * there are pattern labels; if not, a warning is issued
 * Each pattern label must be in the correct range, given the number of
 * probands.
 */
int test_probgam_pat (SIM_PTR p_sel)
{
   int   i;

   PROBGAM_PTR p_pg;
   WNDPAT_PTR p_wp;

   if (!(p_wp = first_windpat_node))
      return TRUE;

   if (!NLocWndw)
   {
      err_typ = ERA_MLW;
      return FALSE;
   }

   p_pg = first_probgam_node;

   while (p_wp)
   {
      while ((p_pg->p_next && p_pg->component < p_wp->component)
             || (p_pg->p_next && p_pg->component == p_wp->component
                 && p_pg->scoreset < p_wp->scoreset))
         p_pg = p_pg->p_next;

      if (p_wp->component != p_pg->component
          || p_wp->scoreset != p_pg->scoreset)
      {
         err_typ = ERA_MSS;
         sprintf (errnumb, "component %d scoreset %d", p_wp->component,
                  p_wp->scoreset);
         err_txt = errnumb;
      }
      else
      {
         for (i = 1; i < p_wp->no_pats; ++i)
            if (*(p_wp->p_pats + i) > CDgg[p_pg->no_pgams])
            {
               err_typ = ERA_BPC;
               sprintf (errnumb, "%d", *(p_wp->p_pats + i));
               err_txt = errnumb;
            }
      }
      p_wp = p_wp->p_next;
   }

   if (err_typ)
      return FALSE;
   return TRUE;
}

/******************************************************************************/
/*                                                                            */
/* The test_num_liab_classes subroutine checks whether or not any liability   */
/* classes penetrance data has been provided.                                 */
/*                                                                            */
/******************************************************************************/

int test_num_liab_classes (void)
{
   if (NumbLiabClasses == UNDEF)
   {
      err_typ = ERD_LCM;
      return FALSE;
   }

   return TRUE;
}

/**
 * Checks scalar parameters and sets defaults.
 */
int test_auto_pars (void)
{
   if (MARKERDROP)
      return TRUE;

   if (IBD_HAPLO)
   {
      if (PhaseofData == UNDEF)
         err_typ = ERA_SPD;

      else if (DoingAllPairs && PhaseofData == PHASED)
      {
         err_typ = ERA_AIP;
         err_txt = "phased";
      }

      else if (StateTransMat == UNDEF)
         err_typ = ERA_STM;

      else if (PopKinship == UNDEF)
         err_typ = ERA_PKS;

      else if (KinshipChanRat == UNDEF)
         err_typ = ERA_KCR;

      else if (TransMatNullFr == UNDEF)
         err_typ = ERA_TNF;

      else if (GenoErrRat == UNDEF)
         err_typ = ERD_GER;

      if (err_typ)
         return FALSE;
      return TRUE;
   }

   if (IBD_CREATE)
   {
      if (SubProgChoice == 'p')         /* Then we are running simpop_fgl.    */
      {
         if (CMperMbp == UNDEF)
            err_typ = ERD_MCM;

         else if (NumOutGens == UNDEF)
            err_typ = ERD_MFG;

         else if (NumOffGens == UNDEF)
            err_typ = ERD_MOG;

         else if (NumFemsGen == UNDEF)
            err_typ = ERA_MFG;

         else if (ChromLen_bp == UNDEF)
            err_typ = ERD_MCL;

         if (RematWeight == UNDEF)
            RematWeight = 1.0;          /* Then set RematWeight to default.   */
      }

      else if (SubProgChoice == 'f')    /* Then we are running fgl2haplo.     */
      {
         if (HaploInput == UNDEF)
            err_typ = ERD_MIH;

         else if (CMperMbp == UNDEF)
            err_typ = ERD_MCM;
      }

      else if (SubProgChoice == 'i')    /* Then we are running fgl2ibd.       */
      {
         if (CMperMbp == UNDEF)
            err_typ = ERD_MCM;
      }

      else if (SubProgChoice == 's')    /* Then we are running beaglesim.     */
      {
         if (IbdCreateOutput == UNDEF)
            err_typ = ERD_MOH;

         if (LDRelaxProb == UNDEF)
            LDRelaxProb = 0.0;
      }

      if (err_typ)
         return FALSE;
      return TRUE;
   }

   if (NIt == UNDEF)
   {
      if (!LM_MAP)
      {
         err_typ = ERA_NIT;
         return FALSE;
      }
      else
         NIt = MCEMIT;
   }
   if (IBDDROP)
   {
      if (err_typ)
         return FALSE;
      return TRUE;
   }

   if (SmplSetup == UNDEF)
      SmplSetup = SEQIMP;

   if (NSeqImp == UNDEF)
   {
      NSeqImp = RSEQIMP * NIt;
      if (NSeqImp < 2)
         NSeqImp = 2;
   }

   if (Scan == UNDEF)
      Scan = Y;
   if (ScorIt == UNDEF)
   {
      if (!LM_MAP && !LM_IBDTESTS)
         ScorIt = 1;
      else if (LM_IBDTESTS)
         ScorIt = NIt / 50 > 10 ? 50 : NIt / 10;
   }

   if (SeqImpIt == UNDEF)
      SeqImpIt = 0;

   if (GlobalMcmc == UNDEF)
      GlobalMcmc = FALSE;

   if (LM_AUTO && MMSampler == UNDEF)
   {
      err_typ = ERA_MMS;
      return FALSE;
   }

   if (NBrnIt == UNDEF)
   {
      if (LM_MAP)
         NBrnIt = RBRNIT_MAP * NIt;
      else if (LM_PVAL)
         NBrnIt = RBRNIT_PVAL * NIt;
      else if (LM_IBDTESTS)
         NBrnIt = RBRNIT_IBDTESTS * NIt;
      else if (LM_AUTO)
         NBrnIt = RBRNIT_AUTO * NIt;
   }

   if (LM_MAP)
   {
      if (GenoErrRat == UNDEF)          /* If genotyping error rate not set,  */
         GenoErrRat = 0.0;              /* use lm_map's default of 0.0.       */

      if (MapEstMod == UNDEF)           /* Checks for with or wthout mistypng */
         err_typ = ERA_MMT;
      else if (MapEstMod != MISTYP)     /* No mistyping, checks algorithm.    */
      {
         if (MaximAlgo == UNDEF)
            err_typ = ERA_MMA;
      }
      else if (MapEstMod == MISTYP)     /* Mistyping, checks genotyping error */
      {                                 /* rate to be other than 0.0.         */
         if (GenoErrRat == 0.0)
            err_typ = ERA_NGE;
      }

      if (MapTyp == UNDEF)              /* If output map type not specified,  */
         MapTyp = BOTH;                 /* sets default.                      */

      if (err_typ)
         return FALSE;
   }

   if (LM_PVAL)
   {
      if (SpairsOn == UNDEF && RobdomOn == UNDEF && MaxFGLOn == UNDEF)
      {
         MaxFGLOn = 1;                  /* assigns default values */
         SpairsOn = 0;
         RobdomOn = 0;
      }
   }

   if (LM_IBDTESTS)
   {
      if (IBDStats == UNDEF && LikRatStats == UNDEF)
      {
         err_typ = ERA_NCS;
         return FALSE;
      }
      else if (IBDStats == 1)           /* checks IBD statistics options      */
      {
         LikRatStats = 0;
         if (DeltaOn != UNDEF || Lambda_pOn != UNDEF) /* first measures     */
            err_typ = ERA_WIM;
         DeltaOn = 0;
         Lambda_pOn = 0;

         if (NormalEnable == UNDEF || PermutEnable == UNDEF)
         {
            NormalEnable = 1;           /* assigns tests defaults             */
            PermutEnable = 0;
         }

         if (SpairsOn == UNDEF || RobdomOn == UNDEF || SAffUnaffOn == UNDEF
             || LambdaOn == UNDEF || MaxFGLOn == UNDEF)
         {
            SpairsOn = 1;               /* assigns meas defaults              */
            RobdomOn = 0;
            SAffUnaffOn = 0;
            LambdaOn = 0;
            MaxFGLOn = 0;
         }

         if (NormalEnable && SAffUnaffOn)
            err_typ = ERA_NSU;

         else if (ChiSqEnable != UNDEF) /* checks tests                       */
            err_typ = ERA_WIT;
         ChiSqEnable = 0;
      }

      else if (LikRatStats == 1)        /* checks lhood-ratio stat options    */
      {
         IBDStats = 0;
         if (SpairsOn != UNDEF || RobdomOn != UNDEF /* first measures         */
             || SAffUnaffOn != UNDEF || LambdaOn != UNDEF || MaxFGLOn != UNDEF)
            err_typ = ERA_WLM;
         SpairsOn = 0;
         RobdomOn = 0;
         SAffUnaffOn = 0;
         LambdaOn = 0;
         MaxFGLOn = 0;

         if (DeltaOn == UNDEF || Lambda_pOn == UNDEF)
         {
            DeltaOn = 1;                /* assigns meas defaults              */
            Lambda_pOn = 0;
         }
         if (NormalEnable != UNDEF || PermutEnable != UNDEF) /* checks tests  */
            err_typ = ERA_WLT;
         NormalEnable = 0;              /* assigns tests defaults             */
         PermutEnable = 0;

         if (ChiSqEnable == UNDEF)
            ChiSqEnable = 1;

         if (Lambda_pOn)                /* assigns grid spacing               */
         {
            IntervalSegL = .5 / (NumOfSegL - 1.);
            IntervalSegP = .8 / (NumOfSegP - 1.);
         }
      }
   }

   if (ChkIt == UNDEF)
      ChkIt = NIt / 5;

   if (FilIt == UNDEF)
      FilIt = NIt / ScorIt;

   if (err_typ)
      return FALSE;
   return TRUE;
}

/* This subroutine uses the trait number passed as its only argument to find  */
/* a trait_data_node for the trait if one exists.  If it does exist, the data */
/* from the trait_data_node is then used to initialize the PedDatTyp and      */
/* TraitTyp global variables.  If there is no trait_data_node for the trait,  */
/* the two global variables are not modified.                                 */

void set_trait_typ (int t_no)
{
   TRAITDATA_PTR p_td;

   p_td = first_trait_data_node;
   while (p_td && p_td->trait_no < t_no)
      p_td = p_td->p_next;

   if (p_td && p_td->trait_no == t_no)
   {
      PedDatTyp = p_td->peddat_typ;
      TraitTyp = p_td->trait_typ;
   }

   return;
}

/**
 * checks and sets TraitTyp (trait data type)
 * and PedDatTyp (integers, integer pairs, reals)
 */
int test_trait_typ (void)
{
   if (TraitTyp == UNDEF)
   {
      if (MARKERDROP)
      {
         if (TRAIT_SELCTD)
            TraitTyp = DISCRETE;
         else if (INHER_SELCTD)
            TraitTyp = INHERITANCE;
      }

      if (LM_IBDTESTS)
         TraitTyp = DISCRETE;

      if (LM_AUTO || LM_PVAL)
         TraitTyp = GENOTYPIC;

      if (LM_MARKERS || LM_BAYES || LM_LODS)
         TraitTyp = GENOTYPIC;

      if (LM_TWOQTL || LM_SCHNELL)
         TraitTyp = QUANTITATIVE;

      if (TRAIT_DATA)
         PedDatTyp = TraitTyp;
   }
   else if (MARKERDROP)
   {
      if (TRAIT_SELCTD && TraitTyp != DISCRETE)
      {
         TraitTyp = DISCRETE;
         PedDatTyp = TraitTyp;
         err_typ = ERA_IPD;
      }
      else if (INHER_SELCTD && TraitTyp == DISCRETE)
      {
         TraitTyp = INHERITANCE;
         PedDatTyp = TraitTyp;
         err_typ = ERA_IPT;
      }
   }

   if (TraitTyp == GENOTYPIC)
      NAforMarkers = TRUE;

   if (!TRAIT_DATA)
      PedDatTyp = UNDEF;

   if (err_typ)
      return FALSE;
   return TRUE;
}

/**
 * checks pedigree record parameters for trait.
 * tests for trait record positions compatible with no. of integers or reals.
 * allows for gender added by Pedchk/ function in pedigree checking step.
 */
int test_ped_trtdat (void)
{
   int   nd = 1;
   int   tno = 0;
   TRTPED_PTR p_p;

   if (sim_traits)                      /*gets trait or inheritance no.       */
      tno = *(sim_traits->p_trait);

   if (PedDatTyp == INHERITANCE || PedDatTyp == DISCR_LIAB
       || PedDatTyp == DISCR_COVAR)
      ++nd;

   if ((p_p = find_trtped (tno)))       /* find record position for data      */
   {
      switch (PedDatTyp)
      {
          case QUANTITATIVE:
             if (p_p->dat_typ != REALDAT)
             {
                err_typ = ERA_BRR;
                sprintf (errnumb, "%d", tno);
                err_txt = errnumb;
             }
             else if (p_p->pos[0] > NReals) /* checks pos against no reals    */
             {
                err_typ = ERA_NEK;
                sprintf (errnumb, "%d", p_p->pos[0]);
                err_txt = errnumb;
             }
             break;

          case INHERITANCE:
          case DISCR_LIAB:
          case DISCR_COVAR:
             if ((PedDatTyp == DISCR_LIAB || PedDatTyp == DISCR_COVAR)
                 && p_p->dat_typ != INTDAT2)
             {
                err_typ = ERA_BI2;
                sprintf (errnumb, "%d", tno);
                err_txt = errnumb;
             }
             else if (PedDatTyp == INHERITANCE && p_p->dat_typ != INHDAT2)
             {
                err_typ = ERA_BH2;
                sprintf (errnumb, "%d", tno);
                err_txt = errnumb;
             }

             else if (p_p->pos[0] > NInts) /* checks pos against no. ints     */
             {
                err_typ = ERA_NEJ;
                sprintf (errnumb, "%d", p_p->pos[0]);
                err_txt = errnumb;
             }
             else if (p_p->pos[1] > NInts) /* checks 2nd pos and no. ints     */
             {
                err_typ = ERA_NEJ;
                sprintf (errnumb, "%d", p_p->pos[1]);
                err_txt = errnumb;
             }
             else if (p_p->pos[0] == p_p->pos[1])
             {
                err_typ = ERA_DRP;
                sprintf (errnumb, "%d", p_p->pos[0]);
                err_txt = errnumb;
             }
             break;

          case GENOTYPIC:
          case DISCRETE:
          default:
             if (p_p->dat_typ != INTDAT)
             {
                err_typ = ERA_BIR;
                sprintf (errnumb, "%d", p_p->no);
                err_txt = errnumb;
             }
             if (p_p->pos[0] > NInts)   /* checks position against no. ints   */
             {
                err_typ = ERA_NEJ;
                sprintf (errnumb, "%d", p_p->pos[0]);
                err_txt = errnumb;
             }
             break;
      }
      if (err_typ)
         return FALSE;
   }
   else                                 /* didn't find data pos for trait/inh */
   {
      switch (TraitTyp)
      {
          case QUANTITATIVE:
             err_typ = ERA_TER;
             break;
          case INHERITANCE:
             err_typ = ERA_TE2;
             break;
          default:
             err_typ = ERA_TEI;
             break;
      }
      sprintf (errnumb, "%d", tno);
      err_txt = errnumb;

      if (PedDatTyp != QUANTITATIVE)    /* note: SexIn may be UNDEF           */
      {
         if (tno + (SexIn != N ? 1 : 0) > NInts - nd + 1)
         {
            err_typ = ERA_NEI;
            sprintf (errnumb, "%d", tno);
            err_txt = errnumb;
         }
      }
      else if (tno > NReals)
      {
         err_typ = ERA_NER;
         sprintf (errnumb, "%d", tno);
         err_txt = errnumb;
      }
   }

   if (err_typ)
      return FALSE;

   return TRUE;
}

/**
 * sets mapping information for tloc relative to the "selected" marker.
 * determines:
 *    the selected reference marker for mapping
 *    the dist from the referenced marker
 */
void sel_tloc_locus (PRDIST_PTR p_pl)
{
   p_pl->selectd = 0;                   /* presently no selection for tlocs   */

   if (p_pl->map_mark == 0)             /* tloc precedes 1st marker:          */
   {
      if (p_pl->p_next->selectd)        /* sets selected marker and dist      */
         p_pl->sel_map_mark = p_pl->p_next->no;
      else
         p_pl->sel_map_mark = p_pl->p_next->sel_map_mark;
      p_pl->sel_dist = p_pl->dist + p_pl->p_next->sel_dist;
   }

   else if (p_pl->p_prev)               /* tloc follows 1st marker:           */
   {
      if (p_pl->p_prev->selectd)        /* preceding marker selected:         */
      {
         p_pl->sel_map_mark = p_pl->p_prev->no;
         p_pl->sel_dist = p_pl->dist;   /* sets sel dist to map dist          */
      }
      else if (p_pl->p_prev->sel_map_mark < p_pl->p_prev->no) /* some prev    */
      {                                 /* marker selected:                   */
         p_pl->sel_map_mark = p_pl->p_prev->sel_map_mark; /* adds sel dist    */
         p_pl->sel_dist = p_pl->dist + p_pl->p_prev->sel_dist; /* to map dist */
      }
      else if (p_pl->p_prev->sel_map_mark > p_pl->p_prev->no) /* no prev      */
      {                                 /* marker selected                    */
         p_pl->sel_map_mark = p_pl->p_prev->sel_map_mark; /* subtrcts map dst */
         p_pl->sel_dist = p_pl->p_prev->sel_dist - p_pl->dist; /* from sel    */
      }
   }

   return;
}

/**
 * Removes prntmap locus nodes for non-selected markers.
 * adjusts forward and backward pointers.
 */
void sel_prntmap_mark (PRNTMAP_PTR p_pm)
{
   PRDIST_PTR p, p_sav;

   while ((p = p_pm->p_locus) && (p->locus_typ == MARK_LOC) && (!p->selectd))
   {                                    /* looks for start of list            */
      p_pm->p_locus = p->p_next;        /* advances ptr to first locus        */
      free (p);                         /* removes this locus                 */
   }
   p->p_prev = NULL;

   p_sav = p;                           /* checks loci for deletion           */
   while ((p = p_sav->p_next))
   {
      if (p->locus_typ == MARK_LOC && !p->selectd)
      {
         p_sav->p_next = p->p_next;     /* advances last locus' forward ptr   */
         free (p);                      /* removes current locus              */
      }
      else
      {
         p->p_prev = p_sav;             /* updates current locus' backwrd ptr */
         p_sav = p;                     /* keeps current locus                */
      }
   }

   return;
}

/*
 * removes non-selected marker loci from list of loci
 */
void shrink_prntmap (PRNTMAP_PTR p_pmap)
{
   PRDIST_PTR p_cur, p_nxt, p_sav;

   p_cur = p_pmap->p_locus;
   while (p_cur && p_cur->locus_typ == MARK_LOC && !p_cur->selectd)
   {
      p_nxt = p_cur->p_next;
      free (p_cur);
      p_cur = p_nxt;
   }
   p_pmap->p_locus = p_cur;
   p_sav = p_cur;
   if (p_cur)                           /* 1st selected marker          */
   {
      p_cur->p_prev = NULL;
      p_cur->dist = p_cur->sel_dist;
   }

   while (p_cur)
   {
      if (p_cur->locus_typ == MARK_LOC && !p_cur->selectd)
      {
         p_sav->p_next = p_cur->p_next;
         p_nxt = p_cur->p_next;
         free (p_cur);
         p_cur = p_nxt;
         if (p_cur)
            p_cur->p_prev = p_sav;
      }
      else
      {
         p_sav = p_cur;                 /* next selected marker        */
         p_cur->dist = p_cur->sel_dist;
         p_cur = p_cur->p_next;
      }
   }

   return;
}

int test_penetrnc (void)
{
   TLOCPARS_PTR p_tlocp;
   int   i1;

   if (P_trait_pars != NULL)
      p_tlocp = P_trait_pars->p_tloc_pars;
   else
      p_tlocp = NULL;

   if (NTraits && TraitTyp == DISCRETE && PedDatTyp != DISCR_LIAB
       && PedDatTyp != DISCR_COVAR)
   {
      if (p_tlocp)
      {
         if (p_tlocp->num_penetrances < 3)
            err_typ = ERA_MPV;
         else
            for (i1 = 0; i1 < p_tlocp->num_penetrances; ++i1)
               if (p_tlocp->penetrances[i1] == 0.0)
               {
                  NAforMarkers = TRUE;
                  err_typ = ERA_ZPV;
               }
      }
   }

   if (MARKERDROP && PedDatTyp == INHERITANCE)
   {
      if (p_tlocp != NULL)
         p_tlocp->num_penetrances = 0;
      err_typ = ERA_PNI;
   }

   if (err_typ)
      return FALSE;

   return TRUE;
}

void prnt_penetrnc (void)
{
   TLOCPARS_PTR p_tlocp;
   double *penetrances;

   if (P_trait_pars != NULL)
      p_tlocp = P_trait_pars->p_tloc_pars;
   else
      p_tlocp = NULL;

   if (p_tlocp == NULL || p_tlocp->num_penetrances == 0)
      return;

   if (PrintLevel >= 2)
   {
      penetrances = p_tlocp->penetrances;
      fprintf (stdout, " Incomplete penetrances for the trait are:\n");
      fprintf (stdout, "        genotype  code  penetrance\n");
      fprintf (stdout, "        --------  ----  ----------\n");
      fprintf (stdout, "           (1 1)     0       %5.3f\n", penetrances[0]);
      fprintf (stdout, "    (1 2), (2 1)     1       %5.3f\n", penetrances[1]);
      fprintf (stdout, "           (2 2)     2       %5.3f\n", penetrances[2]);
      fprintf (stdout, "\n");
   }

   return;
}

/**
 * allocates and sets MarkNam[][] which contains the names of the
 * selected markers, either provided names or default names.
 * returns true if marker names given in par statements.
 */
int mak_marknam (SIM_PTR p_sel)
{
   MARKNAM_PTR p_mnm;
   int   i, no, nc;
   char *p_n = NULL;

   if (!NMarks)
      return FALSE;

   MarkNam = malloc (NMarks * sizeof (char *));
   alloc_chk (WHERE, MarkNam);
   for (i = 0; i < NMarks; ++i)
   {
      MarkNam[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, MarkNam[i]);
   }

   if ((p_mnm = first_marknam_node))    /* checks for names given for chrom   */
   {
      while (p_mnm->p_next && p_mnm->chrom_no < p_sel->chrom_no)
         p_mnm = p_mnm->p_next;
      if (p_mnm->chrom_no == p_sel->chrom_no)
         p_n = p_mnm->p_nams;
   }

   for (i = 0; i < NMarks; ++i)
   {
      if (!p_sel->p_mark)
         no = i + 1;
      else
         no = *(p_sel->p_mark + i);
      if (p_n)
      {
         if ((nc = get_marknam (MarkNam[i], p_n, no)) > LocNamLen)
            LocNamLen = nc;
      }
      else if ((nc = fmt_marknam (MarkNam[i], no)) > LocNamLen)
         LocNamLen = nc;
   }

   if (p_n)
      return TRUE;
   else
      return FALSE;
}

/**
 * allocates and sets TraitNam[][], which contains the name
 * of the selected trait.
 */
void mak_traitnam (TLLIST_PTR p_tr)
{
   int   i, nc;

   if (!NTraits)
      return;

   TraitNam = malloc (NTraits * sizeof (char *));
   alloc_chk (WHERE, TraitNam);

   for (i = 0; i < NTraits; ++i)
   {
      TraitNam[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, TraitNam[i]);
   }
   for (i = 0; i < NTraits; ++i)
   {
      if ((nc = fmt_traitnam (TraitNam[i], p_tr->no)) > LocNamLen)
         LocNamLen = nc;
   }

   return;
}

/**
 * allocates and sets TLocNam[][], which contains the name
 * of the selected tloc.
 */
void mak_tlocnam (int tln)
{
   int   i, nc;

   if (!NTLocs)
      return;

   TLocNam = malloc (NTLocs * sizeof (char *));
   alloc_chk (WHERE, TLocNam);

   for (i = 0; i < NTLocs; ++i)
   {
      TLocNam[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, TLocNam[i]);
   }
   for (i = 0; i < NTLocs; ++i)
   {
      if ((nc = fmt_tlocnam (TLocNam[i], tln)) > LocNamLen)
         LocNamLen = nc;
   }

   return;
}

/* Allocates global arrays for marker positions.                              */
/*   Marek:   2 sexes, number of marker recombination fractions               */
/*            plus beginning and ending unlinked positions.                   */
/*   MarkPos: 2 sexes, number of marker positions.                            */

void alloc_marek (void)
{
   int   sx;

   Marek = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, Marek);
   MarkPos = malloc (2 * sizeof (double *));
   alloc_chk (WHERE, MarkPos);

   for (sx = FEM; sx <= MAL; ++sx)
   {
      Marek[sx] = malloc (NLoci * sizeof (double));
      alloc_chk (WHERE, Marek[sx]);
      MarkPos[sx] = malloc ((NLoci - 1) * sizeof (double));
      alloc_chk (WHERE, MarkPos[sx]);
   }

   return;
}

/* Stores recombination fractions for markers.                                */
/* Makes Marek[][]: sex-specific recombination fractions.                     */
/* All tloc loci and the final entry in Marek[][] are initialized to 0.5.     */

void mak_marek (void)
{
   PRNTMAP_PTR p_pm;
   PRDIST_PTR p_loc;
   double pos;
   int   first_marker_ndx;
   int   entry, sx, i1;

   first_marker_ndx = max (NTLocs, 1);

   for (sx = FEM; sx <= MAL; ++sx)
   {
      entry = FALSE;
      if ((p_pm = nxt_prntmap (&entry, sx)))
      {
         pos = 0.0;

         /* Initialize MarkPos and Marek so that tlocs are ignored if they    */
         /* are linked.                                                       */
         for (i1 = 0; i1 < first_marker_ndx; ++i1)
         {
            MarkPos[sx][i1] = pos;
            Marek[sx][i1] = 0.5;
         }

         i1 = first_marker_ndx;         /* Start with the first marker locus. */

         p_loc = p_pm->p_locus;
         while (p_loc)
         {
            if (p_loc->locus_typ == MARK_LOC && p_loc->dist)
            {
               pos += p_loc->dist;
               MarkPos[sx][i1] = pos;
               Marek[sx][i1] = dist2rf (p_loc->dist);
               ++i1;
            }

            p_loc = p_loc->p_next;
         }

         Marek[sx][i1] = 0.5;
      }
   }

   return;
}

/* Stores recombination fractions for tlocs.                                  */
/* Makes Int0, Rec0[2] - tloc position and sex-specific recomb fraction.      */

void mak_tarek (void)
{
   int   sx, entry, j;
   PRNTMAP_PTR p_pm;
   PRDIST_PTR p_loc;

   if (!NLnkdTL)
   {
      Loc0 = 0;
      Rec0[0] = .5;
      Rec0[1] = .5;
      TraitPos[0] = .5;
      TraitPos[1] = .5;
   }
   else
      for (sx = FEM; sx <= MAL; ++sx)
      {
         entry = FALSE;
         if ((p_pm = nxt_prntmap (&entry, sx)))
         {
            p_loc = p_pm->p_locus;
            for (j = 0; p_loc; p_loc = p_loc->p_next)
            {
               if (p_loc->locus_typ == TLOC_LOC)
               {
                  Loc0 = j;
                  Rec0[sx] = dist2rf (p_loc->dist);
                  TraitPos[sx] = p_loc->dist;
               }
               else if (p_loc->locus_typ == MARK_LOC && p_loc->selectd)
                  ++j;
            }
         }
      }

   return;
}

/* Allocates storage for MarSel[], which will be used to contain the indices  */
/* of those markers selected for inclusion in the analysis.                   */

void alloc_marsel (void)
{
   if (!NMarks)
      return;

   MarSel = malloc (NMarks * sizeof (int));
   alloc_chk (WHERE, MarSel);

   return;
}

/* Allocates arrays and stores allele frequencies for all selected loci.      */
/* On first entry, malloc NAll[NLoci] and Qq[NLoci] (NLoci is NMarks+NTLocs). */
/*    NAll: 0 ... NLoci-1                                                     */
/*    Qq: 1st subscript: 0 ... NLoci-1                                        */
/*    Qq: 2nd: 1 ... NAll[1st subscript]                                      */
/* Note: NAll[<NTLocs], Qq[<NTLocs] are tlocs).                               */

void mak_allefrq (PRNTMAP_PTR p_pmap, PRDIST_PTR p_ploc, int no)
{
   static int l_ndx;                    /* Subscript to current marker.       */
   static int t_ndx;                    /* Subscript to current tloc.         */

   double sum;
   int   i1;

   LOCUS_PTR p_locus;

   if (!Qq)
   {
      NAll = malloc (NLoci * sizeof (int));
      alloc_chk (WHERE, NAll);
      Qq = malloc (NLoci * sizeof (double *));
      alloc_chk (WHERE, Qq);

/* Arrange it so that the marker allele frequencies will be stored beginning  */
/* at the l_ndx'th row of NAll and Qq.  Leave room in the first NTLocs' rows  */
/* for the trait loci frequencies, making sure that at least the first row    */
/* (locus 0) is not used for marker allele frequencies.                       */

      l_ndx = max (NTLocs, 1);
      t_ndx = 0;
   }

   /* If the locus is a marker, set the number of alleles in Nall and store   */
   /* the allele frequencies in Qq.                                           */

   if (p_ploc && p_ploc->locus_typ == MARK_LOC)
   {
      p_locus = first_mark_locus_node;
      while (((p_locus->p_next
               && (p_locus->chrom_no < p_pmap->chrom_no))
              || (p_locus->chrom_no == p_pmap->chrom_no
                  && p_locus->no < p_ploc->no)))
         p_locus = p_locus->p_next;

      NAll[l_ndx] = p_locus->no_alleles;

      Qq[l_ndx] = malloc ((NAll[l_ndx] + 1) * sizeof (double));
      alloc_chk (WHERE, Qq[l_ndx]);

      Qq[l_ndx][0] = 0.0;               /* First element is unused.           */
      sum = p_locus->p_freq_node->sum;
      for (i1 = 0; i1 < NAll[l_ndx]; ++i1)
      {
         Qq[l_ndx][i1 + 1] = *(p_locus->p_freq_node->p_freq + i1);
         if (alle_frq_norm)             /* Normalize frequencies, if needed.  */
            Qq[l_ndx][i1 + 1] /= sum;
      }
      ++l_ndx;
   }

   /* If there is no trait locus, we still need NAll[0] and Qq[0][].          */

   else if (no == 0)
   {
      NAll[0] = 2;

      Qq[0] = malloc (3 * sizeof (double));
      alloc_chk (WHERE, Qq[0]);

      Qq[0][0] = 0.0;                   /* First element is unused.           */
      Qq[0][1] = 0.5;
      Qq[0][2] = 0.5;
   }

   /* If the locus is a tloc, set the number of alleles in Nall and store the */
   /* allele frequencies in Qq.                                               */

   else
   {
      p_locus = first_tloc_locus_node;
      while (p_locus->p_next && p_locus->no < no)
         p_locus = p_locus->p_next;

      NAll[t_ndx] = p_locus->no_alleles;

      Qq[t_ndx] = malloc ((NAll[t_ndx] + 1) * sizeof (double));
      alloc_chk (WHERE, Qq[t_ndx]);

      Qq[t_ndx][0] = 0.0;               /* First element is unused.           */
      sum = p_locus->p_freq_node->sum;
      for (i1 = 0; i1 < NAll[t_ndx]; ++i1)
      {
         Qq[t_ndx][i1 + 1] = *(p_locus->p_freq_node->p_freq + i1);
         if (alle_frq_norm)             /* Normalize frequencies, if needed.  */
            Qq[t_ndx][i1 + 1] /= sum;
      }
      ++t_ndx;
   }

   return;
}

/* For the chromosome selected in the sim_chrom_node passed as the only       */
/* argument to mak_mdist, this subroutine creates a double precision vector   */
/* containing the sex-averaged distances between the selected markers on the  */
/* chromosome.  The resulting vector is saved in the Mdist[] global variable. */
/* This subroutine also saves the position of the first selected marker on    */
/* the chromosome in the FirstMarkPos global variable.                        */

void mak_mdist (SIM_PTR p_sim)
{
   MAP_PTR p_map;
   DIST_PTR p_d;

   int   entry;
   int   num_dists;

   int   md_ndx;                        /* Subscript into Mdist vector.       */
   int   msc_ndx;                       /* Subscript to current marker.       */
   int   msp_ndx;                       /* Subscript to preceding marker.     */

   /* Find the sex-averaged marker map_node for the selected chromosome.      */

   entry = FALSE;
   while ((p_map = nxt_mark_map (&entry, AVG))
          && p_map->chrom_no < p_sim->chrom_no)
      ;
   if (p_map && p_map->chrom_no == p_sim->chrom_no)
   {
      /* For the selected chromosome, allocate enough storage for a double    */
      /* precision vector to store the distance from the preceding selected   */
      /* marker for each selected marker.  Since this is a vector consisting  */
      /* of distances it needs to contain one less than the total number of   */
      /* selected markers.                                                    */

      num_dists = NMarks - 1;
      Mdist = alloc_dbl1d (num_dists);

      /* Save a pointer to the dist_node array of marker distances.           */

      p_d = p_map->p_dist;

      /* Keep the position of the first selected marker.  Using cumulative    */
      /* distances to each following selected marker this can be used to      */
      /* determine positions of each of the selected markers.                 */

      FirstMarkPos = p_d[MarSel[0] - 1].pos;

      /* For each selected marker after the first marker, calculate the       */
      /* current marker's distance from the preceding selected marker.        */

      for (md_ndx = 0; md_ndx < num_dists; ++md_ndx)
      {
         msc_ndx = MarSel[md_ndx + 1] - 1;
         msp_ndx = MarSel[md_ndx] - 1;
         Mdist[md_ndx] = p_d[msc_ndx].pos - p_d[msp_ndx].pos;
      }
   }

   return;
}

void mak_markdat (SIM_PTR p_sel)
{
   int   i, j, k;
   int  *p_a;
   char *p_o;
   MARKDAT_PTR p_mdt;

   if (NObs == 0)
      return;

   if (!(p_mdt = first_markdat_node))   /* finds marker data for this chrom   */
      return;

   while (p_mdt->p_next && p_mdt->chrom_no < p_sel->chrom_no)
      p_mdt = p_mdt->p_next;
   if (p_mdt->chrom_no != p_sel->chrom_no)
      return;

   ObsNam = malloc (NObs * sizeof (char *)); /* mallocs:                     */
   alloc_chk (WHERE, ObsNam);           /*        ObsNam [NObs][MAXNAMLEN+1]  */
   ObsIx = malloc (NObs * sizeof (int)); /*      ObsIx  [NObs]               */
   alloc_chk (WHERE, ObsIx);
   ObsMark = malloc (NObs * sizeof (int **)); /* ObsMark[NObs][NMarks][2]    */
   alloc_chk (WHERE, ObsMark);

   for (i = 0; i < NObs; ++i)
   {
      ObsNam[i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
      alloc_chk (WHERE, ObsNam[i]);
      ObsMark[i] = malloc (NMarks * sizeof (int *));
      alloc_chk (WHERE, ObsMark[i]);
      for (j = 0; j < NMarks; ++j)
      {
         ObsMark[i][j] = malloc (2 * sizeof (int));
         alloc_chk (WHERE, ObsMark[i][j]);
      }
   }

   p_o = p_mdt->p_obs;
   p_a = p_mdt->p_alls;
   for (i = 0; i < NObs; ++i)           /* stores indices, names, alleles     */
   {
      ObsIx[i] = 0;
      strncpy (ObsNam[i], p_o, MAXNAMLEN + 1);
      p_o += MAXNAMLEN + 1;

      for (j = 0, k = 1; k <= (p_sel->no_mark_map ?
                               p_sel->no_mark_map : 1); ++k)
      {
         if (!p_sel->p_mark ||
             (j < p_sel->no_mark_sel && k == *(p_sel->p_mark + j)))
         {
            ObsMark[i][j][0] = *p_a;
            ObsMark[i][j][1] = *(p_a + 1);
            ++j;
         }
         p_a = p_a + 2;
      }
   }

   return;
}

void mak_probgam (void)
{
   int   c, s, i;
   PROBGAM_PTR p_pg;
   PROBLIST_PTR p;

   if ((p_pg = first_probgam_node) == NULL)
   {
      /* If there are no probgam_node's we just need to set the number of     */
      /* components with scoresets to 0, before returning to proc_auto.       */
      NScCmp = 0;
      NScCmpAlloc = NScCmp;
      return;
   }

   while (p_pg->p_next)                 /* Initializes number of components   */
      p_pg = p_pg->p_next;              /* with proband gametes.              */
   NScCmp = p_pg->component;
   NScCmpAlloc = NScCmp;

   NScSet = malloc (NScCmp * sizeof (int)); /* Allocations for components.    */
   alloc_chk (WHERE, NScSet);
   ScSetNo = malloc (NScCmp * sizeof (int *));
   alloc_chk (WHERE, ScSetNo);
   NPrGam = malloc (NScCmp * sizeof (int *));
   alloc_chk (WHERE, NPrGam);
   ProbNam = malloc (NScCmp * sizeof (char ***));
   alloc_chk (WHERE, ProbNam);
   ProbIx = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbIx);
   ProbGam = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, ProbGam);

   p_pg = first_probgam_node;
   for (c = 0; c < NScCmp; ++c)         /* sets no. of scoresets for ea comp  */
   {
      NScSet[c] = 0;
      while (p_pg && p_pg->component <= c + 1)
      {
         ++NScSet[c];
         p_pg = p_pg->p_next;
      }
      if (!NScSet[c])                   /* gets no. of scoresets for ea comp  */
      {
         ScSetNo[c] = NULL;
         NPrGam[c] = NULL;              /* none: sets this component to empty */
         ProbNam[c] = NULL;
         ProbIx[c] = NULL;
         ProbGam[c] = NULL;
      }
      else
      {
         ScSetNo[c] = malloc (NScSet[c] * sizeof (int)); /* allocs for sets  */
         alloc_chk (WHERE, ScSetNo[c]);
         NPrGam[c] = malloc (NScSet[c] * sizeof (int));
         alloc_chk (WHERE, NPrGam[c]);
         ProbNam[c] = malloc (NScSet[c] * sizeof (char **));
         alloc_chk (WHERE, ProbNam);
         ProbIx[c] = malloc (NScSet[c] * sizeof (int *));
         alloc_chk (WHERE, ProbIx[c]);
         ProbGam[c] = malloc (NScSet[c] * sizeof (int *));
         alloc_chk (WHERE, ProbGam[c]);
      }
   }
   p_pg = first_probgam_node;           /* sets no. of proband gametes for    */
   for (c = 0; c < NScCmp; ++c)         /* ea scoreset                        */
      for (s = 0; s < NScSet[c]; ++s)
      {
         ScSetNo[c][s] = p_pg->scoreset;
         NPrGam[c][s] = p_pg->no_pgams; /* allocs for proband gametes         */
         ProbNam[c][s] = malloc (NPrGam[c][s] * sizeof (char *));
         alloc_chk (WHERE, ProbNam[c][s]);
         ProbIx[c][s] = malloc (NPrGam[c][s] * sizeof (int));
         alloc_chk (WHERE, ProbIx[c][s]);
         ProbGam[c][s] = malloc (NPrGam[c][s] * sizeof (int));
         alloc_chk (WHERE, ProbGam[c][s]);

         p = p_pg->p_pglist;            /* stores names and haplotypes of     */
         for (i = 0; i < p_pg->no_pgams; ++i) /* proband gametes             */
         {
            ProbNam[c][s][i] = malloc ((MAXNAMLEN + 1) * sizeof (char));
            alloc_chk (WHERE, ProbNam[c][s][i]);
            strncpy (ProbNam[c][s][i], p->nam, MAXNAMLEN + 1);
            ProbGam[c][s][i] = p->gamete;
            ++p;
         }
         if (NPrGam[c][s] > MxGam)      /* looks for max no. prob gametes     */
            MxGam = NPrGam[c][s];
         p_pg = p_pg->p_next;
      }

   return;
}

void mak_patlab (void)
{
   int   c, s, i;
   int  *p;
   WNDPAT_PTR p_wp;

   p_wp = first_windpat_node;

   NPtLab = malloc (NScCmp * sizeof (int *)); /* allocs for components       */
   alloc_chk (WHERE, NPtLab);
   PatLab = malloc (NScCmp * sizeof (int **));
   alloc_chk (WHERE, PatLab);

   for (c = 0; c < NScCmp; ++c)
      if (!NScSet[c])                   /* tests for probands for this scset  */
      {
         NPtLab[c] = NULL;
         PatLab[c] = NULL;
      }
      else
      {
         NPtLab[c] = malloc (NScSet[c] * sizeof (int)); /* allocs for scsets */
         alloc_chk (WHERE, NPtLab[c]);
         PatLab[c] = malloc (NScSet[c] * sizeof (int *));
         alloc_chk (WHERE, PatLab);

         for (s = 0; s < NScSet[c]; ++s) /* sets no. of patterns for ea      */
         {                              /* component scoreset                 */
            while (p_wp
                   && ((p_wp->p_next && p_wp->component <= c)
                       || (p_wp->p_next && p_wp->component == c + 1
                           && p_wp->scoreset < ScSetNo[c][s])))
               p_wp = p_wp->p_next;

            if (p_wp && p_wp->component == c + 1
                && p_wp->scoreset == ScSetNo[c][s])
            {
               NPtLab[c][s] = p_wp->no_pats; /* allocs for pattern labels    */
               PatLab[c][s] = malloc (NPtLab[c][s] * sizeof (int));
               alloc_chk (WHERE, PatLab[c][s]);

               p = p_wp->p_pats;        /* stores pattern labels              */
               for (i = 0; i < p_wp->no_pats; ++i, ++p)
                  PatLab[c][s][i] = *p;
            }
            else
            {
               NPtLab[c][s] = 0;        /* sets to no patterns for component  */
               PatLab[c][s] = NULL;     /* and scoreset (or none at all)      */
            }
         }
      }

   return;
}

/**
 * makes SelInt[], SelReal[] and NColData.
 * SelInt[0] ... SelInt[n], where n is the number of integers in the ped file
 * the ith integer is to be stored in data[][i]
 * NColData is the allocation: data[][NColData]
 * SelReal[0] ... SelReal[n], where n is the number of reals in the ped file
 * the ith real is to be stored in quant[][i]
 */
void mak_ped_dat (void)
{
   TRTPED_PTR p_p;
   int   tno;
   int   j;

   alloc_sel ();

   SelInt[0] = 4;                       /* (gender)                           */
   NColData = 5;                        /* NColData is 5 or greater           */

   if (!sim_traits)                     /* Return if no sim_traits_node.      */
      return;

   if (!TLOC_SELCTD)
      tno = *(sim_traits->p_trait);
   else
      tno = (SexIn != N) ? 1 : 0;

   if ((p_p = find_trtped (tno)))       /* looks for ped record pos for data  */
   {
      while (p_p->p_next && p_p->no < tno)
         p_p = p_p->p_next;
      if (p_p->no != tno)
         p_p = NULL;
   }

   if (PedDatTyp != QUANTITATIVE)       /* sets up SelInt[]                   */
   {
      if (!p_p)
      {
         SelInt[tno] = NColData;        /* sets record pos from trait no.     */
         ++NColData;

         if (PedDatTyp == INHERITANCE || PedDatTyp == DISCR_LIAB
             || PedDatTyp == DISCR_COVAR)
         {
            SelInt[tno + 1] = NColData; /* sets 2nd record pos for data       */
            ++NColData;
         }
      }
      else
      {
         j = p_p->pos[0];               /* Sets specified record position.    */
         if (SexIn != N)                /* Note: SexIn may be UNDEF.          */
         {
            --j;
            if (j == 0)
            {
               err_typ = ERA_NES;
               sprintf (errnumb, "%d", p_p->pos[0]);
               err_txt = errnumb;
            }
         }
         SelInt[j] = NColData;
         ++NColData;

         if (PedDatTyp == INHERITANCE || PedDatTyp == DISCR_LIAB
             || PedDatTyp == DISCR_COVAR)
         {
            j = p_p->pos[1];            /* Sets second record pos for data.   */
            if (SexIn != N)             /* Note: SexIn may be UNDEF.          */
               --j;
            SelInt[j] = NColData;
            ++NColData;
         }
      }
   }

   else if (PedDatTyp == QUANTITATIVE)  /* sets up SelReal[]                  */
   {
      if (!p_p)
         SelReal[tno - 1] = 1;          /* sets record pos from trait no.     */
      else
         SelReal[p_p->pos[0] - 1] = 1;  /* sets specified record pos          */
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Prints the pedigree or individual's record position(s) for trait data.     */
/*                                                                            */
/******************************************************************************/

void prnt_ped_dat (void)
{
   char  trait_source[18];              /* Identifies trait data source.      */

   int   ic[2], rcol;
   int   i, j;

   /* Determine whether an input pedigree file or an input individuals file   */
   /* is being used as a source for the trait data.                           */

   if (FromIndivsFile == TRUE)
      sprintf (trait_source, "%s", "each individual's");
   else
      sprintf (trait_source, "%s", "the pedigree");

   j = 0;
   ic[0] = 0;
   ic[1] = 0;
   if (SexIn != N)                      /* Note: SexIn may be UNDEF.          */
      ++j;
   for (i = j; i < NInts; ++i)
      if (SelInt[i])
      {
         if (!ic[0])
            ic[0] = i;
         else if (SelInt[ic[0]] < SelInt[i])
            ic[1] = i;
         else
         {
            ic[1] = ic[0];
            ic[0] = i;
         }
      }

   for (i = 0; i < NReals; ++i)
      if (SelReal[i])
         rcol = i;

   switch (PedDatTyp)
   {
       case INHERITANCE:
          printf (" Inheritance indicators at trait locus are to be read\n"
                  "    from integers %d (pat) and %d (mat)"
                  " in %s record\n\n", ic[0] + 1, ic[1] + 1, trait_source);
          break;
       case GENOTYPIC:
          printf (" Genotypic data at trait locus are to be read"
                  " from integer %d\n"
                  "    in %s record\n\n", ic[0] + 1, trait_source);
          break;
       case DISCRETE:
          printf (" Discrete data at trait locus are to be read"
                  " from integer %d\n"
                  "    in %s record\n\n", ic[0] + 1, trait_source);
          break;
       case DISCR_LIAB:
          printf (" Discrete data and liability classes at trait locus"
                  " are to be read\n"
                  "    from integers %d and %d in %s record\n\n",
                  ic[0] + 1, ic[1] + 1, trait_source);
          break;
       case DISCR_COVAR:
          printf (" Discrete data and covariate at trait locus are to be read\n"
                  "    from integers %d and %d in %s record\n\n",
                  ic[0] + 1, ic[1] + 1, trait_source);
          break;
       case QUANTITATIVE:
          printf (" Quantitative data at trait locus are to be read from"
                  " real %d\n" "    in %s record\n\n", rcol + 1, trait_source);
          break;
   }

   return;
}

/**
 * Prints the names of the selected markers, if input.
 */
void prnt_marknam (SIM_PTR p_sel)
{
   int   i, mw, nc, no;

   if (PrintLevel >= 5)
   {
      mw = intWid (1, p_sel->no_mark_map);
      nc = get_colfmt2 (LineWidth, 2, mw, 1, LocNamLen);

      if (p_sel->chrom_no)
         fprintf (stdout, " Marker names for chromosome %d:\n",
                  p_sel->chrom_no);
      else
         fprintf (stdout, " Marker names:\n");

      for (i = 0; i < NMarks; ++i)
      {
         if (!p_sel->p_mark)
            no = i + 1;
         else
            no = *(p_sel->p_mark + i);

         if (!(i % nc))
            fprintf (stdout, "\n");
         fprintf (stdout, "  %*d %*s", mw, no, LocNamLen, MarkNam[i]);
      }
      fprintf (stdout, "\n\n\n");
      fflush (stdout);
   }

   return;
}

/**
  prints the genotypes for the selected markers for observed individuals.
 */
void prnt_markdat (SIM_PTR p_sel)
{
   int   na, wa, nc, nm, ns;
   int   i, j, n;

   if (PrintLevel >= 5)
   {
      if (NObs == 0)
         return;

      if (p_sel->chrom_no)
         fprintf (stdout, " Marker data for chromosome %d:\n\n",
                  p_sel->chrom_no);
      else
         fprintf (stdout, " Marker data:\n\n");

      na = 0;
      if (MARK_ALLELE)
      {
         for (i = 1; i < NLoci; ++i)    /* Gets no. chars for printng alleles */
            if (NAll[i] > na)
               na = NAll[i];
      }
      wa = intWid (0, na);

      nc = 0;
      for (i = 0; i < NObs; ++i)
      {
         n = strlen (ObsNam[i]);        /* gets no. chars for printing names  */
         nc = n > nc ? n : nc;
      }

      nm = intWid (0, NMarks);          /* and no. chars for prntng mark no.s */
      ns = 2 + 2 * wa - nm;
      if (ns < 0)
         ns = 0;

      fprintf (stdout, "%*s ", nc, " ");
      for (i = 0; i < NMarks; ++i)
      {
         if (!p_sel->p_mark)
            fprintf (stdout, "%*sM%-*d", ns, " ", nm, i + 1);
         else
            fprintf (stdout, "%*sM%-*d", ns, " ", nm, *(p_sel->p_mark + i));
      }
      fprintf (stdout, "\n\n");

      for (i = 0; i < NObs; ++i)
      {
         fprintf (stdout, " %-*s", nc, ObsNam[i]);
         for (j = 0; j < NMarks; ++j)
         {
            if (!((j - 1) % 14) && j != 1)
               fprintf (stdout, " %*s\n", nc, " ");
            fprintf (stdout, "  %*d %*d", wa, ObsMark[i][j][0], wa,
                     ObsMark[i][j][1]);
         }
         fprintf (stdout, "\n");
      }
      fprintf (stdout, "\n\n");
      fflush (stdout);
   }

   return;
}

void prnt_probgam (void)
{
   int   c, s, i, n, nn, np, nd, nc = 0;

   if (NLocWndw)                        /* prints window info                 */
   {
      NWndw = NLoci + 1 - NLocWndw;
      fprintf (stdout, " Autozygosity will be scored for %d windows of "
               "%d loci\n\n\n", NWndw, NLocWndw);
   }

   for (c = 0; c < NScCmp; ++c)
   {
      for (s = 0; s < NScSet[c]; ++s)
      {
         for (i = 0; i < NPrGam[c][s]; ++i)
         {
            n = strlen (ProbNam[c][s][i]); /* gets no. chars in names for    */
            nc = n > nc ? n : nc;       /* printing                           */
         }
      }
   }

   nn = LineWidth / (nc + 4);
   nd = intWid (1, MxGam);
   np = (LineWidth - 20) / (nd + 2);

   fprintf (stdout, " Proband gametes:\n\n");

   if (!DoingAllPairs)
   {
      for (c = 0; c < NScCmp; ++c)
      {
         for (s = 0; s < NScSet[c]; ++s) /* For each component scoreset,      */
         {
            for (i = 0; i < NPrGam[c][s]; ++i) /* print proband gametes.      */
            {
               if (i == 0)
               {
                  fprintf (stdout, " Component %d, scoreset %d:\n\n ",
                           c + 1, ScSetNo[c][s]);
               }
               else if (!(i % nn))
                  fprintf (stdout, "\n ");
               fprintf (stdout, "  %*s", nc, ProbNam[c][s][i]);
               fprintf (stdout, " %d", ProbGam[c][s][i]);
               if (i == NPrGam[c][s] - 1)
                  fprintf (stdout, "\n\n");
            }

            if ((LM_AUTO) && NLocWndw && NPrGam[c][s] && NPtLab[c][s])
               for (i = 0; i < NPtLab[c][s]; ++i) /* Print pattern labels.    */
               {
                  if (i == 0)
                     fprintf (stdout, " Window patterns: ");
                  else if (!(i % np))
                     fprintf (stdout, "\n                  ");
                  fprintf (stdout, "  %*d", nd, PatLab[c][s][i]);
                  if (i == NPtLab[c][s] - 1)
                     fprintf (stdout, "\n\n");
               }

            if (NPrGam[c][s])
               fprintf (stdout, "\n");
         }
      }
   }
   else
   {
      fprintf (stdout, "    Proband gametes for all individual pairs will be");
      fprintf (stdout, " included in the analysis\n\n\n");
   }
   fflush (stdout);

   return;
}

void prnt_affind (void)
{
   fprintf (stdout, " Affected individuals to be identified by");
   if (TraitTyp == GENOTYPIC)
      fprintf (stdout, " genotypic");
   else if (TraitTyp == DISCRETE)
      fprintf (stdout, " discrete");
   fprintf (stdout, " data\n    for trait %d in the pedigree file\n\n",
            *(sim_traits->p_trait));

   fflush (stdout);

   return;
}

void prnt_trt_typ (void)
{
   fprintf (stdout, "\n");
   if (TraitTyp == INHERITANCE)
      fprintf (stdout, " Data are inheritance indicators at trait locus\n\n");
   else if (TraitTyp == GENOTYPIC)
      fprintf (stdout, " Trait data are genotypic\n\n");
   else if (TraitTyp == DISCRETE)
   {
      if (PedDatTyp == DISCR_LIAB)
         fprintf (stdout, " Trait data are discrete with liability classes"
                  " provided\n\n");
      else if (PedDatTyp == DISCR_COVAR)
         fprintf (stdout, " Trait data are discrete with covariates"
                  " provided\n\n");
      else
         fprintf (stdout, " Trait data are discrete\n\n");
   }
   else if (TraitTyp == QUANTITATIVE)
      fprintf (stdout, " Trait data are quantitative\n\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The det_output_order subroutine can be called when running as ibd_haplo in */
/* order to determine which set of messages about order of output need to be  */
/* printed to the stdout file.                                                */
/*                                                                            */
/* Note that since ibd_haplo does not run with more than one component, in    */
/* this subroutine the component subscript will always be 0.                  */
/*                                                                            */
/* The integer value returned by det_output_order will be one of:             */
/*                                                                            */
/* 0 -> There are both lexicographically ordered and Jacquard ordered output. */
/* 1 -> All output will be lexicographically ordered.                         */
/* 2 -> All output will be Jacquard ordered.                                  */
/*                                                                            */
/******************************************************************************/

int det_output_order (void)
{
   int   num_4gam_ss;                   /* Number of 4-gamete score sets.     */
   int   result;                        /* The intger value returned.         */

   int   c1 = 0;                        /* Component subscript (always == 0). */
   int   s1;                            /* Current score set subscript.       */

   if (DoingAllPairs == FALSE)
   {
      /* Count the number of score sets provided by the user that contain     */
      /* precisely four gametes.                                              */

      num_4gam_ss = 0;

      for (s1 = 0; s1 < NScSet[c1]; ++s1)
         if (NPrGam[c1][s1] == 4)
            ++num_4gam_ss;

      /* Now determine which of the three scenarios described above is        */
      /* actually the case.                                                   */

      if (Lexico == TRUE || num_4gam_ss == 0)
         result = 1;
      else if (num_4gam_ss == NScSet[c1])
         result = 2;
      else
         result = 0;
   }
   else                                 /* DoingAllPairs == TRUE              */
   {
      /* When ibd_haplo is "DoingAllPairs" it will always be the situation    */
      /* that there will be just one score set and it will have precisely     */
      /* four gametes.  Unfortunately the storage for the proband gamete      */
      /* statements will not have been allocated before this subroutine is    */
      /* called, so it is necessary to treat this as a special situation.     */
      /* When "DoingAllPairs" we need to determine how to set the result      */
      /* variable as if (num_4gam_ss == 1) and (NScSet[c1] == 1).  Here it    */
      /* would not be valid for the result variable to be set to 0.           */

      if (Lexico == TRUE)
         result = 1;
      else
         result = 2;
   }

   return result;
}

void prnt_auto_pars (void)
{
   if (IBDDROP)
   {
      fprintf (stdout, " Number of MC iterations requested:  %d\n\n", NIt);
      return;
   }

   if (IBD_HAPLO)
   {
      switch (PhaseofData)
      {
          case 0:
             fprintf (stdout, " The data will be analyzed as phased");
             fprintf (stdout, " data\n");
             break;
          case 1:
             fprintf (stdout, " The data will be analyzed as unphased");
             fprintf (stdout, " only for sets of whole individuals\n");
             break;
          case 2:
             fprintf (stdout, " The data will be analyzed as partially phased");
             fprintf (stdout, " only for sets of whole individuals\n");
             break;
      }

      fprintf (stdout, " The population kinship is %5.3f\n", PopKinship);
      fprintf (stdout, " The kinship change rate is %5.3f\n\n", KinshipChanRat);

      switch (StateTransMat)
      {
          case 0:
             fprintf (stdout, " The 2011");
             break;
          case 1:
             fprintf (stdout, " The 2009");
             break;
      }
      fprintf (stdout, " state transition matrix will be used\n");

      fprintf (stdout, " The transition matrix null fraction is %5.3f\n\n",
               TransMatNullFr);

      switch (det_output_order ())
      {
          case 0:
             fprintf (stdout, " In all cases except the 4-gamete case the");
             fprintf (stdout, " output is in lexicographical order\n");
             fprintf (stdout, " In the 4-gamete case the output is in");
             fprintf (stdout, " traditional (Jacquard) order\n\n");
             break;
          case 1:
             fprintf (stdout, " In all cases the output is in lexicographical");
             fprintf (stdout, " order\n\n");
             break;
          case 2:
             fprintf (stdout, " In all cases the output is in traditional");
             fprintf (stdout, " (Jacquard) order\n\n");
             break;
      }

      fprintf (stdout, "\n");

      return;
   }

   if (LM_MAP || LM_PVAL || LM_IBDTESTS || LM_AUTO)
   {
      if (LM_AUTO)
      {
         if (MMSampler == Y)
            fprintf (stdout, " Multiple meiosis sampler is to be used with\n");
         else if (MMSampler == N)
            fprintf (stdout, " Single meiosis sampler is to be used with\n");
         fprintf (stdout, " L-sampler probability set to %.3f\n\n", LSamplProb);
      }

      if (SmplSetup == INDEP_LOC)
         fprintf (stdout, " Locus-by-locus sampling to be used for setup\n\n");
      else if (SmplSetup == SEQIMP)
         fprintf (stdout, " Sequential imputation to be used for setup,"
                  " using %d samples\n\n", NSeqImp);

      if (LM_AUTO)
      {
         fprintf (stdout, " MCMC is being done ");
         if (GlobalMcmc)
            fprintf (stdout, "globally over all components\n\n");
         else
            fprintf (stdout, "for each component separately\n\n");
      }

      if (Scan)
      {
         fprintf (stdout, " Sampling is to be by scan");
         if (ScorIt == UNDEF)
            fprintf (stdout, "\n");
         else if (ScorIt == 1)
            fprintf (stdout, ", with scores estimated every scan\n");
         else
            fprintf (stdout, ", with scores estimated every %d scans\n",
                     ScorIt);
      }
      else
      {
         fprintf (stdout, " Sampling is to be by single step");
         if (ScorIt == UNDEF)
            fprintf (stdout, "\n");
         if (ScorIt == 1)
            fprintf (stdout, ", with scores estimated every step\n");
         else
            fprintf (stdout, ", with scores estimated every %d steps\n",
                     ScorIt);
      }
      fprintf (stdout, "\n");

      if (!LM_AUTO)
         fprintf (stdout, " L-sampler probability set to %.3f\n\n", LSamplProb);

      if (NBrnIt != UNDEF)
      {
         fprintf (stdout, " The number of MC burn-in iterations is %d\n",
                  NBrnIt);
         fprintf (stdout, "\n");
      }
   }

   if (!LM_MAP && NIt != UNDEF)
   {
      fprintf (stdout, " Number of MC iterations requested is %d\n", NIt);
      if (ChkIt != UNDEF)
         fprintf (stdout, " with progress checked every %d iterations\n",
                  ChkIt);
   }

   if (LM_MAP)
   {
      if (MapEstMod != MISTYP)
      {
         fprintf (stdout, "\n Map estimation model is without mistyping\n");
         fprintf (stdout, "\n Number of iterations to compute likelihood"
                  " ratio test (LRT) statistics is %d\n", NChiSqIt);
         fprintf (stdout, "\n %d map-estimation %s\n", RedoEst,
                  RedoEst == 1 ? "run is to be made" : "passes are to be made");

         if (MaximAlgo == HYBRID)
         {
            fprintf (stdout, "\n The hybrid algorithm, using both SA and"
                     " MCEM for maximization, is to be used\n");
            fprintf (stdout, " The number of curvature iterations for the"
                     " hybrid algorithm is %d\n", NHybCurvIt);
            fprintf (stdout, " The number of ascent iterations for hybrid"
                     " algorithm is %d\n", NHybAscentIt);
            fprintf (stdout, " The number of gradient iterations for SA"
                     " is %d\n", NSAGradIt);
            fprintf (stdout, " SA convergence for successive estimates is set"
                     " to %6.4f\n", SA_bound);
         }
         else if (MaximAlgo == MCEM)
            fprintf (stdout, "\n MCEM only is to be used for maximization\n");
      }
      else if (MapEstMod == MISTYP)
      {
         fprintf (stdout, "\n Map estimation model is with possible"
                  " mistyping\n");
         fprintf (stdout, " The initial genotyping error rate is %6.4f\n",
                  GenoErrRat);
      }

      fprintf (stdout, "\n The number of MC iterations for the first MCEM"
               " step is %d\n", NIt);
      fprintf (stdout, " Recombination fraction estimates are restricted"
               " to the interval [%.2f, %.2f]\n", RF_bound, .5 - RF_bound);
      fprintf (stdout, " EM convergence for successive estimates is set"
               " to %.4f\n", EM_bound);

      switch (MapTyp)
      {
          case SEX_SPEC:
             fprintf (stdout, " Gender-specific");
             break;
          case SEX_AVG:
             fprintf (stdout, " Gender-averaged");
             break;
          case BOTH:
             fprintf (stdout, " Both gender-averaged and -specific");
             break;
      }
      fprintf (stdout, " maps will be estimated\n");
   }

   if (LM_PVAL)
   {
      fprintf (stdout, "\n IBD statistics are to be computed\n");
      fprintf (stdout, "    measures:   ");
      fprintf (stdout, "MaxFGL - %s   Spairs - %s   Srobdom - %s\n",
               MaxFGLOn ? "yes" : "no", SpairsOn ? "yes" : "no",
               RobdomOn ? "yes" : "no");
   }
   else if (LM_IBDTESTS)
   {
      if (IBDStats)
      {
         fprintf (stdout, "\n IBD statistics are to be computed\n");
         fprintf (stdout, "    measures:   ");
         fprintf (stdout, "Spairs - %s   Srobdom - %s   Saffect - %s"
                  "   Slambda - %s\n",
                  SpairsOn ? "yes" : "no", RobdomOn ? "yes" : "no",
                  SAffUnaffOn ? "yes" : "no", LambdaOn ? "yes" : "no");
         fprintf (stdout, "    tests:      ");
         fprintf (stdout, "normal - %s   permutation - %s\n",
                  NormalEnable ? "yes" : "no", PermutEnable ? "yes" : "no");
         if (PermutEnable)
            fprintf (stdout, "\n    the number of permutations is %d\n",
                     NPermutIt);
         fprintf (stdout, "\n");
      }
      else if (LikRatStats)
      {
         fprintf (stdout,
                  "\n Likelihood-ratio statistics are to be computed\n");
         fprintf (stdout, "    measures:   ");
         fprintf (stdout, "delta - %s   lambda-p - %s\n",
                  DeltaOn ? "yes" : "no", Lambda_pOn ? "yes" : "no");
         fprintf (stdout, "    tests:      ");
         fprintf (stdout, "chi-squared - %s\n", ChiSqEnable ? "yes" : "no");
         if (Lambda_pOn)
         {
            fprintf (stdout, "\n    lambda_a, lambda_u interval:"
                     "  gridpoints - %d  spacing %.4f\n",
                     NumOfSegL, IntervalSegL);
            fprintf (stdout, "    p interval:                 "
                     "  gridpoints - %d  spacing %.4f\n",
                     NumOfSegP, IntervalSegP);
         }
      }
   }
   fprintf (stdout, "\n");

   return;
}

int get_marknam (char *nam, char *p_nm, int imrk)
{
   strncpy (nam, p_nm + (imrk - 1) * (MAXNAMLEN + 1), MAXNAMLEN + 1);

   return strlen (nam);
}

/* Returns a pointer to the trait/inheritance pedigree data node.             */

TRTPED_PTR find_trtped (int ti)
{
   TRTPED_PTR p;
   int   found;

   found = FALSE;                       /* Initialize to not found.           */

   if ((p = first_trtped_node))
   {
      if (TRAIT_SELCTD)
      {
         while (p->p_next
                && (p->no < ti || (p->no == ti && p->trait_typ != SS_TRLIST)))
            p = p->p_next;
         if (p->no == ti && p->trait_typ == SS_TRLIST)
            found = TRUE;
      }

      else if (INHER_SELCTD)
      {
         while (p->p_next
                && (p->no < ti || (p->no == ti && p->trait_typ != SS_INHER)))
            p = p->p_next;
         if (p->no == ti && p->trait_typ == SS_INHER)
            found = TRUE;
      }
   }
   if (found)
      return p;
   else
      return NULL;
}

/* Allocates the trait parameters data structures.                            */

void alloc_trait_pars (int n_trait_pars)
{
   P_trait_pars = calloc (n_trait_pars, sizeof (TRAITPARS));
   alloc_chk (WHERE, P_trait_pars);

   return;
}

/* The pass_trait subroutine should be called once for each selected DISCRETE */
/* or QUANTITATIVE trait.  It assumes that enough storage has been allocated  */
/* for all trait_pars_nodes required for each of these traits.  Storage will  */
/* be allocated for each needed tloc_pars_nodes.  This subroutine copies:     */
/*  the additive and residual variances for the trait to the trait_pars_node, */
/*  incomplete penetrances for the trait at this tloc to the tloc_pars_nodes, */
/*  and genotype means for the trait at this tloc to the tloc_pars_nodes.     */
/* This subroutine should work even if there are no additive or residual      */
/* variances, incomplete penetrancs or genotype means provided by the user.   */

void pass_trait (int t_no, int trait_ndx)
{
   TRAITPARS_PTR p_traitp;
   TLOCPARS_PTR p_tlocp;
   TRAITTLOC_PTR p_tt;

   VARI_PTR p_vari;
   INCO_PENE_PTR p_p;
   GENO_MEANS_PTR p_m;

   int   entry1, entry2;
   int   tloc_ndx;
   int   tl_no;
   int   i1;

   double mini;
   double maxi;

   p_traitp = P_trait_pars + trait_ndx;
   p_traitp->t_no = t_no;
   p_traitp->num_tlocs = NTLocs;

/* Store the additive and residual variances in the current trait_pars_node.  */

   if ((p_vari = first_addi_node))
   {
      while (p_vari->p_next && p_vari->no < t_no)
         p_vari = p_vari->p_next;
      if (p_vari->no == t_no)
      {
         p_traitp->addi_var = p_vari->vari;
         p_traitp->addi_term_found = sqrt (p_vari->vari);
         p_traitp->addi_term = sqrt (p_vari->vari / 2.0);
      }
   }

   if ((p_vari = first_resi_node))
   {
      while (p_vari->p_next && p_vari->no < t_no)
         p_vari = p_vari->p_next;
      if (p_vari->no == t_no)
      {
         p_traitp->resi_var = p_vari->vari;
         p_traitp->resi_term = sqrt (p_vari->vari);
      }
   }

/* Allocate any storage required for tloc_pars_node's for this trait.         */

   p_traitp->p_tloc_pars = calloc (p_traitp->num_tlocs, sizeof (TLOCPARS));
   alloc_chk (WHERE, p_traitp->p_tloc_pars);

   tloc_ndx = 0;

   entry1 = FALSE;
   while ((tl_no = nxt_tr_tl_no (&entry1, t_no)))
   {
      p_tlocp = p_traitp->p_tloc_pars + tloc_ndx;
      p_tlocp->tl_no = tl_no;

/* Store incomplete penetrances for each selected tloc in its tloc_pars_node. */

      if ((p_p = first_inco_pene_node))
      {
         while (p_p->p_next && (p_p->trait_no < t_no
                                || (p_p->trait_no == t_no
                                    && p_p->tloc_no < tl_no)))
            p_p = p_p->p_next;

         if (p_p->trait_no == t_no && p_p->tloc_no == tl_no)
         {
            if ((p_tlocp->num_penetrances = p_p->num_penetrances))
            {
               p_tlocp->penetrances =
                  malloc (p_tlocp->num_penetrances * sizeof (double));
               alloc_chk (WHERE, p_tlocp->penetrances);

               for (i1 = 0; i1 < p_tlocp->num_penetrances; ++i1)
                  p_tlocp->penetrances[i1] = p_p->penetrances[i1];
            }
         }
      }

/* Store the genotype means for each selected tloc in its tloc_pars_node.     */

      mini = 0;
      maxi = 0;

      if ((p_m = first_geno_means_node))
      {
         while (p_m->p_next && (p_m->trait_no < t_no
                                || (p_m->trait_no == t_no
                                    && p_m->tloc_no < tl_no)))
            p_m = p_m->p_next;

         if (p_m->trait_no == t_no && p_m->tloc_no == tl_no)
         {
            if ((p_tlocp->num_means = p_m->num_means))
            {
               p_tlocp->means = malloc (p_tlocp->num_means * sizeof (double));
               alloc_chk (WHERE, p_tlocp->means);

               for (i1 = 0; i1 < p_tlocp->num_means; ++i1)
                  p_tlocp->means[i1] = p_m->means[i1];

               mini = p_tlocp->means[0];
               for (i1 = 1; i1 < p_tlocp->num_means; ++i1)
                  if (p_tlocp->means[i1] < mini)
                     mini = p_tlocp->means[i1];

               maxi = p_tlocp->means[0];
               for (i1 = 1; i1 < p_tlocp->num_means; ++i1)
                  if (p_tlocp->means[i1] > maxi)
                     maxi = p_tlocp->means[i1];
            }

            if (p_m->joint)
            {
               entry2 = FALSE;
               if ((p_tt = nxt_tr_tl (&entry2)))
                  if (p_tt->trait_no == t_no && p_tt->selectd)
                     copy_quant_means (p_tt);
            }
         }
      }

      p_tlocp->dec = decWid (mini, maxi, SIGFIG);
      p_tlocp->wid = fldWid (mini, maxi, p_tlocp->dec) + 1;
      p_tlocp->ixG = get_tl_sim_loc_no (tl_no);
      p_tlocp->ixPh = trait_ndx;

      ++tloc_ndx;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_trait_pars subroutine should be called when it is time to free up */
/* the storage allocated by the alloc_trait_pars and pass_trait subroutines.  */
/*                                                                            */
/******************************************************************************/

void free_trait_pars (void)
{
   TLOCPARS_PTR p_tlocp;

   int   trait_ndx;

   if (P_trait_pars)
   {
      for (trait_ndx = 0; trait_ndx < NTraits; ++trait_ndx)
      {
         p_tlocp = P_trait_pars[trait_ndx].p_tloc_pars;
         if (p_tlocp)
         {
            if (p_tlocp->penetrances)
               free (p_tlocp->penetrances);

            if (p_tlocp->means)
               free (p_tlocp->means);

            free (p_tlocp);
         }
      }

      free (P_trait_pars);
      P_trait_pars = NULL;
   }

   return;
}
