/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/* These functions are called by the parameter statement functions, appearing
 * in proc_pars.c and named in parsedefs.h
 */

/* ---- check fncs:  check contents of parameter statements for errors ----- */

int chk_stat (void)
{                                       /* increments counter for this       */
   ++ppp->count;                        /* statement type                    */

   if (strchr (ppp->pcods, psx->pcod))  /* checks for this statement         */
      return TRUE;                      /* implemented for program           */
   else
   {
      err_typ = ERR_NOT;
      err_txt = psx->pnam;
      return FALSE;
   }
}

int chk_repeat (void)                   /* test for redundant statement      */
{
   if (ppp->count > 1)
   {
      err_typ = ERR_RED;
      return FALSE;
   }
   else
      return TRUE;
}

void chk_fil (char *f_typ)              /* tests for file name already given */
{                                       /* stores file name                  */
   pft = get_filent (f_typ);

   if (!pft)
      return;

   if (pft->command)
   {
      err_typ = ERR_CFO;
      sprintf (errnumb, "%s", pft->fdesc);
      err_txt = errnumb;
   }
   else
   {
      if (!chk_repeat ())
         free_fnam (pft);
      pft->fname = alloc_copy (s_par);
      pft->done = N;
      pft->rewrit = N;
   }
   pft->param = Y;

   return;
}

void set_rewrt (char *f_typ)
{
   pft = get_filent (f_typ);            /* sets overwrite file option */

   if (!pft)
      return;

   pft->rewrit = Y;

   return;
}

int chk_trait (void)                    /* tests for qualitative trait no.   */
                                        /* out of range                      */
{
   if (j_par[0] < 1 || j_par[0] > FMAX_REAL)
   {
      err_typ = ERD_TRO;
      return FALSE;
   }
   else
      return j_par[0];
}

int chk_trtped (int typ)                /* tests list of trait or inher no.'s */
{                                       /* vs. pos for ints, pairs, or reals  */
   int   m, i;

   m = 2;
   if (typ == INTDAT2 || typ == INHDAT2)
      ++m;

   if (m == 2)
   {
      if (nk_par % 2)
         err_typ = ERR_WTI;
   }
   else if (nk_par % 3)
      err_typ = ERR_WT2;
   if (err_typ)
      return FALSE;

   if (typ != REALDAT)                  /* checks record positions            */
   {
      for (i = nk_par / m; i < nk_par; ++i)
         if (j_par[i] < 1 || j_par[i] > FMAX_INT)
         {
            err_typ = ERR_TRI;
            return TRUE;
         }
   }
   else
      for (i = nk_par / m; i < nk_par; ++i)
         if (j_par[i] < 1 || j_par[i] > FMAX_REAL)
         {
            err_typ = ERR_TRK;
            return TRUE;
         }

   for (i = 0; i < nk_par / m; ++i)     /* checks trait numbers               */
      if (j_par[i] < 1 || j_par[i] > FMAX_INT)
      {
         err_typ = ERD_TRO;
         return TRUE;
      }
   return TRUE;
}

TRTPED *chk_trtped_node (int no, int tt) /* Tests for record position         */
                                        /* already given for this trait.      */
{
   TRTPED_PTR p_node;

   p_node = first_trtped_node;
   if (p_node)
   {
      while (p_node->p_next && (p_node->no < no
                                || (p_node->no == no
                                    && p_node->trait_typ != tt)))
         p_node = p_node->p_next;

      if (p_node->no == no && p_node->trait_typ == tt)
      {
         if (p_node->trait_typ != SS_INHER)
         {
            err_typ = ERR_TRP;
            return p_node;
         }
         else
         {
            err_typ = ERR_IRP;
            return p_node;
         }
      }
   }
   return NULL;
}

int chk_ff (int m)                      /* Tests for fixed effect number out  */
{                                       /* of range.                          */
   int   x1;

   /* Check for "set trait k1 effects k2 k3 ..." fixed effect number out of   */
   /* range.                                                                  */
   if (!m)
   {
      for (x1 = 1; x1 <= nk_par; ++x1)
         if (j_par[x1] < 1 || j_par[x1] > FMAX_FEF)
         {
            err_typ = ERR_FFO;
            return FALSE;
         }
      return TRUE;
   }

   /* Check for "start trait k1 effect k2 x1 x2 ..." fixed effect number out  */
   /* of range.                                                               */
   else if (j_par[1] < 1 || j_par[1] > FMAX_FEF)
   {
      err_typ = ERR_FFO;
      return FALSE;
   }
   else
      return TRUE;
}

int chk_levels_ff (void)                /* Tests for too many starting levels */
{                                       /* for a fixed effect.                */
   if (nx_par > LEVELS)
   {
      err_typ = ERR_LVM;
      return FALSE;
   }
   else
      return TRUE;
}

int chk_var (void)                      /* Tests for too many starting values */
{                                       /* for variances.                     */
   if (nx_par > psx->max_trait)
   {
      err_typ = ERR_VAM;
      return FALSE;
   }
   else
      return TRUE;
}

int chk_covar (void)                    /* Tests for too many starting values */
{                                       /* for covariances.                   */
   if (nx_par > psx->max_trait * (psx->max_trait - 1) / 2)
   {
      err_typ = ERR_COM;
      return FALSE;
   }
   else
      return TRUE;
}

int chk_covar_i (void)                  /* Tests for too many fit flags for   */
{                                       /* covariances.                       */
   if (nk_par > psx->max_trait * (psx->max_trait - 1) / 2)
   {
      err_typ = ERR_CFM;
      return FALSE;
   }
   else
      return TRUE;
}

/* Tests for chromosome number, number of markers, number of tlocs or number  */
/* of distances out of range.  Also tests for distances or recombination      */
/* fractions out of range.                                                    */

int chk_map (int chrom_num_included, int d, int m)
{
   int   i, ix = 0;
   int   err_ret = FALSE;

   if (chrom_num_included == CH)        /* Checks chromosome number.          */
   {
      if (j_par[0] < 1 || j_par[0] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ix++;
   }
   if (err_ret)
      return FALSE;

   if (m == MARK_LOC)                   /* checks map marker statement        */
      switch (d)
      {
          case RM:
             if (nx_par >= psx->max_mark)
             {
                err_typ = ERD_MRM;
                err_ret = TRUE;
             }
             break;
          case DM:
          case KDM:
             if (nx_par >= psx->max_mark)
             {
                err_typ = ERD_MDM;
                err_ret = TRUE;
             }
             break;
          case PM:
          case KPM:
             if (nx_par > psx->max_mark)
             {
                err_typ = ERD_MPM;
                err_ret = TRUE;
             }
             else if (nx_par == 1)
             {
                err_typ = ERD_O1M;
                err_ret = TRUE;
             }
             break;
          default:
             break;
      }

   if (err_ret)
      return FALSE;

   if (m == TLOC_LOC)                   /* checks map tloc statement          */
   {
      if (nk_par != 2 * nx_par)
      {
         err_typ = ERD_TMM;
         err_ret = TRUE;
      }
      else if (nk_par / 2 > MAXTL)
      {
         err_typ = ERD_MXT;
         err_ret = TRUE;
      }
      if (!err_ret)
         for (i = 0; i < nk_par / 2; i++)
            if (j_par[ix + i] < 1 || j_par[ix + i] > MAXTL)
            {
               err_typ = ERD_TLO;
               err_ret = TRUE;
            }
      if (!err_ret)
         for (i = 0; i < nk_par / 2; i++)
            if (j_par[ix + nk_par / 2 + i] < 0 ||
                j_par[ix + nk_par / 2 + i] > psx->max_mark)
            {
               err_typ = ERD_MKO;
               err_ret = TRUE;
            }
   }

   if (err_ret)
      return FALSE;

   for (i = 0; i < nx_par; i++)         /* checks fractions, distances        */
   {                                    /* or positions                       */
      switch (d)
      {
          case RM:
             if (d_par[i] < 0. || d_par[i] >= .5)
             {
                if (m == MARK_LOC)      /* tests for recomb fraction >= 0.    */
                   err_typ = ERD_MFO;   /* and < .5                           */
                else
                   err_typ = ERD_TFO;
                err_ret = TRUE;
             }
             break;
          case DM:
          case KDM:
             if (d_par[i] < 0. || d_par[i] > 500.)
             {
                if (m == MARK_LOC)      /* tests for distance >= 0.           */
                   err_typ = ERD_MDO;   /* and <= 500. cM                     */
                else
                   err_typ = ERD_TDO;
                err_ret = TRUE;
             }
             break;
          case PM:
          case KPM:
             if (d_par[i] < -500. || d_par[i] > 500.)
             {
                err_typ = ERD_MPO;      /* tests for position >= -500.        */
                err_ret = TRUE;         /* and <= 500 cM                      */
             }
             else if (i && d_par[i] < d_par[i - 1])
             {
                err_typ = ERD_MPD;      /* also tests for increasing order    */
                err_ret = TRUE;
             }
             break;

          default:
             break;
      }
   }
   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

/**
 * tests for tloc no.s out of range and for dup tloc no.s.
 */
int chk_map_unlnkd (void)
{
   int   i;

   if (nk_par > MAXTL)
   {
      err_typ = ERD_MXT;
      return FALSE;
   }

   for (i = 0; i < nk_par; i++)
      if (j_par[i] < 1 || j_par[i] > MAXTL)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }

   qsort (&(j_par[0]), nk_par, sizeof (int), qfunc_int);
   for (i = 0; i < nk_par - 1; i++)
      if (j_par[i] == j_par[i + 1])
      {
         err_typ = ERD_DTW;
         sprintf (errnumb, "%d", j_par[i]);
         err_txt = errnumb;
         return FALSE;
      }

   return TRUE;
}

/* Tests for chromosome number, tloc number or interval number out of range.  */
/* Then tests that all interval proportions, recombination fractions and map  */
/* distances are within the appropriate range.                                */

int chk_vmap (int chrom_num_included, int map_parameter_type)
{
   int   i1, ix = 0;
   int   err_ret = FALSE;

   if (chrom_num_included == CH)        /* Checks chromosome number.          */
   {
      if (j_par[0] < 1 || j_par[0] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }

   if (err_ret)
      return FALSE;

   if (j_par[ix] < 1 || j_par[ix] > MAXTL) /* Checks tloc number.             */
   {
      err_typ = ERD_TLO;
      err_ret = TRUE;
   }
   ++ix;

   if (err_ret)
      return FALSE;

/* Checks that interval numbers are within the correct range.                 */
   if (map_parameter_type == IM)
      for (i1 = 0; i1 < nk_par; ++i1, ++ix)
         if (j_par[ix] < 1 || j_par[ix] > psx->max_mark - 1)
         {
            err_typ = ERA_TIN;
            err_ret = TRUE;
         }

   if (err_ret)
      return FALSE;

/* Checks map proportions, recombination fractions or distances.              */
   for (i1 = 0; i1 < nx_par; i1++)
   {
      switch (map_parameter_type)
      {

/* Tests that proportions, either one or all intervals, are > 0.0 and < 1.0.  */
          case AM:
          case IM:
             if (d_par[i1] <= 0. || d_par[i1] >= 1.)
             {
                err_typ = ERA_TIO;
                err_ret = TRUE;
             }
             break;

/* Tests for recombination fraction >= 0.0 and < 0.5.                         */
          case RM:
          case BRM:
          case ERM:
          case XRM:
             if (d_par[i1] < 0. || d_par[i1] >= .5)
             {
                err_typ = ERA_TFO;
                err_ret = TRUE;
             }
             break;

/* Tests that distances are >= 0.0 and <= 500.0 cM.                           */
          case DM:
          case BDM:
          case EDM:
          case XDM:
          case KDM:
          case KBDM:
          case KEDM:
          case KXDM:
             if (d_par[i1] < 0. || d_par[i1] > 500.)
             {
                err_typ = ERA_TDO;
                err_ret = TRUE;
             }
             break;

#if 0
/* Tests that positions are >= -500.0 and <= 500.0 cM.                        */
          case PM:
          case KPM:
             if (d_par[i1] < -500. || d_par[i1] > 500.)
             {
                err_typ = ERA_TPO;
                err_ret = TRUE;
             }
             break;
#endif

          default:
             break;
      }
   }

/* Tests whether or not positions are in increasing order.                    */
   if (!err_ret)
      for (i1 = 1; i1 < nx_par; i1++)
         if (d_par[i1] <= d_par[i1 - 1])
         {

            switch (map_parameter_type)
            {
                case AM:
                case IM:
                   err_typ = ERA_TID;
                   break;

                case RM:
                case BRM:
                case ERM:
                case XRM:
                   err_typ = ERA_TFD;
                   break;

                case DM:
                case BDM:
                case EDM:
                case XDM:
                case KDM:
                case KBDM:
                case KEDM:
                case KXDM:
                   err_typ = ERA_TDD;
                   break;

                case PM:
                case KPM:
                   break;

                default:
                   break;
            }
         }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

/* Tests for chromosome number, tloc number(s) or at marker numbers out of    */
/* range.                                                                     */

int chk_jmap (int chrom_num_included, int map_parameter_type)
{
   int   i1, ix = 0, no_tlocs;
   int   err_ret = FALSE;

   if (chrom_num_included == CH)        /* Checks chromosome number.          */
   {
      if (j_par[0] < 1 || j_par[0] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }

   if (err_ret)
      return FALSE;

/* Checks that the potentially joint tloc numbers are within range.           */

   no_tlocs = 2;                        /* Assume joint tloc numbers.         */
   if (map_parameter_type == MN)
      no_tlocs = 1;

   for (i1 = 0; i1 < no_tlocs; i1++)
   {
      if (j_par[ix] < 1 || j_par[ix] > MAXTL)
      {
         err_typ = ERD_TLO;
         err_ret = TRUE;
      }
      ++ix;
   }

   if (err_ret)
      return FALSE;

#if 0
   if (map_parameter_type == JPM || map_parameter_type == JKPM)
   {
/* Checks for pairs of joint positions.                                       */
      if (nx_par % 2)
      {
         err_typ = ERA_WJP;
         err_ret = TRUE;
      }

      for (i1 = 0; i1 < nx_par; i1++)   /* Checks map positions.              */
         if (d_par[i1] < -500. || d_par[i1] > 500.)
         {
            err_typ = ERA_TPO;
            err_ret = TRUE;
         }
   }
#endif

/* Checks that at marker numbers are within the correct range.                */
   if (map_parameter_type == MN || map_parameter_type == JMN)
      for (i1 = 0; i1 < nk_par; ++i1, ++ix)
         if (j_par[ix] < 1 || j_par[ix] > psx->max_mark)
         {
            err_typ = ERD_MKO;
            err_ret = TRUE;
         }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

/**
 * tests for statement already given for this chromosome no. and gender
 * for specified locus type (for markers or tlocs).
 * returns map node, if found
 */
MAP_PTR chk_map_node (int chrom_num_included, int m, int sx)
{
   MAP_PTR p_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   if (m == MARK_LOC)
      p_node = first_mark_map_node[sx];
   else
      p_node = first_tloc_map_node[sx];

   if (p_node)
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERD_RMP;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

/**
 * tests for statement already given for unlinked tlocs.
 * returns map node, if found
 */
UMAP_PTR chk_map_unl_node (void)
{
   UMAP_PTR p_node;

   if ((p_node = unlnkd_tloc_map_node))
      ;                                 /* prev warning for repeated statemnt */
   return p_node;
}

/* Tests for any external or interval map statement already given for this    */
/* chromosome, tloc and gender.  Then returns the vmap node, if found.        */

VMAP *chk_vmap_node (int chrom_num_included, int sx)
{
   VMAP_PTR p_node;

   int   chrom_no;
   int   ix = 0;

   if (chrom_num_included == CH)
      chrom_no = j_par[ix++];
   else
      chrom_no = 0;

   p_node = first_tloc_vmap_node[sx];

   if (p_node)
   {
      while (p_node->p_next             /* Checks for vmap node.              */
             && (p_node->chrom_no < chrom_no
                 || (p_node->chrom_no == chrom_no && p_node->tlno < j_par[ix])))
         p_node = p_node->p_next;

      if (p_node->chrom_no == chrom_no && p_node->tlno == j_par[ix])
         return p_node;
   }
   return NULL;
}

int chk_vext (XMAP_PTR p_x, int chrom_num_included)
/*
 * checks for external or beginning or ending distances & recomb fractions
 */
{
   if (p_x)
   {
      if (chrom_num_included == CH)
         err_typ = ERA_OXC;
      else
         err_typ = ERA_OXP;
   }

   if (err_typ)
      return TRUE;
   else
      return FALSE;
}

/* Checks for list of proportions already present for specified interval      */
/* (0 = all).                                                                 */

IMAP_PTR chk_vint (VMAP_PTR p_v, int chrom_num_included, int d, int l)
{
   IMAP_PTR p_i;

   p_i = p_v->p_intv;

   if (!p_i)
      return NULL;

   if (d == AM)                         /* Checks for all intervals present.  */
   {
      if (p_i->intv_no == 0)
      {
         if (chrom_num_included == CH)
            err_typ = ERA_OAC;
         else
            err_typ = ERA_OAP;
      }
      else
         p_i = NULL;
   }

   if (d == IM)                         /* Checks for this interval present.  */
   {
      while (p_i && p_i->p_next && p_i->intv_no < l)
         p_i = p_i->p_next;
      if (p_i && p_i->intv_no == l)
      {
         if (chrom_num_included == CH)
            err_typ = ERA_OIC;
         else
            err_typ = ERA_OIP;
      }
      else
         p_i = NULL;
   }

   return p_i;                          /* returns ptr to interval (or NULL)  */
}

/* Tests for chromosome number, marker numbers, tloc numbers and number of    */
/* allele frequencies within range.  Also checks sum of allele frequencies.   */

int chk_locus (int chrom_num_included, int md)
{
   int   i, ix = 0;

   if (chrom_num_included == CH)
   {
      if (j_par[ix] < 1 || j_par[ix] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ix++;
   }

   for (i = 0; i < nk_par; i++)
   {
      if (md == MARK_LOC)
      {
         if (j_par[ix] < 1 || j_par[ix] > psx->max_mark)
         {
            err_typ = ERD_MKO;
            return FALSE;
         }
      }
      else if (j_par[ix] < 1 || j_par[ix] > MAXTL)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }
      ix++;
   }

   if (md == MARK_LOC)
   {
      if (nx_par < 2 || nx_par > pmx->max_mk_all)
      {
         err_typ = ERD_MLO;
         return FALSE;
      }
   }
   else
   {
      if (nx_par != 2)
      {
         err_typ = ERD_TLA;
         return FALSE;
      }
   }

   for (i = 0; i < nx_par; i++)
      if (d_par[i] < 0)
      {
         err_typ = ERD_AFO;
         return FALSE;
      }

   return TRUE;
}

/* Checks for allele frequencies already given for this locus.                */

LOCUS *chk_locus_node (int no, int chrom_num_included, int md)
{
   LOCUS_PTR p_node;
   int   n = 0;

   if (md == MARK_LOC)                  /* Marker allele frequencies.         */
   {
      if (chrom_num_included == CH)
         n = j_par[0];

      if (no > 0)
         p_node = first_mark_locus_node;
      else
         p_node = all_mark_locus_node;
   }
   else
   {                                    /* tloc allele frequencies            */
      n = NO_CH_NO;
      p_node = first_tloc_locus_node;
   }

   if (p_node)
   {
      while (p_node->p_next
             && (p_node->chrom_no < n
                 || (p_node->chrom_no == n && p_node->no < no)))
         p_node = p_node->p_next;
      if (p_node->chrom_no == n && p_node->no == no)
      {
         if (md == MARK_LOC)
            err_typ = ERD_MAF;          /* overrides previous allele freqs    */
         else                           /* .. for marker                      */
            err_typ = ERD_TAF;          /* .. or for tloc                     */
         sprintf (errnumb, "%d", p_node->no);
         err_txt = errnumb;
         return p_node;
      }
   }

   return NULL;                         /* no match for id                    */
}

int chk_locus_norm (void)               /* tests for unnormalized freqs       */
{
   int   i;
   double sum = 0.;

   for (i = 0; i < nx_par; i++)         /* this test is done last             */
      sum += d_par[i];

   if (unnorm_frq (sum))
      err_typ = ERD_SMA;

   return TRUE;
}

/* Tests trait, tloc numbers in range, and for number of penetrances = 3.     */

int chk_inco_pene (void)
{
   int   i1;

   if (nk_par % 2)
   {
      err_typ = ERD_WNM;
      return FALSE;
   }

   for (i1 = 0; i1 < nk_par / 2; i1++)
      if (j_par[i1] < 1 || j_par[i1] > MAXTR)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }

   for (i1 = nk_par / 2; i1 < nk_par; i1++)
      if (j_par[i1] < 1 || j_par[i1] > MAXTL)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }

   if (nx_par != 3)
   {
      err_typ = ERA_NPV;
      return FALSE;
   }

   for (i1 = 0; i1 < nx_par; i1++)
      if (d_par[i1] < 0.0 || d_par[i1] > 1.0)
      {
         err_typ = ERA_RPV;
         return FALSE;
      }
      else if (d_par[1] > d_par[2] || d_par[0] > d_par[1])
         err_typ = ERA_OPV;

   return TRUE;
}

/* Tests for valid number of liability classes, for correct number of         */
/* penetrances and that all penetrances are between 0.0 and 1.0.              */

int chk_liab_classes (void)
{
   int   needed_reals;
   int   ix;

   if (j_par[0] < 1 || j_par[0] > 24)
   {
      err_typ = ERD_NLC;
      return FALSE;
   }

   else
   {
      if (j_par[0] == 1)
         err_typ = ERD_1LC;
   }

   NumbLiabClasses = j_par[0];
   needed_reals = NumbLiabClasses * 3;

   if (needed_reals != nx_par)
   {
      err_typ = ERD_NPV;
      sprintf (errnumb, "%d", needed_reals);
      err_txt = errnumb;
      return FALSE;
   }

   for (ix = 0; ix < nx_par; ++ix)
   {
      if (d_par[ix] <= 0.0 || d_par[ix] >= 1.0)
      {
         err_typ = ERD_RPV;
         return FALSE;
      }
   }

   return TRUE;
}

/* Tests trait, tloc numbers in range, and for number of means = 3.           */

int chk_geno_means (void)
{
   int   i1;

   if (nk_par % 2)
   {
      err_typ = ERD_WNM;
      return FALSE;
   }

   for (i1 = 0; i1 < nk_par / 2; i1++)
      if (j_par[i1] < 1 || j_par[i1] > MAXTR)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }

   for (i1 = nk_par / 2; i1 < nk_par; i1++)
      if (j_par[i1] < 1 || j_par[i1] > MAXTL)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }

   if (nx_par != 3)
   {
      err_typ = ERD_TGM;
      return FALSE;
   }

   return TRUE;
}

/* Tests trait, tloc numbers in range, and for number of means = 9.           */

int chk_joint_geno_means (void)
{
   int   i1;

   if (ni_par != 1 || nk_par != 2)
   {
      err_typ = ERD_JNM;
      return FALSE;
   }

   if (j_par[0] < 1 || j_par[0] > MAXTR)
   {
      err_typ = ERD_TRO;
      return FALSE;
   }

   for (i1 = 1; i1 < 3; i1++)
      if (j_par[i1] < 1 || j_par[i1] > MAXTL)
      {
         err_typ = ERD_TLO;
         return FALSE;
      }

   if (nx_par != 9)
   {
      err_typ = ERD_JGM;
      return FALSE;
   }

   return TRUE;
}

/* Tests for incomplete penetrances already given for the trait at the tloc.  */
/* This subroutine is called whenever a "set traits ... for tlocs ...         */
/* incomplete penetrances X1 X2 X3" statement is being parsed.                */

INCO_PENE_PTR chk_inco_pene_node (int t_no, int tl_no)
{
   INCO_PENE_PTR p_node;

   p_node = first_inco_pene_node;
   while (p_node && p_node->p_next
          && (p_node->trait_no < t_no
              || (p_node->trait_no == t_no && p_node->tloc_no < tl_no)))
      p_node = p_node->p_next;

   if (p_node && p_node->trait_no == t_no && p_node->tloc_no == tl_no)
   {
      err_typ = ERD_TPR;
      return p_node;
   }

   return NULL;
}

/* Tests for trait genotype means already given for the trait at the tloc.    */
/* This subroutine is called whenever a "set traits ... for tlocs ...         */
/* genotype means X1 X2 X3" statement is being parsed.                        */

GENO_MEANS_PTR chk_geno_means_node (int t_no, int tl_no)
{
   GENO_MEANS_PTR p_node;

   p_node = first_geno_means_node;
   while (p_node && p_node->p_next
          && (p_node->trait_no < t_no
              || (p_node->trait_no == t_no && p_node->tloc_no < tl_no)))
      p_node = p_node->p_next;

   if (p_node && p_node->trait_no == t_no && p_node->tloc_no == tl_no)
   {
      if (GxGInteraction == FALSE)
      {
         err_typ = ERD_TGR;
         return p_node;
      }
      else
         err_typ = ERD_JGR;
   }

   return NULL;
}

/* Tests for trait genotype means already given for the trait at the tloc.    */
/* This subroutine is called whenever a "set trait K1 for tlocs L1 L2 joint   */
/* genotype means X11 ... X33" statement is being parsed.                     */

GENO_MEANS_PTR chk_joint_geno_means_node (int t_no, int tl_no)
{
   GENO_MEANS_PTR p_node;

   p_node = first_geno_means_node;
   while (p_node && p_node->p_next
          && (p_node->trait_no < t_no
              || (p_node->trait_no == t_no && p_node->tloc_no < tl_no)))
      p_node = p_node->p_next;

   if (p_node && p_node->trait_no == t_no && p_node->tloc_no == tl_no)
   {
      err_typ = ERD_TGR;
      return p_node;
   }

   return NULL;
}

int chk_vari (void)                     /* tests for trait no. in range,     */
{
   int   i;

   for (i = 0; i < nk_par; i++)
      if (j_par[i] < 1 || j_par[i] > MAXTR)
      {
         err_typ = ERD_TRO;
         return FALSE;
      }
   if (d_par[0] < 0.)
   {
      err_typ = ERD_VMN;
      return FALSE;
   }
   return TRUE;
}

VARI *chk_vari_node (int no, int v)     /* tests for variance already given  */
                                       /* for this trait                     */
{
   VARI_PTR p_node;

   if (v == AV)
      p_node = first_addi_node;
   else
      p_node = first_resi_node;

   if (p_node)
   {
      while (p_node->p_next && p_node->no < no)
         p_node = p_node->p_next;

      if (p_node->no == no)
      {
         if (v == AV)
            err_typ = ERD_AVR;
         else
            err_typ = ERD_RVR;
         return p_node;
      }
   }
   return NULL;
}

/* Tests for chromosome number out of range and for number of markers out of  */
/* range.  Tests for each pair of alleles greater than 0, if observed, or 0   */
/* if unobserved.  Checks for duplicate names.                                */

int chk_mrkdat (int chrom_num_included)
{
   int   i, j, k, ix = 0;
   int   err_ret = FALSE;

   if (chrom_num_included == CH)        /* Checks chromosome number.          */
   {
      if (j_par[0] < 1 || j_par[0] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (!err_ret)                        /* checks no. markers                */
      if (j_par[ix] < 1 || j_par[ix] > psx->max_mark)
      {
         err_typ = ERA_MNO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   if (!err_ret)                        /* checks for alleles for each indiv */
   {                                    /* either unobserved, or all > 0     */
      j = ix + 1;
      for (i = 0; i < nm_par; ++i)
      {
         for (k = 0; k < j_par[ix]; ++k, j += 2)
         {
            if (j_par[j] < 0 || j_par[j + 1] < 0)
            {
               err_typ = ERA_PAL;
               sprintf (errnumb, "%s", m_par[i]);
               err_txt = errnumb;
               err_ret = TRUE;
            }
            else if (!j_par[j] != !j_par[j + 1])
            {
               err_typ = ERA_UAL;
               sprintf (errnumb, "%s", m_par[i]);
               err_txt = errnumb;
               err_ret = TRUE;
            }
         }
      }
   }
   if (!err_ret)                        /* checks for duplicate names        */
      for (i = 1; i < nm_par; ++i)
         for (j = 0; j < i; ++j)
            if (!(strncmp (m_par[i], m_par[j], MAXNAMLEN + 1)))
            {
               err_typ = ERA_DIN;
               sprintf (errnumb, "%s", m_par[j]);
               err_txt = errnumb;
               err_ret = TRUE;
            }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

/* Tests for statement already given for this chromosome number.              */

MARKDAT *chk_mrkdat_node (int chrom_num_included)
{
   MARKDAT_PTR p_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   if ((p_node = first_markdat_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RMD;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

/* Tests for chromosome number out of range and for number of markers out of  */
/* range.  Then checks for duplicate names.                                   */

int chk_mrknam (int chrom_num_included)
{
   int   i, j, ix = 0;
   int   err_ret = FALSE;

   if (chrom_num_included == CH)        /* Checks chromosome number.          */
   {
      if (j_par[0] < 1 || j_par[0] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (!err_ret)                        /* checks no. markers                */
      if (nm_par < 1 || nm_par > psx->max_mark)
      {
         err_typ = ERA_MNO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   if (!err_ret)                        /* checks for duplicate names        */
      for (i = 1; i < nm_par; ++i)
         for (j = 0; j < i; ++j)
            if (!(strncmp (m_par[i], m_par[j], MAXNAMLEN + 1)))
            {
               err_typ = ERA_DMN;
               sprintf (errnumb, "%s", m_par[j]);
               err_txt = errnumb;
               err_ret = TRUE;
            }

   if (err_ret)
      return FALSE;
   else
      return TRUE;
}

/* Tests for statement already given for this chromosome number.              */

MARKNAM *chk_mrknam_node (int chrom_num_included)
{
   MARKNAM_PTR p_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   if ((p_node = first_marknam_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RMN;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_prbgam (int cp, int ss)
/*                                       * tests for component no. out of
 *                                         range
 *                                         checks name & gamete (proband
 *                                         gametes cannot be duplicated
 *                                         within the set, & gametes must be
 *                                         0 (maternal) or 1 (paternal))      */
{
   int   i, j, ix = 0;
   int   err_ret = FALSE;

   if (cp == CMP)                       /* Checks component number.           */
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (ss == SC)                        /* Checks scoreset number.            */
   {
      if (j_par[ix] < 1)
      {
         err_typ = ERA_SCS;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (nm_par < 2)                      /* checks min no. proband gametes     */
   {
      err_typ = ERA_PGL;
      err_ret = TRUE;
   }
   else if (nm_par > MAXGAM)            /* checks max no. proband gametes     */
   {
      err_typ = ERA_PGX;
      sprintf (errnumb, "%d", MAXGAM);
      err_txt = errnumb;
      err_ret = TRUE;
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nm_par; ++i)
   {
      if (j_par[ix + i] != 0 && j_par[ix + i] != 1) /* checks haplotypes     */
      {
         err_typ = ERA_BGM;
         sprintf (errnumb, "%d", j_par[ix + i]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      for (j = 0; j < i; ++j)
         if (!(strcmp (m_par[i], m_par[j])) && j_par[ix + i] == j_par[ix + j])
         {                              /* checks for dup proband gametes  */
            err_typ = ERA_DPG;
            sprintf (errnumb, "%s %d", m_par[j], j_par[ix + i]);
            err_txt = errnumb;
            err_ret = TRUE;
         }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

/* Tests for statement already given for this component and scoreset.         */

PROBGAM *chk_prbgam_node (int cp, int ss)
{
   PROBGAM_PTR p_node;

   int   c, s, ix = 0;

   if (cp == CMP)
      c = j_par[ix++];
   else
      c = 1;
   if (ss == SC)
      s = j_par[ix++];
   else
      s = 1;

   if ((p_node = first_probgam_node))
   {
      while (p_node->p_next &&
             (p_node->component < c
              || (p_node->component == c && p_node->scoreset < s)))
         p_node = p_node->p_next;

      if (p_node->component == c && p_node->scoreset == s)
      {
         err_typ = ERA_OSC;
         return p_node;
      }
   }
   return NULL;
}

int chk_indnam (int cp)                 /* tests for component no. out of     */
                                        /* range                              */
                                        /* checks names for duplication       */
{
   int   i, j;
   int   err_ret = FALSE;

   if (cp == CMP)                       /* Checks component number.           */
   {
      if (j_par[0] < 1 || j_par[0] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[0]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nm_par; ++i)
   {
      for (j = 0; j < i; ++j)
         if (!(strcmp (m_par[i], m_par[j])))
         {                              /* checks for dup named individuals  */
            err_typ = ERA_DAI;
            sprintf (errnumb, "%s", m_par[j]);
            err_txt = errnumb;
            err_ret = TRUE;
         }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

/* Tests for statement already given for this component.                      */

INDNAM *chk_indnam_node (int cp, int nt)
{
   INDNAM_PTR p_node;

   int   c, ix = 0;

   if (cp == CMP)
      c = j_par[ix++];
   else
      c = 1;

   if (nt == DISPIND)
      p_node = first_dispnam_node;
   if (p_node)
   {
      while (p_node->p_next && p_node->component < c)
         p_node = p_node->p_next;

      if (p_node->component == c)
      {
         err_typ = ERA_OAI;
         return p_node;
      }
   }
   return NULL;
}

/* Tests for chromosome number or locus window size out of range.             */

int chk_wind (int chrom_num_included)
{
   int   ix = 0;

   if (chrom_num_included == CH)
   {
      if (j_par[ix] < 1 || j_par[ix] > psx->max_chrom)
      {
         err_typ = ERD_CHO;
         return FALSE;
      }
      ix++;
   }
   if (j_par[ix] < 2)
   {
      err_typ = ERR_MNW;
      return FALSE;
   }
   else if (j_par[ix] > psx->max_mark + 1)
   {
      err_typ = ERR_MXW;
      return FALSE;
   }

   if (err_typ)
      return FALSE;
   else
      return TRUE;
}

/* Tests for statement already given for this chromosome number.              */

LOCWND *chk_wind_node (int chrom_num_included)
{
   LOCWND_PTR p_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   if ((p_node = first_locwind_node))
   {
      while (p_node->p_next && p_node->chrom_no < n)
         p_node = p_node->p_next;

      if (p_node->chrom_no == n)
      {
         if (n == 0)
            err_typ = ERA_RLW;
         else
            err_typ = ERD_RCS;
         return p_node;
      }
   }
   return NULL;
}

int chk_wndpat (int cp, int ss)         /* tests for component no., scoreset  */
                                        /* no. and no. of patterns out of     */
                                        /* range ???                          */
                                        /* tests for pattern label > 0        */
{
   int   i, ix = 0;
   int   err_ret = FALSE;

   if (cp == CMP)                       /* Checks component number.           */
   {
      if (j_par[ix] < 1 || j_par[ix] > psx->max_chrom)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   if (ss == SC)                        /* Checks scoreset number.            */
   {
      if (j_par[ix] < 1)
      {
         err_typ = ERA_SCS;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
      ++ix;
   }
   if (err_ret)
      return FALSE;

   for (i = 0; i < nk_par; ++i, ++ix)   /* checks pattern labels out of range */
   {
      if (j_par[ix] < 0 || j_par[ix] > CDgg[MAXGAM])
      {
         err_typ = ERA_WPO;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
         err_ret = TRUE;
      }
   }
   if (err_ret)
      return FALSE;

   return TRUE;
}

/* Tests for statement already given for this component and scoreset.         */

WNDPAT *chk_wndpat_node (int cp, int ss)
{
   WNDPAT_PTR p_node;

   int   c, s, ix = 0;

   if (cp == CMP)
      c = j_par[ix++];
   else
      c = 1;
   if (ss == SC)
      s = j_par[ix++];
   else
      s = 1;

   if ((p_node = first_windpat_node))
   {
      while (p_node->p_next &&
             (p_node->component < c
              || (p_node->component == c && p_node->scoreset < s)))
         p_node = p_node->p_next;

      if (p_node->component == c && p_node->scoreset == s)
      {
         err_typ = ERA_OSP;
         return p_node;
      }
   }
   return NULL;
}

int chk_kinshp (int cmp, int m)         /* tests for component out of range   */
{                                       /* checks for pairs of names          */
   int   ix = 0;

   if (cmp == CMP)
   {
      if (j_par[ix] < 1 || j_par[ix] > pmx->max_n_cmp)
      {
         err_typ = ERA_CMP;
         sprintf (errnumb, "%d", j_par[ix]);
         err_txt = errnumb;
      }
      ix++;
   }
   if (err_typ)
      return FALSE;

   if (nm_par % m)
   {
      sprintf (errnumb, "pairs");
      err_txt = errnumb;
      err_typ = ERK_KNN;
   }

   if (err_typ)
      return FALSE;
   return TRUE;
}

KINS *chk_kinshp_node (int cmp, int kc) /* tests for statement already        */
{                                       /* given for this component           */
   KINS *p_node;

   int   n = 0;

   if (cmp == CMP)
      n = j_par[0];

   if ((kc == KINSHP && (p_node = first_kins_node)) /* determines node type  */
       || (kc == INBRD && (p_node = first_inbreed_node))
       || (kc == INBRD2 && (p_node = first_2loc_node)))
   {
      while (p_node->p_next && p_node->component < n)
         p_node = p_node->p_next;

      if (p_node->component == n)
      {
         if (n == 0)
            err_typ = ERR_RED;
         else
            err_typ = ERK_RCP;
         return p_node;
      }
   }
   return NULL;
}

int chk_reco_frq (void)
{
   int   i;

   for (i = 0; i < nx_par; ++i)
      if (d_par[i] < 0.)
         err_typ = ERK_NRF;
      else if (d_par[i] > .5)
         err_typ = ERK_XRF;

   if (err_typ)
      return FALSE;
   return TRUE;
}

int chk_pseudo (void)
{
   int   i;

/*
   for (i = 0; i < nx_par; ++i)
   if (d_par[i] == 0)
   {
   d_par[i] = DIFF;
   err_typ = ERA_PPW;
   }
 */
   for (i = 0; i < nx_par; ++i)
      if (d_par[i] < 0)
         err_typ = ERA_PPN;

   if (err_typ)
      return FALSE;
   return TRUE;
}

void chk_seeds (void)
{
   if (u_par[0] % 2 == 0)
   {
      u_par[0] |= 1;
      err_typ = ERD_SEE;
   }
   if (u_par[1] == 0)
   {
      u_par[1] = 1;
      err_typ = ERD_SEE;
   }
   return;
}

int unnorm_frq (double q_sum)
{
   if (q_sum < Q_SUMMN || q_sum > Q_SUMMX) /* checks unnormlzd */
      return TRUE;
   else
      return FALSE;
}

int unnorm_frq_mx (int norm_flg, double q_sum)
{
   if (!norm_flg && q_sum > Q_SUMMX)    /* checks sum > upper limit           */
      return TRUE;
   else
      return FALSE;
}

/* ---------------- free fncs:  frees list nodes or arrays ----------------- */

void free_dist (MAP_PTR p_node)         /* frees list of map distances        */
{
   if (p_node && p_node->p_dist)
   {
      free (p_node->p_dist);
      p_node->p_dist = NULL;
      p_node->no_loci = 0;
   }
   return;
}

void free_udist (UMAP_PTR p_node)       /* frees list of map distances        */
{
   if (p_node && p_node->p_udist)
   {
      free (p_node->p_udist);
      p_node->p_udist = NULL;
      p_node->no_loci = 0;
   }
   return;
}

void free_vdist (XMAP_PTR p_x)          /* frees list of external dists       */
{
   if (p_x && p_x->no_pos)
   {
      free (p_x->p_dist);
      p_x->p_dist = NULL;
      free (p_x->p_rf);
      p_x->p_rf = NULL;
      p_x->no_pos = 0;
   }
   return;
}

void free_vprop (IMAP_PTR p_i)          /* frees list of interval proportions */
{

   if (p_i && p_i->p_prop)
      free (p_i->p_prop);
   p_i->p_prop = NULL;
   p_i->no_prop = nx_par;

   return;
}

void free_freq (LOCUS_PTR p_locus)      /* if there are no loci for the      */
                                        /* frequencies, frees list           */
{
   if ((p_locus->p_freq_node->no_loci -= 1) == 0)
   {                                    /* decrement & test no. loci         */
      free (p_locus->p_freq_node->p_freq);
      free (p_locus->p_freq_node);
   }
   p_locus->p_freq_node = NULL;
   p_locus->no_alleles = 0;

   return;
}

void free_mrkdat_list (MARKDAT_PTR p_node)
{
   if (p_node->p_obs)
   {
      free (p_node->p_obs);
      p_node->p_obs = NULL;
   }
   if (p_node->p_alls != NULL)
   {
      free (p_node->p_alls);
      p_node->p_alls = NULL;
   }
   return;
}

void free_mrknam_list (MARKNAM_PTR p_node)
{
   if (p_node->p_nams)
   {
      free (p_node->p_nams);
      p_node->p_nams = NULL;
   }
   return;
}

void free_prbgam_list (PROBGAM_PTR p_node)
{
   if (p_node->p_pglist)
   {
      free (p_node->p_pglist);
      p_node->p_pglist = NULL;
   }
   p_node->no_pgams = 0;

   return;
}

void free_indnam_list (INDNAM_PTR p_node)
{
   if (p_node->p_namlist)
   {
      free (p_node->p_namlist);
      p_node->p_namlist = NULL;
   }
   p_node->no_nams = 0;

   return;
}

void free_wndpat_set (WNDPAT_PTR p_node)
{
   if (p_node->p_pats)
   {
      free (p_node->p_pats);
      p_node->p_pats = NULL;
   }
   p_node->no_pats = 0;

   return;
}

void free_nams_list (KINS * p_node)
{
   if (p_node->p_nams != NULL)
   {
      free (p_node->p_nams);
      p_node->p_nams = NULL;
   }
   return;
}

void free_reco_frq (double **r)
{
   if (*r)
   {
      free (*r);
      *r = NULL;
   }
   return;
}

/* --------------- alloc fncs:  malloc list nodes or arrays ---------------- */

TRTPED *alloc_trtped_node (int no)      /* allocates the traitped node        */
                                        /* inserts in list in numeric order   */
                                        /* (note:  may be dup nos' when both  */
                                        /* traits and inher are present)      */
{
   TRTPED_PTR new_node, cur_node;

   new_node = malloc (sizeof (TRTPED));
   alloc_chk (WHERE, new_node);

   if (!first_trtped_node || first_trtped_node->no > no)
   {
      new_node->p_next = first_trtped_node;
      first_trtped_node = new_node;
   }
   else
   {
      cur_node = first_trtped_node;
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

/**
 * allocates the map node;  inserts in list in numeric order
 */

MAP_PTR add_map_node (int chrom_num_included, int m, int sx)
{
   MAP_PTR new_node, cur_node;
   MAP_PTR *first_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   new_node = malloc (sizeof (MAP));
   alloc_chk (WHERE, new_node);

   if (m == MARK_LOC)                   /* Checks for marker or tloc map.     */
      first_node = &first_mark_map_node[sx];
   else
      first_node = &first_tloc_map_node[sx];

   if (!*first_node || (*first_node)->chrom_no > n)
   {
      new_node->p_next = *first_node;   /* puts at start of list              */
      *first_node = new_node;           /* (may be only node)                 */
   }
   else
   {                                    /* inserts node in list or adds       */
      cur_node = *first_node;           /* to end of list                     */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   new_node->chrom_no = n;

   return new_node;
}

/**
 * allocates the map node for unlinked tlocs.
 */

UMAP_PTR add_map_unl_node (void)
{
/* note:  most items in the map structure do not apply to unlinked loci       */

   unlnkd_tloc_map_node = malloc (sizeof (UMAP));
   alloc_chk (WHERE, unlnkd_tloc_map_node);

   unlnkd_tloc_map_node->no_loci = 0;
   unlnkd_tloc_map_node->p_udist = NULL;

   return unlnkd_tloc_map_node;
}

/**
 * allocates the test position map node;  inserts in list in chromosome
 * and tloc order.
 */
VMAP *add_vmap_node (int chrom_num_included, int sx)
{
   VMAP_PTR new_node, cur_node, first_node;

   int   chrom_no;
   int   ix = 0;

   if (chrom_num_included == CH)
      chrom_no = j_par[ix++];
   else
      chrom_no = 0;

   new_node = malloc (sizeof (VMAP));
   alloc_chk (WHERE, new_node);

   first_node = first_tloc_vmap_node[sx];

   if (!first_node                      /* Checks start of list.              */
       || first_node->chrom_no > chrom_no
       || (first_node->chrom_no == chrom_no && first_node->tlno > j_par[ix]))
   {
      new_node->p_next = first_node;    /* Puts at start (may be only node).  */
      first_tloc_vmap_node[sx] = new_node;
   }
   else
   {
      cur_node = first_node;            /* Inserts node in list or at end.    */

      while (cur_node->p_next
             && (cur_node->p_next->chrom_no < chrom_no
                 || (cur_node->p_next->chrom_no == chrom_no
                     && cur_node->p_next->tlno < j_par[ix])))
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }

   new_node->chrom_no = chrom_no;
   new_node->tlno = j_par[ix++];
   new_node->selectd = FALSE;
   new_node->p_vprntmap[FEM] = NULL;
   new_node->p_vprntmap[MAL] = NULL;
   new_node->p_vprntmap[AVG] = NULL;
   new_node->jointly = FALSE;
   new_node->joined_tlno = j_par[ix];   /* Save second tloc just in case.     */
   new_node->p_joined = NULL;
   new_node->p_pos = NULL;
   new_node->p_beg = NULL;
   new_node->p_end = NULL;
   new_node->p_intv = NULL;
   new_node->no_marker_nums = 0;
   new_node->p_marker_nums = NULL;

   return new_node;
}

/**
 * allocates the dist node.
 */
void alloc_dist (MAP_PTR p_node, int m)
{
   if (m == MARK_LOC)
      p_node->p_dist = malloc ((nx_par + 1) * sizeof (DIST));
   else
      p_node->p_dist = malloc (nx_par * sizeof (DIST));
   alloc_chk (WHERE, p_node->p_dist);

   return;
}

/**
 * Allocates the Udist_node for unlinked tlocs.
 */
void alloc_unl_dist (UMAP_PTR p_node)
{
   p_node->p_udist = malloc (nk_par * sizeof (UDIST));
   alloc_chk (WHERE, p_node->p_udist);

   return;
}

XMAP_PTR alloc_vext (void)
/*                                       * allocates the beg or ending map    */
{
   XMAP_PTR p;

   p = malloc (sizeof (XMAP));
   alloc_chk (WHERE, p);

   return p;
}

IMAP_PTR add_vint (VMAP_PTR p_v, int l)
/*                                       * adds  the interval prop list       */
{
   IMAP_PTR p_new, p_i;

   p_new = malloc (sizeof (IMAP));
   alloc_chk (WHERE, p_new);

   p_i = p_v->p_intv;

   if (!p_i)
   {
      p_v->p_intv = p_new;              /* adds as only interval              */
      p_new->p_next = NULL;
   }
   else if (p_i->intv_no > l)
   {
      p_new->p_next = p_v->p_intv;      /* puts at start of list of intervals */
      p_v->p_intv = p_new;
   }
   else
   {
      while (p_i->p_next && p_i->p_next->intv_no < l)
         p_i = p_i->p_next;
      p_new->p_next = p_i->p_next;      /* adds to list of intervals          */
      p_i->p_next = p_new;
   }
   p_new->intv_no = l;
   p_new->no_prop = nx_par;

   return p_new;
}

void alloc_vdist (XMAP_PTR p_x)
/*                                       * allocates extern dist & rf lists   */
{
   p_x->p_dist = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_x->p_dist);
   p_x->p_rf = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_x->p_rf);

   return;
}

void alloc_vprop (IMAP_PTR p_i)
/*                                       * allocates internal prop list       */
{
   p_i->p_prop = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, p_i->p_prop);

   return;
}

/* Allocates storage for an array of at marker numbers.                       */

void alloc_vmarknums (VMAP_PTR p_vmap)
{
   p_vmap->no_marker_nums = nk_par;

   p_vmap->p_marker_nums = malloc (p_vmap->no_marker_nums * sizeof (int));
   alloc_chk (WHERE, p_vmap->p_marker_nums);

   return;
}

/*
******>>>>>>>>>>>>>>>>>><<<<<<<<<<<<<<<<<<<<<<
*/
FREQ_PTR alloc_frq (void)               /* allocates frequency header        */
{                                       /* and array                         */
   return (alloc_frqx (nx_par));
}

FREQ_PTR alloc_dflt_frq (void)          /* allocates frequency header        */
{                                       /* and array for default values      */
   return (alloc_frqx (2));
}

FREQ_PTR alloc_frqx (int n_alleles)     /* allocates frequency header        */
{                                       /* and array                         */
   FREQ_PTR p_f;

   p_f = malloc (sizeof (FREQ));
   alloc_chk (WHERE, p_f);
   p_f->p_freq = malloc (n_alleles * sizeof (double));
   alloc_chk (WHERE, p_f->p_freq);
   return p_f;
}

LOCUS_PTR alloc_locus_node (int no, int chrom_num_included, int md)
{
   int   cno = 0;

   if (md == MARK_LOC)                  /* Marker allele frequencies.         */
   {
      if (chrom_num_included == CH)
         cno = j_par[0];
   }
   else                                 /* Tloc allele frequencies.           */
      cno = NO_CH_NO;

   return (alloc_locus_nodex (cno, no, md));
}

LOCUS_PTR alloc_dflt_locus_node (int cno, int no, int md)
{
   return (alloc_locus_nodex (cno, no, md));
}

LOCUS *alloc_locus_nodex (int chrom, int no, int md)
{
   LOCUS_PTR new_node, cur_node;
   LOCUS_PTR *first_node;

   if (md == MARK_LOC)
   {
      if (no > 0)
         first_node = &first_mark_locus_node;
      else
         first_node = &all_mark_locus_node;
   }
   else
      first_node = &first_tloc_locus_node;

   new_node = malloc (sizeof (LOCUS));
   alloc_chk (WHERE, new_node);

   if (!*first_node
       || (*first_node)->chrom_no > chrom
       || ((*first_node)->chrom_no == chrom && (*first_node)->no > no))
   {
      new_node->p_next = *first_node;   /* puts at start of list             */
      *first_node = new_node;
   }
   else
   {
      cur_node = *first_node;

      while (cur_node->p_next
             && (cur_node->p_next->chrom_no < chrom
                 || (cur_node->p_next->chrom_no == chrom
                     && cur_node->p_next->no < no)))
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

/* Allocate an inco_pene_node for this trait and tloc.  Then insert it in the */
/* list in increasing trait/tloc order.                                       */

INCO_PENE_PTR alloc_inco_pene_node (int t_no, int tl_no)
{
   INCO_PENE_PTR new_node, cur_node;

   new_node = malloc (sizeof (INCO_PENE));
   alloc_chk (WHERE, new_node);

   new_node->trait_no = t_no;
   new_node->tloc_no = tl_no;
   new_node->num_penetrances = 0;
   new_node->penetrances = NULL;
   new_node->p_next = NULL;

   if (!first_inco_pene_node || first_inco_pene_node->trait_no > t_no
       || (first_inco_pene_node->trait_no == t_no
           && first_inco_pene_node->tloc_no >= tl_no))
   {
      new_node->p_next = first_inco_pene_node;
      first_inco_pene_node = new_node;
   }
   else
   {
      cur_node = first_inco_pene_node;
      while (cur_node->p_next
             && (cur_node->p_next->trait_no < t_no
                 || (cur_node->p_next->trait_no == t_no
                     && cur_node->p_next->tloc_no < tl_no)))
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }

   return new_node;
}

/* Allocates the array of incomplete penetrance data for the trait and tloc.  */

void alloc_inco_pene_array (INCO_PENE_PTR p_node)
{
   p_node->num_penetrances = nx_par;

   p_node->penetrances = malloc (p_node->num_penetrances * sizeof (double));
   alloc_chk (WHERE, p_node->penetrances);

   return;
}

/* Free the array of incomplete penetrance data for this trait and tloc.      */

void free_inco_pene_array (INCO_PENE_PTR p_p)
{
   if (p_p && p_p->penetrances)
   {
      p_p->num_penetrances = 0;
      free (p_p->penetrances);
      p_p->penetrances = NULL;
   }

   return;
}

/* Store the liability classes penetrance data.                               */

void stor_liab_classes (void)
{
   int   ix;
   int   p1;
   int   p2;

   for (ix = 0; ix < nx_par; ++ix)
   {
      p1 = ix / 3;
      p2 = ix % 3;
      PenMat[p1][p2] = d_par[ix];
   }

   return;
}

/* Stores the array of incomplete penetrance data for the trait and tloc.     */

void stor_inco_pene_array (INCO_PENE_PTR p_node)
{
   double *p_p;
   int   ix = 0;

   for (p_p = p_node->penetrances; p_p < p_node->penetrances +
        p_node->num_penetrances; ++p_p)
      *p_p = d_par[ix++];

   return;
}

/* Allocate a geno_means_node for this trait and tloc.  Then insert it in the */
/* list in increasing trait/tloc order.                                       */

GENO_MEANS_PTR alloc_geno_means_node (int t_no, int tl_no)
{
   GENO_MEANS_PTR new_node, cur_node;

   new_node = malloc (sizeof (GENO_MEANS));
   alloc_chk (WHERE, new_node);

   new_node->trait_no = t_no;
   new_node->tloc_no = tl_no;
   new_node->joint = FALSE;
   new_node->joined_tlno = 0;
   new_node->p_joined = NULL;
   new_node->num_means = 0;
   new_node->means = NULL;
   new_node->p_next = NULL;

   if (!first_geno_means_node || first_geno_means_node->trait_no > t_no
       || (first_geno_means_node->trait_no == t_no
           && first_geno_means_node->tloc_no >= tl_no))
   {
      new_node->p_next = first_geno_means_node;
      first_geno_means_node = new_node;
   }
   else
   {
      cur_node = first_geno_means_node;
      while (cur_node->p_next
             && (cur_node->p_next->trait_no < t_no
                 || (cur_node->p_next->trait_no == t_no
                     && cur_node->p_next->tloc_no < tl_no)))
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }

   return new_node;
}

/* Allocates the array of genotype means data for the trait and tloc.         */

void alloc_geno_means_array (GENO_MEANS_PTR p_node)
{
   p_node->num_means = nx_par;

   p_node->means = malloc (p_node->num_means * sizeof (double));
   alloc_chk (WHERE, p_node->means);

   return;
}

/* Free the array of genotype means data for this trait and tloc.             */

void free_geno_means_array (GENO_MEANS_PTR p_m)
{
   if (p_m && p_m->means)
   {
      p_m->num_means = 0;
      free (p_m->means);
      p_m->means = NULL;
   }

   return;
}

/* Stores the array of genotype means data for the trait and tloc.            */

void stor_geno_means_array (GENO_MEANS_PTR p_node)
{
   double *p_m;
   int   ix = 0;

   for (p_m = p_node->means; p_m < p_node->means + p_node->num_means; ++p_m)
      *p_m = d_par[ix++];

   return;
}

/* Allocates the variance node.  Then inserts it in the list in numeric trait */
/* order.                                                                     */

VARI *alloc_vari_node (int no, int v)
{
   VARI_PTR new_node, cur_node;
   VARI_PTR *first_node;

   if (v == AV)
      first_node = &first_addi_node;
   else
      first_node = &first_resi_node;

   new_node = malloc (sizeof (VARI));
   alloc_chk (WHERE, new_node);

   if (!*first_node || (*first_node)->no > no)
   {
      new_node->p_next = *first_node;
      *first_node = new_node;
   }
   else
   {
      cur_node = *first_node;
      while (cur_node->p_next && cur_node->p_next->no < no)
         cur_node = cur_node->p_next;
      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

/* Allocates marker data node.  Then inserts it in the list in numeric order. */

MARKDAT *alloc_mrkdat_node (int chrom_num_included)
{
   MARKDAT_PTR new_node, cur_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   new_node = malloc (sizeof (MARKDAT));
   alloc_chk (WHERE, new_node);

   if (!first_markdat_node || first_markdat_node->chrom_no > n)
   {
      new_node->p_next = first_markdat_node; /* puts at start of list       */
      first_markdat_node = new_node;    /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list or           */
      cur_node = first_markdat_node;    /* adds to end of list               */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_mrkdat_list (MARKDAT_PTR p_node) /* allocates marker data array */
{
   p_node->p_obs = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_obs);

   p_node->p_alls = malloc (nk_par * sizeof (int));
   alloc_chk (WHERE, p_node->p_alls);

   return;
}

/* Allocates marker name node.  Then inserts it in the list in numeric order. */
MARKNAM *alloc_mrknam_node (int chrom_num_included)
{
   MARKNAM_PTR new_node, cur_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   new_node = malloc (sizeof (MARKNAM));
   alloc_chk (WHERE, new_node);

   if (!first_marknam_node || first_marknam_node->chrom_no > n)
   {
      new_node->p_next = first_marknam_node; /* puts at start of list       */
      first_marknam_node = new_node;    /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list or           */
      cur_node = first_marknam_node;    /* adds to end of list               */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_mrknam_list (MARKNAM_PTR p_node) /* allocates marker name array */
{
   p_node->p_nams = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_nams);

   return;
}

PROBGAM *alloc_prbgam_node (int cp, int ss) /* allocates proband gamete node  */
                                        /* inserts in list in order of        */
                                        /* component and scoreset             */
{
   PROBGAM_PTR new_node, cur_node;

   int   c, s, ix = 0;

   if (cp == CMP)
      c = j_par[ix++];
   else
      c = 1;
   if (ss == SC)
      s = j_par[ix++];
   else
      s = 1;

   new_node = malloc (sizeof (PROBGAM));
   alloc_chk (WHERE, new_node);

   if (!first_probgam_node || first_probgam_node->component > c
       || (first_probgam_node->component == c
           && first_probgam_node->scoreset > s))
   {
      new_node->p_next = first_probgam_node; /* puts at start of list        */
      first_probgam_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_probgam_node;    /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && (cur_node->p_next->component < c
                 || (cur_node->p_next->component == c
                     && cur_node->p_next->scoreset < s)))
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_prbgam_list (PROBGAM_PTR p_node) /* allocates the proband        */
{                                       /* gamete array                       */

   p_node->p_pglist = malloc ((nm_par) * sizeof (PROBANDS));
   alloc_chk (WHERE, p_node->p_pglist);

   return;
}

INDNAM *alloc_indnam_node (int cp, int nt) /* allocates named individs node   */
                                        /* (affected or display individuals)  */
                                        /* inserts in list in order of compnt */
{
   INDNAM_PTR *pp_first;
   INDNAM_PTR new_node, cur_node;

   int   c;

   if (cp == CMP)
      c = j_par[0];
   else
      c = 1;

   new_node = malloc (sizeof (INDNAM));
   alloc_chk (WHERE, new_node);

   if (nt == AFFIND)
      pp_first = &first_affnam_node;
   else if (nt == DISPIND)
      pp_first = &first_dispnam_node;

   if (!*pp_first || (*pp_first)->component > c)
   {
      new_node->p_next = *pp_first;     /* puts at start of list              */
      *pp_first = new_node;             /* (may be only node)                 */
   }
   else
   {
      cur_node = *pp_first;             /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && cur_node->p_next->component < c)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_indnam_list (INDNAM_PTR p_node) /* allocates the named individuals */
{                                       /* list                               */

   p_node->p_namlist = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_namlist);

   return;
}

/* Allocates window locus node.  Then inserts it in the list in numeric order. */

LOCWND *alloc_wind_node (int chrom_num_included)
{
   LOCWND_PTR new_node, cur_node;

   int   n = 0;

   if (chrom_num_included == CH)
      n = j_par[0];

   new_node = malloc (sizeof (LOCWND));
   alloc_chk (WHERE, new_node);

   if (!first_locwind_node || first_locwind_node->chrom_no > n)
   {
      new_node->p_next = first_locwind_node; /* puts at start of list         */
      first_locwind_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_locwind_node;    /* inserts node in list               */
      /*                                 * or adds to end of list             */
      while (cur_node->p_next && cur_node->p_next->chrom_no < n)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

WNDPAT *alloc_wndpat_node (int cp, int ss) /* allocates window pattern node   */
{                                       /* inserts in list in order of        */
   WNDPAT_PTR new_node, cur_node;       /* component and scoreset             */

   int   c, s, ix = 0;

   if (cp == CMP)
      c = j_par[ix++];
   else
      c = 1;
   if (ss == SC)
      s = j_par[ix++];
   else
      s = 1;

   new_node = malloc (sizeof (WNDPAT));
   alloc_chk (WHERE, new_node);

   if (!first_windpat_node || first_windpat_node->component > c
       || (first_windpat_node->component == c
           && first_windpat_node->scoreset > s))
   {
      new_node->p_next = first_windpat_node; /* puts at start of list         */
      first_windpat_node = new_node;    /* (may be only node)                 */
   }
   else
   {
      cur_node = first_windpat_node;    /* inserts node in list               */
      while (cur_node->p_next           /* or adds to end of list             */
             && (cur_node->p_next->component < c
                 || (cur_node->p_next->component == c
                     && cur_node->p_next->scoreset < s)))
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_wndpat_set (WNDPAT_PTR p_node) /* allocates window pattern set     */
{
   p_node->p_pats = malloc ((nk_par) * sizeof (int));
   alloc_chk (WHERE, p_node->p_pats);

   return;
}

KINS *alloc_kinshp_node (int cmp, int kc) /* allocates kinship node.          */
                                        /* inserts in list after last set     */
                                        /* for the chromosome                 */
{
   KINS *new_node, *cur_node;
   KINS **p_first;
   int   cp = 1;

   if (cmp == CMP)
      cp = j_par[0];

   new_node = malloc (sizeof (KINS));
   alloc_chk (WHERE, new_node);

   if (kc == KINSHP)
      p_first = &first_kins_node;
   else if (kc == INBRD)
      p_first = &first_inbreed_node;
   else if (kc == INBRD2)
      p_first = &first_2loc_node;

   if (!*p_first || (*p_first)->component > cp)
   {
      new_node->p_next = *p_first;      /* puts at start of list             */
      *p_first = new_node;              /* (may be only node)                */
   }
   else
   {                                    /* inserts node in list              */
      cur_node = *p_first;              /* or adds to end of list            */
      while (cur_node->p_next && cur_node->p_next->component <= cp)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }
   return new_node;
}

void alloc_nams_list (KINS * p_node)    /* allocates the list of pedigree    */
{                                       /* member names                      */

   p_node->p_nams = malloc (nm_par * (MAXNAMLEN + 1) * sizeof (char));
   alloc_chk (WHERE, p_node->p_nams);

   return;
}

double *alloc_reco_frq (void)           /* allocates list of recomb freqs    */
{
   double *r;

   r = malloc (nx_par * sizeof (double));
   alloc_chk (WHERE, r);

   return r;
}

/* -------------- stor fncs:  stores data in nodes or arrays ---------------- */

void stor_trtped_pos (int typ, int no, int tt, int *pos, TRTPED_PTR p_node)
{
   p_node->no = no;
   p_node->dat_typ = typ;
   p_node->trait_typ = tt;
   p_node->pos[0] = pos[0];
   if (typ == INTDAT2 || typ == INHDAT2)
      p_node->pos[1] = pos[1];

   return;
}

/**
 * stores no. of loci; stores an array of tloc or marker no's,
 * preceding marker no's, probabilities & distances in the distance structure.
 * for tlocs, sorts the array for increasing position.
 */
void stor_map (MAP_PTR p_node, int chrom_num_included, int d,
               int map_parameter_type)
{
   int   i;
   int   ix = 0, jx = 0;
   DIST_PTR p_r;

   if (chrom_num_included == CH)
      ++ix;

   if (map_parameter_type == MARK_LOC)
   {
      if (d != PM && d != KPM)
         p_node->no_loci = nx_par + 1;
      else
         p_node->no_loci = nx_par;
   }
   else
      p_node->no_loci = nx_par;

   p_r = p_node->p_dist;
   for (i = 0; i < p_node->no_loci; ++i)
   {
      p_r->selectd = 0;
      if (map_parameter_type == MARK_LOC)
      {
         p_r->no = i + 1;
         p_r->map_mark = i;
      }
      else
      {
         p_r->no = j_par[ix];
         p_r->map_mark = j_par[ix + p_node->no_loci];
         ++ix;
      }

      if (i == 0 && map_parameter_type == MARK_LOC)
      {
         if (d == PM || d == KPM)
            p_r->pos = d_par[jx];
         else
            p_r->pos = 0.0;
         p_r->dist = 0.0;
         p_r->prob = 0.5;
      }
      else
      {
         switch (d)
         {
             case RM:
                p_r->pos = 0.0;
                p_r->dist = -.5 * log (1. - 2. * d_par[jx]) * 100.;
                p_r->prob = d_par[jx];
                break;
             case DM:
                p_r->pos = 0.0;
                p_r->dist = d_par[jx];
                p_r->prob = .5 * (1. - exp (-2. * d_par[jx] / 100.));
                break;
             case PM:
                p_r->pos = d_par[jx + 1];
                p_r->dist = d_par[jx + 1] - d_par[jx];
                p_r->prob = .5 *
                   (1. - exp (-2. * (d_par[jx + 1] - d_par[jx]) / 100.));
                break;
             case KDM:
                p_r->pos = 0.0;
                p_r->prob = .5 *
                   (((exp (4. * d_par[jx] / 100.)) - 1.)
                    / ((exp (4. * d_par[jx] / 100.)) + 1.));
                p_r->dist = -.5 * log (1. - 2. * p_r->prob) * 100.;
                break;
             case KPM:
                p_r->pos = d_par[jx + 1];
                p_r->dist = d_par[jx + 1] - d_par[jx];
                p_r->prob = .5 *
                   (((exp (4. * p_r->dist / 100.)) - 1.)
                    / ((exp (4. * p_r->dist / 100.)) + 1.));
                p_r->dist = -.5 * log (1. - 2. * p_r->prob) * 100.;
                break;
             default:
                break;
         }
         ++jx;
      }
      ++p_r;
   }

   if (map_parameter_type == TLOC_LOC && d != PM && d != KPM)
      qsort (p_node->p_dist, p_node->no_loci, sizeof (DIST), compar_mkid_d);

   return;
}

/**
 * stores no. of unlinked tlocs; stores list of tlocs no's in dist list.
 */
void stor_map_unl (UMAP_PTR p_node)
{
   int   i;
   UDIST_PTR p_u;

   p_node->no_loci = nk_par;

   p_u = p_node->p_udist;
   for (i = 0; i < p_node->no_loci; ++i)
   {
      p_u->no = j_par[i];
      p_u->selectd = 0;
      ++p_u;
   }

   return;
}

/* Stores the list of external distances and recombination fractions.         */

void stor_vdist (XMAP_PTR p_x, int d, int qsf ())
{
   int   i;

   p_x->no_pos = nx_par;

   qsort (d_par, nx_par, sizeof (double), qsf); /* sorts: incr or decr order  */

   if (XDM == d || BDM == d || EDM == d) /* converts Haldane dist to rec frct */
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_dist[i] = d_par[i];
         p_x->p_rf[i] = dist2rf (d_par[i]);
      }
   else if (XRM == d || BRM == d || ERM == d) /* converts rec fract to dist   */
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_dist[i] = rf2dist (d_par[i]);
         p_x->p_rf[i] = d_par[i];
      }
   if (KXDM == d || KBDM == d || KEDM == d) /* converts Kosambi dist to rec f */
      for (i = 0; i < nx_par; ++i)
      {
         p_x->p_rf[i] = distK2rf (d_par[i]);
         p_x->p_dist[i] = rf2dist (p_x->p_rf[i]);
      }
   return;
}

/* Stores the array of interval proportions for the specified interval.       */
/* There is no reason I am able to determine for requiring the user to enter  */
/* the proportions in any specific order, so qsort is being called to put     */
/* them into ascending order.                                                 */

void stor_vprop (IMAP_PTR p_i)
{
   int   ix;

   for (ix = 0; ix < nx_par; ++ix)
      p_i->p_prop[ix] = d_par[ix];

   qsort (p_i->p_prop, nx_par, sizeof (double), qfunc_doub);

   return;
}

/*  Stores the array of marker numbers for the potentially joint test tlocs.  */
/*  For now I don't believe that the marker numbers have already been sorted  */
/*  into ascending marker number order, so qsort is being called.  This may   */
/*  need to be checked out later.                                             */
/*  The at marker numbers should not be sorted; their order is important.     */

void stor_vmarknums (VMAP_PTR p_vmap, int chrom_num_included,
                     int map_parameter_type)
{
   int   ix = 1;                        /* There is always at least one tloc number. */
   int  *p_m;

   if (chrom_num_included == CH)
      ++ix;                             /* Skip the chromosome number.        */
   if (map_parameter_type == JMN)
      ++ix;                             /* Skip the other tloc number.        */

   for (p_m = p_vmap->p_marker_nums;
        p_m < p_vmap->p_marker_nums + p_vmap->no_marker_nums; ++p_m)
      *p_m = j_par[ix++];

   return;
}

int compar_mkid_d (const void *p1, const void *p2) /* called by qsort         */
                                        /* for distance array                 */
                                        /* orders on marker id & dist         */
{
   DIST_PTR pd1;
   DIST_PTR pd2;

   pd1 = (DIST_PTR const) p1;
   pd2 = (DIST_PTR const) p2;

   if (pd1->map_mark < pd2->map_mark
       || (pd1->map_mark == pd2->map_mark && pd1->dist < pd2->dist))
      return -1;

   else if (pd1->map_mark == pd2->map_mark && pd1->dist == pd2->dist)
      return 0;
   else
      return 1;
}

void stor_frq (FREQ_PTR p_f)            /* Stores the number of loci         */
                                        /* for these alleles in the header   */
                                        /* and stores an array of alleles.   */
{
   int   i1;

   p_f->no_loci = max (nk_par, 1);
   p_f->sum = 0.0;

   for (i1 = 0; i1 < nx_par; ++i1)
   {
      *(p_f->p_freq + i1) = d_par[i1];
      p_f->sum += d_par[i1];
   }

   return;
}

void stor_dflt_frq (FREQ_PTR p_f)       /* stores the default allele freqs   */
{
   p_f->no_loci = 1;
   p_f->sum = 1.0;
   *(p_f->p_freq) = .5;
   *(p_f->p_freq + 1) = .5;

   return;
}

/* Stores data in the locus node.                                             */

void stor_locus_data (int no, LOCUS_PTR p_l, FREQ_PTR p_f,
                      int chrom_num_included, int md)
{
   if (md == MARK_LOC)
   {
      if (chrom_num_included == CH)
         p_l->chrom_no = j_par[0];
      else
         p_l->chrom_no = 0;
   }
   else
      p_l->chrom_no = NO_CH_NO;

   p_l->no = no;
   p_l->no_alleles = nx_par;
   p_l->p_freq_node = p_f;

   return;
}

void stor_dflt_locus_data (int chro, int no, LOCUS_PTR p_l, FREQ_PTR p_f)
                                        /*stores locus node data             */
                                        /*for default allele frequencies     */
{
   p_l->chrom_no = chro;
   p_l->no = no;
   p_l->no_alleles = 2;
   p_l->p_freq_node = p_f;

   return;
}

void stor_dflt_geno_mean_dat (int t_no, int tl_no, GENO_MEANS_PTR p_node)
{
   int   i1;

   for (i1 = 0; i1 < p_node->num_means; i1++)
      p_node->means[i1] = 0.0;

   return;
}

void stor_vari_data (int i, VARI_PTR p_node)
{
   p_node->no = j_par[i];
   p_node->vari = d_par[0];

   return;
}

void stor_dflt_vari_data (int no, VARI_PTR p_node)
{
   p_node->no = no;
   p_node->vari = 0.;

   return;
}

/* Stores data in the marker data node.  Also stores the list of observed     */
/* individuals and the lists of marker alleles.                               */

void stor_mrkdat (MARKDAT_PTR p_md, int chrom_num_included)
{
   int   i, j;
   char *p_n;
   int  *p_a;
   int   ix = 0;

   if (chrom_num_included == CH)
   {
      p_md->chrom_no = j_par[ix];
      ++ix;
   }
   else
      p_md->chrom_no = 0;

   p_md->no_obs = nm_par;
   p_md->no_marks = nk_par / (2 * nm_par);
   ++ix;

   p_n = p_md->p_obs;
   p_a = p_md->p_alls;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p_n, m_par[i], MAXNAMLEN + 1);
      p_n += MAXNAMLEN + 1;
   }
   for (j = 0; j < nk_par; ++j)
   {
      *p_a = j_par[ix++];
      ++p_a;
   }

   return;
}

/* Stores the chromosome number, the number of markers and the list of names. */

void stor_mrknam (MARKNAM_PTR p_mn, int chrom_num_included)
{
   int   i;
   char *p_n;

   if (chrom_num_included == CH)
      p_mn->chrom_no = j_par[0];
   else
      p_mn->chrom_no = 0;

   p_mn->no_marks = nm_par;

   p_n = p_mn->p_nams;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p_n, m_par[i], MAXNAMLEN + 1);
      p_n += MAXNAMLEN + 1;
   }

   return;
}

void stor_prbgam (PROBGAM_PTR p_pg, int cmp, int sc) /* stores data in the  */
                                        /* proband gamete node.  Also stores  */
                                        /* the array of program gametes in    */
                                        /* the proband gamete list            */
{
   int   i, ix = 0;
   PROBLIST_PTR p;

   if (cmp == CMP)
      p_pg->component = j_par[ix++];
   else
      p_pg->component = 1;

   if (sc == SC)
      p_pg->scoreset = j_par[ix++];
   else
      p_pg->scoreset = 1;

   p_pg->no_pgams = nm_par;

   p = p_pg->p_pglist;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p->nam, m_par[i], MAXNAMLEN + 1);
      p->gamete = j_par[ix++];
      ++p;
   }
   return;
}

void stor_indnam (INDNAM_PTR p_ind, int cmp) /* stores data in the named      */
                                        /* individuals node.                  */
                                        /* Also stores the names in a list    */
{
   int   i;
   char *p;

   if (cmp == CMP)
      p_ind->component = j_par[0];
   else
      p_ind->component = 1;

   p_ind->no_nams = nm_par;

   p = p_ind->p_namlist;
   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p, m_par[i], MAXNAMLEN + 1);
      p += MAXNAMLEN + 1;
   }
   return;
}

/* Stores the chromosome number and the window size in the locus window node. */

void stor_wind (LOCWND_PTR p_node, int chrom_num_included)
{
   int   ix = 0;

   if (chrom_num_included == CH)
      p_node->chrom_no = j_par[ix++];
   else
      p_node->chrom_no = 0;

   p_node->window = j_par[ix];

   return;
}

void stor_wndpat (WNDPAT_PTR p_pw, int cp, int sc) /* stores data in the    */
                                        /* window pattern node.  Also stores  */
                                        /* the window pattern labels in the   */
                                        /* list of patterns                   */
{
   int   i, ix = 0;

   if (cp == CMP)
      p_pw->component = j_par[ix++];
   else
      p_pw->component = 1;

   if (sc == SC)
      p_pw->scoreset = j_par[ix++];
   else
      p_pw->scoreset = 1;

   p_pw->no_pats = nk_par;

   for (i = 0; i < nk_par; ++i)
      *(p_pw->p_pats + i) = j_par[ix++];

   return;
}

/* Stores data in the kinship, ibreeding or 2-locus inbreeding node.  Also    */
/* stores the list of pedigree members in the names list.                     */

void stor_kinshp (KINS * p_kn, int cmp)
{
   int   i, ix = 0;
   char *p;

   if (cmp == CMP)
      p_kn->component = j_par[ix++];
   else
      p_kn->component = 1;

   p_kn->no_nams = nm_par;
   p = p_kn->p_nams;

   for (i = 0; i < nm_par; ++i)
   {
      strncpy (p, m_par[i], MAXNAMLEN + 1);
      p += MAXNAMLEN + 1;
   }

   return;
}

void stor_reco_frq (int *n, double *r)  /* stores the no. of recomb freqs    */
{                                       /* and the recomb freqs              */
   int   i;

   *n = nx_par;

   for (i = 0; i < nx_par; ++i)
      r[i] = d_par[i];

   return;
}

/* ------------- copy fnc:  copies string to malloc'd memory --------------- */

char *alloc_copy (char *text)           /* first gets memory, then copies    */
{
   char *pstrg;
   int   s;

   s = strlen (text) + 1;
   pstrg = malloc (s * sizeof (char *));
   alloc_chk (WHERE, pstrg);
   strncpy (pstrg, text, s);

   return pstrg;
}

/* ---------------- reset fncs:  reinitialize data to 0's ------------------ */

void reset_start_var (int m)            /* resets the starting value(s)      */
                                        /* for variances                     */
{
   int   j;

   for (j = 1; j <= psx->max_trait; ++j)
      if (!m)                           /* additive                          */
         SIGSQ (j, j) = 0.0;
      else                              /* residual                          */
         TAUSQ (j, j) = 0.0;
   if (!m)
      naddvar = 0;
   else
      nresvar = 0;
   return;
}

void reset_start_cov (int m)            /* resets the starting value(s)      */
{                                       /* for covariances                   */
   int   row, col;

   for (row = 2; row <= psx->max_trait; ++row)
      for (col = 1; col < row; ++col)
         if (!m)                        /* additive                          */
            SIGSQ (row, col) = 0.0;
         else                           /* residual                          */
            TAUSQ (row, col) = 0.0;
   if (!m)
      naddcov = 0;
   else
      nrescov = 0;
   return;
}

void reset_fit_var (int m)              /* resets the fit/nofit flags        */
{                                       /* for variances ( = 1)              */
   int   j;

   for (j = 1; j <= MAXQT; ++j)
      if (!m)                           /* additive                          */
         SIG_MODL (j, j) = 1;
      else                              /* residual                          */
         TAU_MODL (j, j) = 1;
}

void reset_fit_cov (int m)              /* resets the fit/nofit flags        */
{                                       /* for covariances                   */
   int   row, col;

   for (row = 2; row <= MAXQT; ++row)
      for (col = 1; col < row; ++col)
         if (!m)                        /* additive                          */
            SIG_MODL (row, col) = 1;
         else                           /* residual                          */
            TAU_MODL (row, col) = 1;
   if (!m)
      nfitadd = 0;
   else
      nfitres = 0;
   return;
}

void reset_set_eigen (void)             /* Frees the eigenvalue array.        */
{
   free ((void *) peigen);
   peigen = NULL;
   return;
}

void reset_set_pseudo (void)            /* Frees the pseudo-prior array.      */
{
   free ((void *) PseudoPrior);
   PseudoPrior = NULL;
   return;
}

/* -------- set fncs:  store input parameters into data structures ---------  */

void start_trait_eff (void)             /* Stores the starting value(s) for   */
{                                       /* a fixed effect for the trait.      */
   TRAITEFF_PTR p_t;
   FIXEDEFF_PTR p_f;

   /* Find the trait_eff_node for the selected trait number.                  */
   p_t = first_trait_eff_node;
   while (p_t && p_t->trait_num < j_par[0])
      p_t = p_t->p_next;

   /* If there is not already a trait_eff_node for the selected trait, create */
   /* a temporary trait_eff_node for this trait.                              */
   if (!p_t || p_t->trait_num != j_par[0])
      p_t = alloc_tr_eff_node (0);

   p_f = alloc_fixed_eff_node (p_t, 1);
   alloc_fixed_eff_array (p_f);
   stor_fixed_eff_array (p_f);

   return;
}

void start_trait_mean (void)            /* Stores the starting value for the  */
{                                       /* mean for the trait.                */
   TRAITEFF_PTR p_t;

   /* Find a trait_eff_node for the selected trait number if one exists.      */
   p_t = first_trait_eff_node;
   while (p_t && p_t->trait_num < j_par[0])
      p_t = p_t->p_next;

   /* Warn the user if the starting value for the mean is being overridden.   */
   if (p_t && p_t->trait_num == j_par[0])
      if (p_t->start_tr_mean_inp)
         err_typ = ERR_MNR;

   /* If there is not already a trait_eff_node for the selected trait, create */
   /* a temporary trait_eff_node for this trait.                              */
   if (!p_t || p_t->trait_num != j_par[0])
      p_t = alloc_tr_eff_node (0);

   p_t->start_tr_mean_inp = Y;
   p_t->start_tr_mean = d_par[0];

   return;
}

void set_start_var (int m)              /* Stores the starting value(s) for   */
{                                       /* variances.                         */
   int   j;

   for (j = 1; j <= nx_par; ++j)
      if (!m)                           /* Additive                           */
         SIGSQ (j, j) = d_par[j - 1];
      else                              /* Residual                           */
         TAUSQ (j, j) = d_par[j - 1];
   if (!m)
      naddvar = nx_par;
   else
      nresvar = nx_par;

   return;
}

void set_start_cor (int m)              /* Stores the starting value(s) for   */
{                                       /* correlations.                      */
   int   j, row, col;

   for (row = 2, j = 0; j < nx_par; ++row)
      for (col = 1; col < row && j < nx_par; ++col, ++j)
         if (!m)                        /* Additive                           */
            SIGSQ (row, col) = d_par[j];
         else                           /* Residual                           */
            TAUSQ (row, col) = d_par[j];
   if (!m)
   {
      naddcov = nx_par;
      add_cor_flag = Y;
   }
   else
   {
      nrescov = nx_par;
      res_cor_flag = Y;
   }
   return;
}

void set_start_cov (int m)              /* Stores the starting value(s) for   */
{                                       /* covariances.                       */
   int   j, row, col;

   for (row = 2, j = 0; j < nx_par; ++row)
      for (col = 1; col < row && j < nx_par; ++col, ++j)
         if (!m)                        /* Additive                           */
            SIGSQ (row, col) = d_par[j];
         else                           /* Residual                           */
            TAUSQ (row, col) = d_par[j];
   if (!m)
   {
      naddcov = nx_par;
      add_cor_flag = 0;
   }
   else
   {
      nrescov = nx_par;
      res_cor_flag = 0;
   }
   return;
}

void set_fit_cov (int m)                /* Stores the fit/no-fit flag for     */
{                                       /* covariances.                       */
   int   j, row, col;

   for (row = 2, j = 0; j < nk_par; ++row)
      for (col = 1; col < row && j < nk_par; ++col, ++j)
         if (!m)                        /* Additive                           */
         {
            if (j_par[j] != 0)
               SIG_MODL (row, col) = 1;
            else
               SIG_MODL (row, col) = 0;
         }
         else
         {                              /* Residual                           */
            if (j_par[j] != 0)
               TAU_MODL (row, col) = 1;
            else
               TAU_MODL (row, col) = 0;
         }
   if (!m)
      nfitadd = nk_par;
   else
      nfitres = nk_par;

   return;
}

void set_set_eigen (void)               /* allocates the array for the       */
                                        /* eigenvalues and stores the        */
                                        /* starting values                   */
{
   int   k;

   peigen = malloc ((nx_par + 1) * sizeof (double));
   alloc_chk (WHERE, peigen);
   for (k = 1; k <= nx_par; ++k)
      peigen[k] = d_par[k - 1];
   n_eigen = nx_par;

   return;
}

void set_set_pseudo (void)              /* sets the pseudo-prior input flag  */
                                        /* allocates the array for the       */
                                        /* pseudo-priors and stores the      */
                                        /* values                            */
{
   int   k;

   InPseudo = Y;
   PseudoPrior = malloc ((nx_par) * sizeof (double));
   alloc_chk (WHERE, PseudoPrior);
   for (k = 0; k < nx_par; ++k)
      PseudoPrior[k] = d_par[k];
   n_pseudo = nx_par;

   return;
}
