/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/******************************************************************************/
/*                                                                            */
/*  Setup function for Morgan main programs not requiring a pedigree:         */
/*    procParIndivs().                                                        */
/*                                                                            */
/******************************************************************************/

#include "parsie.h"

/******************************************************************************/
/*                                                                            */
/* Process the parameter file and set corresponding globals                   */
/* (these start with capital letters, and are found in globs.h).              */
/*                                                                            */
/******************************************************************************/

void procParIndivs (int argc, char **argv, char *ppn)
{
   /* Parse input parameter statements.                                       */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* Set the global variable indicating we are using an individuals file.    */
   FromIndivsFile = TRUE;

   /* Check statements for completeness and consistency.                      */
   /* Set up global parameters and data structures.                           */

   if (IS_IBD_HAPLO)
   {
      if (!proc_auto ())
         exit (EXIT_FAILURE);
   }

   if (IS_GL_LODS)
   {
      if (!proc_ibd ())
         exit (EXIT_FAILURE);
   }

   /* Check out the individuals file record options.                          */
   if (!test_indfil ())                 /* Test record options.               */
      if (proc_err ())
         exit (EXIT_FAILURE);

   /* Set up the output scores file parameters and call proc_end.             */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

#if 0
   /* Currently there are no Morgan programs not requiring a pedigree which   */
   /* need to set up seed parameters.  If at some time in the future such a   */
   /* program would be needed, then these lines need to be uncommented out.   */

   /* Set up seed parameters.                                                 */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);
#endif

   prntStatusCP ();

   return;
}
