/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 *   Setup functions for Morgan main programs:
 *     procParStat()
 *     setupBasic()
 */

#include "parsie.h"
#include "rans.h"

/**
 * Process the parameter file and set corresponding globals
 * (these start with capital letters, and are found in globs.h).
 */

void procParStat (int argc, char **argv, char *ppn)
{

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* check statements for completeness and consistency.
    * Set up global parameters and data structures     */

   if (IS_MARKERDROP || IS_LM_AUTO || IS_LM_IBDTESTS || IS_LM_MAP || IS_LM_PVAL)
   {
      if (!proc_auto ())
         exit (EXIT_FAILURE);
   }

   if (IS_LM_BAYES || IS_LM_LODS || IS_LM_MARKERS || IS_LM_SCHNELL
       || IS_LM_TWOQTL)
   {
      if (!proc_lods ())
         exit (EXIT_FAILURE);
   }

#if 0
   /* check for 2-D graphical display option */
   if (IS_LM_AUTO)
      if (!proc_disp ())
         exit (EXIT_FAILURE);
#endif

   /* set up for PolyEM programs */
   if (IS_BIVAR || IS_MULTIVAR || IS_UNIBIG || IS_UNIVAR)
   {
      if (!proc_poly ())
         exit (EXIT_FAILURE);
   }

   /* set up pedigree file parameters */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

   /* set up seed parameters */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   return;
}

/******************************************************************************/
/*                                                                            */
/* The init_rans_seeds subroutine initializes the appropriate seeds for the   */
/* rans() subroutine.                                                         */
/*                                                                            */
/******************************************************************************/

void init_rans_seeds (void)
{
   if (IS_LM_IBDTESTS && PermutEnable)
      setSeeds (PERM);
   else
      setSeeds (SAMPL);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The first argument to this subroutine is a null-terminated string.  This   */
/* string consists of option code letters and is used to indicate which parts */
/* of the subroutine the caller wants executed.                               */
/*                                                                            */
/* If one of the option code letters is 'A', then the calling program wants   */
/* this subroutine to call proc_auto to perform semantic analysis.            */
/*                                                                            */
/* If one of the option code letters is 'R', then the calling program will be */
/* using the rans() subroutine, which needs to have its seeds initialized.    */
/*                                                                            */
/******************************************************************************/

void setupBasic (char *options, int argc, char **argv, char *ppn)
{
   initMorgProg (argv);

   /* Parse input parameter statements.                                       */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   if (strchr (options, 'A'))
   {
      /* Check statements for completeness and consistency.                   */
      /* Set up global parameters and data structures.                        */
      if (!proc_auto ())
         exit (EXIT_FAILURE);
   }
   else
   {
      if (!test_auto_pars ())           /* Check computational parameters.    */
         if (proc_err ())
            exit (EXIT_FAILURE);
   }

   /* If the calling program requires use of a pedigree file this would be    */
   /* where the proc_fils subroutine would need to be called.  In any case,   */
   /* we still need to call proc_end so the program can be terminated in case */
   /* there were any errors requiring ending the program.                     */

   if (!proc_end ())
      exit (EXIT_FAILURE);

   if (strchr (options, 'R'))
   {
      /* Set up seed parameters.                                              */
      if (!proc_seeds ())
         exit (EXIT_FAILURE);
   }

   prntStatusCP ();

   if (strchr (options, 'R'))
   {
      /* Call the init_rans_seeds subroutine so that the seeds needed by the  */
      /* rans() function can be initialized.                                  */

      init_rans_seeds ();
   }

   return;
}
