/*************************************************************************
  program name: translink.c    (modified from markerdrop: EAT 2009)

  Program to translate MORGAN-style input files, including
    pedigree, trait, and marker data, marker map etc., into a
  LINKAGE-style input file;
    current format of dat file is for linkmap program.
    if individuals are output by index, then are "named" 1 through n,
    but recoding may still be required for some programs, if there
    are multiple component pedigrees.

  Currently runs under lm_auto name and uses the lm_auto tloc position
    to define the physical locus ordering, and initial positions.

**************************************************************************/

#define PNAM  "lm_auto"
#include "autozyg.h"

int main (int argc, char *argv[])
{
   int   i, j, c, k, cc, ccc;
   int   numnam, siblinks, dexit;       /* numnam= 1 is names are numeric   */
   /* In this case, output by MORGAN index adjusted within component  */
   /* siblinks=1; sib links required   */
   /* dexit= 1; output by newdex       */
   int   sxsp;                          /* indicator of sex-specific maps   */
   int   neval;                         /* number of points for lod score   */
   int  *first_pers;                    /* flag for 1 st persom in each cpnt */

   Ind  *dad, *mom;
   int  *newdex;                        /* indexing order following Marlists */
   int  *cmpdex;                        /* indexing order by component   */

   TRAITPARS_PTR p_traitp;              /* Pointers to trait parameters.      */
   TLOCPARS_PTR p_tlocp;

   FILE *fptr, *iXtra;

   NSw = setupAutozyg (argc, argv, PNAM, 0);

   first_pers = malloc ((nCompnt + 1) * sizeof (int)); /* compnts indexed from 1 */
   for (c = 1; c <= nCompnt; ++c)
      first_pers[c] = 1;

   /* Set pointers to be able to refer to incomplete penetrance parameters or */
   /* to genotype mean parameters.                                            */

   if (TraitTyp == DISCRETE || TraitTyp == QUANTITATIVE)
   {
      p_traitp = P_trait_pars;
      p_tlocp = p_traitp->p_tloc_pars;
   }

   /* set up additional flags etc.  */
   if (!(iXtra = open_fil (XTRA)))
      give_up (ERR_qit);
   fscanf (iXtra, "%d %d %d %d", &numnam, &siblinks, &dexit, &neval);

/*  Can reindex according to the Marlists if desired */
   if (siblinks)
   {
      newdex = malloc (pedsize * sizeof (int));
      for (i = 0; i < pedsize; ++i)
         newdex[i] = 0;

      setup_siblinks (newdex);
   }

/* First we will output the pedigree, trait, and marker data */

   if (!(fptr = fopen ("pedfile.dat", "w")))
      give_up (ERR_qit);

/* If outputting by MORGAN index, will reindex each component
    according to output order requested  */
   if (numnam)
   {
      cmpdex = malloc (pedsize * sizeof (int));
      for (i = 0; i < pedsize; ++i)
         cmpdex[i] = 0;

      cc = 0;                           /* count of inds in previous cmpt */
      ccc = 0;                          /* cmpnt of previous person in list */
      for (k = 0; k < pedsize; k++)     /* go through all individuals */
      {
         if (dexit)
            i = newdex[k];              /* index of k th person to be output */
         else
            i = k;                      /* output by MORGAN index order      */

         c = nghd[i].component;
         if (ccc == 0)
         {
            cmpdex[i] = k + 1;
            ccc = c;
            cc = 0;                     /* first component */
         }
         else
         {
            if (c == ccc)
               cmpdex[i] = k + 1 - cc;  /* same component */
            else
            {
               ccc = c;                 /* new component  */
               cc = k;
               cmpdex[i] = k + 1 - cc;
            }
         }
      }
   }

/* We will now output by chosen order over all components   */
   for (k = 0; k < pedsize; k++)        /* go through all individuals */
   {
      if (dexit)
         i = newdex[k];                 /* index of k th person to be output */
      else
         i = k;                         /* output by MORGAN index order      */

      c = nghd[i].component;            /* pedigree component  of i */

      fprintf (fptr, " %3d", c);
      /* basic trio  as names or index */
      if (!numnam)
         fprintf (fptr, " %5s", nghd[i].self);
      else
         fprintf (fptr, " %5d", cmpdex[i]);

      mom = nghd[i].ma;
      dad = nghd[i].pa;
      if ((!numnam) || (isNullNode (dad))) /* parents offspring links */
         fprintf (fptr, " %5s %5s", dad->self, mom->self);
      else
         fprintf (fptr, " %5d %5d", cmpdex[dad->index], cmpdex[mom->index]);

      if (siblinks)                     /* next-sib links if desired */
      {
         if ((!numnam) || (isNullNode (nghd[i].first_kid)))
            fprintf (fptr, " %5s", (nghd[i].first_kid)->self);
         else
            fprintf (fptr, " %5d", cmpdex[nghd[i].first_kid->index]);

         if ((!numnam) || (isNullNode (nghd[i].next_patsib)))
            fprintf (fptr, " %5s", (nghd[i].next_patsib)->self);
         else
            fprintf (fptr, " %5d", cmpdex[(nghd[i].next_patsib)->index]);

         if ((!numnam) || (isNullNode (nghd[i].next_matsib)))
            fprintf (fptr, " %5s", (nghd[i].next_matsib)->self);
         else
            fprintf (fptr, " %5d", cmpdex[(nghd[i].next_matsib)->index]);
      }
      /* gender and proband status  */
      /* probstatus 1 for first person in cmpt c */
      fprintf (fptr, " %1d", nghd[i].sex);
      fprintf (fptr, " %1d", first_pers[c]);
      first_pers[c] = 0;

      /* Affectation status etc.   */
      if (TraitTyp == DISCRETE)
      {
         if (nghd[i].d_dat)
            fprintf (fptr, " %1d", nghd[i].d_dat[0]);
         else
            fprintf (fptr, " 0");
      }
      else if (TraitTyp == QUANTITATIVE)
      {
         if (nghd[i].q_dat)
            fprintf (fptr, " %f", nghd[i].q_dat[0]);
         else
            fprintf (fptr, " -999.99");
      }

      /* Marker data for marker alleles */
      if (nghd[i].g_dat)
      {
         for (j = 1; j < NLoci; ++j)
            fprintf (fptr, "  %2d %2d", nghd[i].g_dat[2 * j],
                     nghd[i].g_dat[2 * j + 1]);
      }
      else
         for (j = 1; j < NLoci; ++j)
            fprintf (fptr, "   0  0");

      fflush (fptr);

      fprintf (fptr, "\n");
   }
   fclose (fptr);                       /* Finished output of pedigree file.  */

   /* now output the marker info etc.  */
   if (!(fptr = fopen ("datafile.dat", "w")))
      give_up (ERR_qit);

/* first line:  no.loci, risk loc, X-linked, Prog code ??,
    superlink complex loci code  */
/* note; trait locs (0 in MORGAN) becomes first locus in output */
   fprintf (fptr, " %d  0   0  4", NLoci);
   fprintf (fptr, "       << see fastlink documentation; 4 is linkmap code\n");
   fflush (stdout);

/* second line: Mutation locus, male rate, female rate, hap freq  */
   fprintf (fptr, "  0  0  0  0 ");
   fprintf (fptr, "       << see fastlink documentation\n");
   fflush (stdout);

/* third line; physical order of the loci; trait locus 1 is right of
   marker locus Loc0, which is locus Loc0+1  */
   if (Loc0 == 0)                       /* trait locus is first */
   {
      for (j = 1; j <= NLoci; ++j)
         fprintf (fptr, "  %3d", j);
   }
   else
   {
      for (j = 1; j <= Loc0; ++j)
         fprintf (fptr, "  %3d", j + 1);
      fprintf (fptr, "  1");
      for (j = Loc0 + 2; j <= NLoci; ++j)
         fprintf (fptr, "  %3d", j);
   }
   fprintf (fptr, "       << physical locus ordering\n");
   fflush (fptr);

/*  Next we will output the locus info for each locus */
   if (TraitTyp == DISCRETE)
   {
      fprintf (fptr, "  1  2  ");       /* discrete trait; 2 alleles */
      fprintf (fptr, "       << discrete trait locus, 2 alleles\n");
      fprintf (fptr, " %6.4f %6.4f ", Qq[0][1], Qq[0][2]);
      fprintf (fptr, "       << allele frequencies\n");
      fprintf (fptr, "  1 ");           /* single liability class  */
      fprintf (fptr, "                << single liability class\n");
      fprintf (fptr, " %6.4f %6.4f %6.4f ", p_tlocp->penetrances[0],
               p_tlocp->penetrances[1], p_tlocp->penetrances[2]);
      fprintf (fptr, "       << penetrances\n");
      fflush (fptr);
   }
   else if (TraitTyp == QUANTITATIVE)   /* need lm_auto to take quant trait */
   {
      fprintf (fptr, "  0  2 ");        /* qu-trait ; 2 alleles */
      fprintf (fptr, "       << quantitative trait locus, 2 alleles\n");
      fprintf (fptr, " %6.4f %6.4f ", Qq[0][1], Qq[0][2]);
      fprintf (fptr, "       << allele frequencies\n");
      fprintf (fptr, "  1 ");           /* single quantitative variable */
      fprintf (fptr, "                << single quantitative variable\n");
      fprintf (fptr, " %6.4f %6.4f %6.4f ", p_tlocp->means[0], p_tlocp->means[1], p_tlocp->means[2]); /* Genotypic means.     */
      fprintf (fptr, "         << genotypic means\n");
      fprintf (fptr, " %6.4f ", p_traitp->resi_var); /* Residual variance.    */
      fprintf (fptr, "         << residual variance\n");
      fprintf (fptr, " 1.0 ");          /* The variance multiplier.           */
      fprintf (fptr,
               "         << the var multiplier; see fastlink documentation\n");
      fflush (fptr);
   }

   for (j = 1; j <= NMarks; ++j)
   {
      fprintf (fptr, "  3  %d  ", NAll[j]); /* marker locus  j-1 */
      fprintf (fptr, "       << marker locus and number of alleles\n");
      for (k = 1; k <= NAll[j]; ++k)
         fprintf (fptr, " %6.4f", Qq[j][k]);
      fprintf (fptr, "       << allele frequencies\n");
      fflush (stdout);
   }

/*  Next we need the recombination info etc    */
/*  To check if maps sex-specific check equality of Marek[0][j], Marek[1][j] */
   sxsp = 2;
   for (j = 1; j < NMarks; ++j)
      if (fabs (Marek[0][j] - Marek[1][j]) > 1.0E-06)
         break;
   if (j == NMarks)
      sxsp = 0;

   fprintf (fptr, "  %3d   0", sxsp);
   fprintf (fptr, "       << sex-specific map, interference\n");

   if (!Loc0)                           /* trait locus is first */
   {
      if (sxsp == 2)                    /* print male map  */
      {
         fprintf (fptr, "  0.5");       /* initial tloc position */
         for (j = 1; j < NMarks; ++j)
            fprintf (fptr, " %5.4f", Marek[1][j]);
         fprintf (fptr, "       << initial male map\n");
      }

      /* female or average map */
      fprintf (fptr, " %f", 0.5);       /* initial tloc position */
      for (j = 1; j < NMarks; ++j)
         fprintf (fptr, " %5.4f", Marek[0][j]);
      fprintf (fptr, "       << initial female/avg/both map\n");
      fprintf (fptr, "    1   %f  %d", Rec0[0], neval); /* final line  */
      fprintf (fptr, "       << locus 1 varies; final recmb; number evals");
      fprintf (fptr, "\n");
   }
   else if (Loc0 == NMarks)             /* trait locus is last */
   {
      if (sxsp == 2)                    /* print male map  */
      {
         for (j = 1; j < NMarks; ++j)
            fprintf (fptr, " %5.4f", Marek[1][j]);
         fprintf (fptr, " %f", Rec0[0]); /* initial tloc position */
         fprintf (fptr, "       << initial male map\n");
      }

      /* female or average map */
      for (j = 1; j < NMarks; ++j)
         fprintf (fptr, " %5.4f", Marek[0][j]);
      fprintf (fptr, " %f", Rec0[0]);   /* initial tloc position */
      fprintf (fptr, "       << initial female/avg/both map\n");
      fprintf (fptr, "    1   0.5  %d", neval); /* final line  */
      fprintf (fptr, "       << locus 1 varies; final recmb; number evals");
      fprintf (fptr, "\n");
   }
   else
   {                                    /* interval is Loc0 to Loc0+1 */
      if (sxsp == 2)                    /* print male map  */
      {
         for (j = 1; j < Loc0; ++j)
            fprintf (fptr, " %5.4f", Marek[1][j]);
         fprintf (fptr, "  0.0");
         for (j = Loc0; j < NMarks; ++j);
         fprintf (fptr, " %5.4f", Marek[1][j]);
         fprintf (fptr, "       << initial male map\n");
      }
      for (j = 1; j < Loc0; ++j)
         fprintf (fptr, " %5.4f", Marek[0][j]);
      fprintf (fptr, "  0.0");
      for (j = Loc0; j < NMarks; ++j)
         fprintf (fptr, " %5.4f", Marek[0][j]);
      fprintf (fptr, "       << initial female/avg/both map\n");
      fprintf (fptr, "    1   %f  %d", Marek[0][Loc0], neval);
      /* final line  */
      fprintf (fptr, "       << locus 1 varies; final recom ; nevals");
      fprintf (fptr, "\n");
   }

   fflush (fptr);
   fclose (fptr);

   exit (EXIT_SUCCESS);
}

/** @} */
