/******************************************************************************/
/*                                                                            */
/* openChkInd(): Opens the output individuals file.                           */
/*                                                                            */
/******************************************************************************/

#include "pedchk.h"

FILE *openChkInd (void)
{
   FILE *pOF;

   if (!(pOF = open_fil (OIND)))
      return NULL;

   fprintf (pOF, "input pedigree size %d\n", nInd);
   fprintf (pOF, "input individuals record names 2");
   if (NStdOut - 3 + NNInts > 0)
      fprintf (pOF, " integers %d", NStdOut - 3 + NNInts);
   if (NReals)
      fprintf (pOF, " reals %d\n", NReals);
   fprintf (pOF, "\n");

   fprintf (pOF, "********\n");

   return pOF;
}

/******************************************************************************/
/*                                                                            */
/* writChkInd(): Writes the output individuals file.                          */
/*                                                                            */
/* The output individuals file contains the same data as the input pedigree   */
/* file, except that it contains each individual's component number in place  */
/* of father's and mother's names.  In other words, the individuals file      */
/* contains each individual's name and their trait data, without any pedigree */
/* structure.                                                                 */
/*                                                                            */
/******************************************************************************/

void writChkInd (FILE * pOF)
{
   KEYS_PP ppK = ppKeyTab;
   PED_PTR pP;
   PEDI_PTR pPI;

   char  temp_buff[12];                 /* A temporary buffer to format the   */
   /*                                      highest component number into.     */

   int   nc;                            /* Character length of highest        */
   /*                                      component number required.         */

   int   comp_no;                       /* Individual's component number.     */
   int   ixInd;                         /* Current individual's subscript.    */

   int   i1;                            /* For loop variable.                 */
   int   i2;                            /* Subscript in output order array.   */

   /* Initialize the integer variable, nc, to the number of characters in the */
   /* highest component number required in the pedigree.                      */

   sprintf (temp_buff, "%d", nCompnt);
   nc = strlen (temp_buff);

   /* Now find field widths for each of the integers.                         */

   for (i2 = 0; i2 < NNInts; ++i2)
      widPInt[i2] = intWid (minPInt[i2], maxPInt[i2]);

   /* For each individual:                                                    */
   for (i1 = 0; i1 < nInd; ++i1)
   {
      ixInd = (*ppK)->ix - 1;           /* Get index for the next individual. */
      pP = pPedDat + ixInd;             /* Set the individual's pedigree data */
      /*                                   pointer.                           */

      for (i2 = 0; i2 < NStdOut; ++i2)
      {
         switch (StdOut[i2])
         {
             case NAM:
                {
                   fprintf (pOF, "%*s", widPNam, pP->nam);
                   break;
                }
             case PA:
                {
                   /* Replace the name with individual's component number.    */
                   comp_no = (*ppK)->comp; /* Individual's component number.  */
                   fprintf (pOF, "  %*d", nc, comp_no);
                   break;
                }
             case SEX:
                {
                   fprintf (pOF, "  %d", pP->sex);
                   break;
                }
             case OBS:
                {
                   fprintf (pOF, "  %d", pP->obsrvd);
                   break;
                }
             default:
                break;
         }
      }

      fprintf (pOF, " ");               /* Output the integers.               */

      pPI = pIntDat + NNInts * ixInd;
      for (i2 = 0; i2 < NNInts; ++i2)
         fprintf (pOF, " %*d", widPInt[i2], *(pPI + i2));

      if (NReals)                       /* Output the reals.                  */
         fprintf (pOF, "%s", *(ppRSDat + ixInd));

      fprintf (pOF, "\n");              /* End the current record.            */
      fflush (pOF);

      ++ppK;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* mak_oindfil(): Writes out the individuals data to the output individuals   */
/* file.                                                                      */
/*                                                                            */
/******************************************************************************/

int mak_oindfil (void)
{
   FILE *oIndF;

   if (!get_filent (OIND)->fname)       /* Output individuals filename given? */
      return OK;

   if (!(oIndF = openChkInd ()))        /* Open the output individuals file.  */
      return !OK;

   writChkInd (oIndF);                  /* Write out the individuals data.    */
   fclose (oIndF);

   return OK;
}
