/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "rans.h"

void get_seeds (const char *fname)
{
   FILE *file;
   uint32 cseed, tseed;

   if (fname == NULL)
      fname = ".Random.seed";

   if ((file = fopen (fname, "r")) == NULL)
   {
      fprintf (stderr, "Cannot open file %s for read\n", fname);
      perror ((char *) NULL);
      exit (1);
   }

   if (fscanf (file, "%x %x", &cseed, &tseed) != 2)
   {
      fprintf (stderr, "Cannot read seeds from file %s\n", fname);
      exit (1);
   }

   fclose (file);

   printf ("\n Hexadecimal seeds %08x %08x\n", cseed, tseed);

   if (cseed % 2 == 0)
   {
      fprintf (stderr, " cseed must be odd\n");
      exit (1);
   }
   if (tseed == 0)
   {
      fprintf (stderr, " tseed must be nonzero\n");
      exit (1);
   }

   setsd (cseed, tseed);
}

void put_seeds (const char *fname)
{
   FILE *file;
   uint32 cseed, tseed;

   getsd (&cseed, &tseed);

   printf ("\n %08x %08x\n", cseed, tseed);

   if (fname == NULL)
      fname = ".Random.seed";

   if ((file = fopen (fname, "w")) != NULL)
   {
      fprintf (file, " %08x %08x\n", cseed, tseed);
      fclose (file);
      return;
   }
   else
   {
      fprintf (stderr, "Cannot open %s for write\n", fname);
      perror ((char *) NULL);
      exit (1);
   }
}
