/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/**********************************************************************
*  Routine to take a proposed switch of segregation, and update gene descent
*   at a given locus, by traversing gene tree down from that point.
*
*  Modified, June 1997, to have a gene_update segment separate from
*   scoring of the IBD
***********************************************************************/

#include "msample.h"

void  traverse (int j, int gene, int k, int i, int sx);

void gene_update (int j, int i, int k)
/* j is locus, i is non-founder being switched, k is what switch, */
{
   int   par_dex, gene, sw;
   if (k == 0)
   {
      sw = xx0[j][i];
      par_dex = nghd[i].ma->index;
      if (sw == 0)
         gene = gg0[j][par_dex];
      else
         gene = gg1[j][par_dex];
      gg0[j][i] = gene;
   }
   else
   {
      sw = xx1[j][i];
      par_dex = nghd[i].pa->index;
      if (sw == 0)
         gene = gg0[j][par_dex];
      else
         gene = gg1[j][par_dex];
      gg1[j][i] = gene;
   }

/* Now descend tree, changing any descendant gene to "gene" */

   traverse (j, gene, k, i, 2 - nghd[i].sex);
}

int ibd_update (int j, int i, int k)
/* j is locus, i is non-founder being switched, k is what switch, */
{
   gene_update (j, i, k);

/* Return 0 if, after switch, proband will not be IBD;  gene if IBD    */

   if ((gg0[j][pedsize - 1] * gg1[j][pedsize - 1]) == 0)
      return (0);
   if (gg0[j][pedsize - 1] == gg1[j][pedsize - 1])
      return (gg0[j][pedsize - 1]);
   return (0);

}

void traverse (int j, int gene, int k, int i, int sx)
{
   Marlist *mar_ptr;
   Indlist *kid_ptr;

   Ind  *kid;

   int   kid_indx;

   mar_ptr = nghd[i].marriages;

   while (mar_ptr != NULL)
   {                                    /* has at least one spouse */
      kid_ptr = mar_ptr->current->kids;

      while (kid_ptr != NULL)
      {                                 /* has at least one kid */
         kid = kid_ptr->current;
         kid_indx = kid->index;
         if ((sx == 0) && (xx0[j][kid_indx] == k))
         {
            gg0[j][kid_indx] = gene;
            traverse (j, gene, 0, kid_indx, 2 - nghd[kid_indx].sex);
         }
         if ((sx == 1) && (xx1[j][kid_indx] == k))
         {
            gg1[j][kid_indx] = gene;
            traverse (j, gene, 1, kid_indx, 2 - nghd[kid_indx].sex);
         }

         kid_ptr = kid_ptr->next;       /* next kid, if any */
      }

      mar_ptr = mar_ptr->next;          /* next spouse, if any */
   }

   return;
}
