/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

#include "stuff.h"

/******************************************************************************/
/*                                                                            */
/* The free_locnams subroutine frees up the memory allocated for TraitNam,    */
/* TLocNam and MarkNam.                                                       */
/*                                                                            */
/******************************************************************************/

void free_locnams (void)
{
   long  x1;

   if (TraitNam)
   {
      for (x1 = 0; x1 < NTraits; ++x1)
         free (TraitNam[x1]);
      free (TraitNam);
      TraitNam = NULL;
   }

   if (TLocNam)
   {
      for (x1 = 0; x1 < NTLocs; ++x1)
         free (TLocNam[x1]);
      free (TLocNam);
      TLocNam = NULL;
   }

   if (MarkNam)
   {
      for (x1 = 0; x1 < NMarks; ++x1)
         free (MarkNam[x1]);
      free (MarkNam);
      MarkNam = NULL;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_auto subroutine frees up the memory allocated for proc_auto.      */
/*                                                                            */
/******************************************************************************/

void free_auto (void)
{
   int   c, s, i, j;

   /* Free the data structures defined in globs.h by proc_auto for
      the autozyg programs. */

   if (NScSet)                          /* Tests for proband gamete sets.     */
   {
      for (c = 0; c < NScCmpAlloc; ++c) /* Loops through components.          */
      {
         if (NPrGam[c])                 /* Tests for sets for this component. */
         {
            for (s = 0; s < NScSet[c]; ++s) /* Loops through score sets.      */
            {
               if (ProbIx[c][s])        /* Frees list of proband indices.     */
                  free (ProbIx[c][s]);
               if (ProbGam[c][s])       /* Frees list of gametes.             */
                  free (ProbGam[c][s]);
               if (ProbNam[c][s])
               {
                  for (i = 0; i < NPrGam[c][s]; ++i) /* Loops through gametes */
                     free (ProbNam[c][s][i]); /* Frees each proband name.     */
                  free (ProbNam[c][s]);
               }
            }

            free (ProbIx[c]);
            ProbIx[c] = NULL;
            free (ProbGam[c]);
            ProbGam[c] = NULL;
            free (ProbNam[c]);
            ProbNam[c] = NULL;
            free (NPrGam[c]);
            NPrGam[c] = NULL;
            free (ScSetNo[c]);
            ScSetNo[c] = NULL;
         }

         if (NPtLab && NPtLab[c])
         {
            if (PatLab[c])
            {
               for (s = 0; s < NScSet[c]; ++s) /* Loops through score sets.   */
               {
                  if (NPtLab[c][s] && PatLab[c][s])
                  {
                     free (PatLab[c][s]);
                     PatLab[c][s] = NULL;
                  }
               }

               free (PatLab[c]);
               PatLab[c] = NULL;
            }

            free (NPtLab[c]);
            NPtLab[c] = NULL;
         }
      }

      free (ProbIx);
      ProbIx = NULL;
      free (ProbGam);
      ProbGam = NULL;
      free (ProbNam);
      ProbNam = NULL;
      free (NPrGam);
      NPrGam = NULL;
      free (ScSetNo);
      ScSetNo = NULL;
      free (NScSet);
      NScSet = NULL;

      free (PatLab);
      PatLab = NULL;
      free (NPtLab);
      NPtLab = NULL;
   }

   if (NObs && ObsMark)                 /* Tests for observed individuals.    */
   {
      for (i = 0; i < NObs; ++i)
      {
         for (j = 0; j < NMarks; ++j)
            free (ObsMark[i][j]);
         free (ObsNam[i]);
         free (ObsMark[i]);
      }

      free (ObsNam);
      ObsNam = NULL;
      free (ObsIx);
      ObsIx = NULL;
      free (ObsMark);
      ObsMark = NULL;
   }

   return;
}

void free_rescue_filenames (char **names)
{
   /* Free up memory used to hold the 2 alternating output rescue file names  */
   /* if this capability has been used by lm_auto, lm_lods or lm_schnell.     */

   if (names[0])
   {
      free (names[0]);
      names[0] = NULL;
   }

   if (names[1])
   {
      free (names[1]);
      names[1] = NULL;
   }

   return;
}
