/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 NAME
      print_S, print_S_marker, print_mc_eval

 SYNOPSIS

      #include "twoqtl.h"
      void print_S (char **S_0, char **S_1, int n_locus, int comp_no);
      void print_S_marker (int comp_no);
      void print_mc_eval (mc_eval_t out);

 DESCRIPTION
      Print intermediate results in lm_twoqtl
      (just for testing and debugging, not for the ordinary user).

 YJS 07/02/2005.

*/

#include "twoqtl.h"

void print_S (char **S_1, char **S_0, int n_locus, int comp_no)
{
   int   l_ndx, i1;

   fprintf (stderr, "\n S_trait of Component %2d", comp_no);

   for (l_ndx = 0; l_ndx < n_locus; ++l_ndx)
   {
      fprintf (stderr, "\n %2d", l_ndx);
      for (i1 = 0; i1 < size_c[comp_no]; ++i1)
         fprintf (stderr, " %2d%2d", (int) S_1[l_ndx][i1],
                  (int) S_0[l_ndx][i1]);
   }
   fprintf (stderr, "\n");

   return;
}

void print_S_marker (int comp_no)
{
   int   l_ndx, i1;

   fprintf (stderr, "\n S_marker of Component %2d", comp_no);

   /* Note: array index into xx1, xx0 begins after the tlocs (for markers).   */
   for (l_ndx = NTLocs; l_ndx < NLoci; ++l_ndx)
   {
      fprintf (stderr, "\n %2d", l_ndx);
      for (i1 = ibase[comp_no]; i1 < (ibase[comp_no] + size_c[comp_no]); ++i1)
         fprintf (stderr, " %2d%2d", xx1[l_ndx][i1], xx0[l_ndx][i1]);
   }
   fprintf (stderr, "\n");

   return;
}

void print_mc_eval (mc_eval_t out)
{
   double *foo_alt, *max_alt;
   int   i1, i2;

   n_loci_1 = out.n_loci_1;
   n_loci_2 = out.n_loci_2;

   fprintf (stderr, "print mc_eval (n_loci_1 = %2d, n_loci_2 = %2d)\n",
            n_loci_1, n_loci_2);

   for (i1 = 0; i1 < n_loci_1; ++i1)
   {
      foo_alt = out.foo_alt[i1];
      max_alt = out.max_alt[i1];

      fprintf (stderr, "%2d ", i1);
      for (i2 = 0; i2 < n_loci_2; ++i2)
         fprintf (stderr, "%e %e ", foo_alt[i2], max_alt[i2]);
      fprintf (stderr, "\n");
   }
   fprintf (stderr, "nl %e %e \n", out.foo_null, out.max_null);

   return;
}
